/*
 * getput.c --
 *
 *      Este fichero implementa los comandos para examinar o modificar el
 *      estado de la maquina que esta siendo simulada.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#define MAXLINES        500

#include <bwcc.h>
#include <alloc.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include "tcl.h"
#include "asm.h"
#include "gp.h"
#include "dlx.h"
#include "sym.h"

/*
 * Declaraciones de procedimientos definidos mas tarde en este fichero:
 */

extern char far salidas[199];
static long int GetAddress();
static char *   GetString();
static void	ErrMsg();

/*
 *----------------------------------------------------------------------
 *
 * Gp_Vget --
 *
 *      Este procedimiento es invocado para procesar el comando "Examinar
 *      registro vectorial".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores, o TCL_ERROR si hay problemas.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Gp_Vget(machPtr, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{
    long int i, j;
    long int cuenta, result;
    unsigned long int numero;
    long int totalSpace, spaceNeeded;
    char huge *p;                            /* Puntero al caracter NULL, al final
					 * del resultado actual. */
    char *locString;
    char locBuf[100];
    char salida[5];
    char flotante[20];
    result = Sym_GetSym(machPtr, (char far *) NULL, argumentos[1],
		SYM_VREGS_OK, (unsigned long int *) &numero);
    if (result != SYM_REGISTER) {
    BWCCMessageBox(GetFocus(), "No vector register", "Error", MB_OK | MB_ICONEXCLAMATION);
    return TCL_ERROR;
    }
    else {
    cuenta = strtoul(argumentos[2], NULL, 10);
    if (cuenta > MAXLINES) cuenta = MAXLINES;
    if (cuenta <= 0) {cuenta = 1;}

    p = machPtr->result;
    totalSpace = TCL_RESULT_SIZE;
    for (i = numero;((i < NUM_V_REGS) && (cuenta > 0)) ;i++,cuenta--)
	{
	locString = locBuf;
	(void) sprintf(locBuf, "Register V%ld:", i);
	spaceNeeded = (p - machPtr->result) + _fstrlen(locString);
	while (totalSpace <= spaceNeeded) {
	    long int spaceUsed;
            char huge *newSpace;

	    totalSpace *= 2;
	    newSpace = farcalloc(1, (unsigned long int) totalSpace);
	    if (newSpace == NULL) {ErrMsg("Gp_VGet(1)");
				   return TCL_ERROR;}
	    spaceUsed = p-machPtr->result;
	    _fstrncpy((char far *)newSpace, (char far *)machPtr->result, (unsigned int)spaceUsed);
	    if (machPtr->dynamic) {
		farfree(machPtr->result);
	    }
	    machPtr->dynamic = 1;
	    machPtr->result = newSpace;
	    p = newSpace + (unsigned int)spaceUsed;
	}
	strcpy((char far *)p, locString);
	p += _fstrlen((char far *)p);

	for (j = 0; (j < MVL) ; j++)
	{
	    locString=locBuf;
	    if (((j+1) % 4) == 1)
		sprintf(salida,"\r\n");
		else sprintf(salida, "");
	    gcvt(*((double *)&(machPtr->vecregs[(int)i][(int)(j*2)])), 8, flotante);
	    (void) sprintf(locBuf,"%sV%ld[%2ld] = %s\t", salida,i,j, flotante);

	    spaceNeeded = (p - machPtr->result) + _fstrlen(locString);
	    while (totalSpace <= (spaceNeeded + 4)) {
	    long int spaceUsed;
            char huge *newSpace;

	    totalSpace *= 2;
	    newSpace = farcalloc(1, (unsigned long int) totalSpace);
	    if (newSpace == NULL) {ErrMsg("Gp_VGet(2)");
				   return TCL_ERROR;}
	    spaceUsed = p-machPtr->result;
	    _fstrncpy((char far *)newSpace, (char far *)machPtr->result, (unsigned int)spaceUsed);
	    if (machPtr->dynamic) {
		farfree(machPtr->result);
	    }
	    machPtr->dynamic = 1;
	    machPtr->result = newSpace;
	    p = newSpace + (unsigned int)spaceUsed;
	}
	strcpy((char far *)p, locString);
	p += _fstrlen((char far *)p);
	}
	*p=13;p++;*p=10;p++;
	*p=13;p++;*p=10;p++;

	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PCindex --
 *
 *      Este procedimiento es invocado para procesar el comando "Examinar
 *      PC Indexado".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores, o TCL_ERROR si hay problemas.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Gp_PCindex(machPtr, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{
    long int indice, cuenta, tmp, aux, value;
    char *end;
    char salida[5];

    long int totalSpace, spaceNeeded;
    char huge *p;                            /* Puntero al caracter NULL, al final
					 * del resultado actual. */
    char *locString;
    char locBuf[100];

    indice = strtol(argumentos[1], &end, 10);
    aux = tmp = machPtr->regs[(int)PC_REG];
    tmp += indice;
    if (tmp < 0) tmp = 0;
    cuenta = strtoul(argumentos[2], &end, 10);
    if (cuenta > MAXLINES) cuenta = MAXLINES;
    if (cuenta <= 0) cuenta = 1;

    if (tmp >= machPtr->numWords)
     sprintf((char far *)machPtr->result, "Unexisting memory reference\r\n");
    else {
    p = machPtr->result;
    totalSpace = TCL_RESULT_SIZE;
    while (1) {
	long int temp;

	temp = INDEX_TO_ADDR(tmp);
	value = machPtr->memPtr[tmp].value;
	if (tmp == machPtr->regs[(int)PC_REG]) {sprintf(salida,"\r\n");}
	else {sprintf(salida,"");}
	locString = locBuf;
	(void) sprintf(locBuf, "%s(pc)%+ld = %.100s => %.50s%s",
	salida,
	tmp-aux,
	Sym_GetString(machPtr, temp),
	Asm_Disassemble(machPtr, value,
	temp & ~0x3), salida);

	/*
	 * Nos aseguramos de que haya espacio suficiente para este valor en
	 * el buffer de resultados (aumenta el buffer de salida si es
	 * necesario), entonces pone el valor en el buffer.
	 */

	spaceNeeded = (p - machPtr->result) + _fstrlen(locString);
	while (totalSpace <= (spaceNeeded + 2)) {
	    char huge *newSpace;
	    long int spaceUsed;

	    totalSpace *= 2;
	    newSpace = farcalloc(1, (unsigned long int) totalSpace);
	    if (newSpace == NULL) {ErrMsg("Gp_PCIndex");
				   return TCL_ERROR;}
	    spaceUsed = p-machPtr->result;
	    _fstrncpy((char far *)newSpace,(char far *)machPtr->result, (unsigned int)spaceUsed);
	    if (machPtr->dynamic) {
		farfree(machPtr->result);
	    }
	    machPtr->dynamic = 1;
	    machPtr->result = newSpace;
	    p = newSpace + (unsigned int)spaceUsed;
	}
	strcpy((char far *)p, locString);
	p += _fstrlen((char far *)p);
	tmp +=1;
	if (tmp >= machPtr->numWords){
		sprintf((char far *)machPtr->result, "Unexisting memory reference\r\n");
		break;
		}

	/*
	 * Vamos al siguiente valor, si es que hay mas.
	 */

	cuenta -= 1;
	if (cuenta <= 0) {
	    break;
	}
	*p = 13;p++;*p = 10;p++;
     }
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_GetCmd --
 *
 *      Este procedimiento es invocado para procesar el comando "Examinar
 *      Otros".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores, o TCL_ERROR si hay problemas.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Gp_GetCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena con los argumentos. */
{
    unsigned long int address;
    long int isReg, result, totalSpace, spaceNeeded;
    enum {BINARY, HEX, DECIMAL, CHAR, STRING, INS} format = HEX;
    long int length = 4;                /* # de bytes en cada valor
					 * a imprimir. */
    long int count = 1;                 /* # de valores a imprimir. */
    long int addrIsValue = 0;           /* Imprimir la direccion en lugar del
					 * valor. */
    char huge *p;                            /* Apunta al caracter NULL al final
					 * del resultado actual. */
    char *locString, *valueString;
    char locBuf[100], valueBuf[100];

    if ((numarg != 2) && (numarg != 3)) {
	wsprintf(salidas, "Incorrect # argument: must be \"%.50s posicion [flags]\"", argumentos[0]);
	BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
    }
    result = GetAddress(machPtr, argumentos[1], &address, &isReg);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Analiza los flags.
     */

    if (numarg == 3) {
	for (p = argumentos[2]; *p != 0; p++) {
	    if (*p == 'b') {
		length = 1;
	    } else if (*p == 'h') {
		length = 2;
	    } else if (*p == 'w') {
		length = 4;
	    } else if (*p == 'x') {
		format = HEX;
	    } else if (*p == 'd') {
		format = DECIMAL;
	    } else if (*p == 'B') {
		format = BINARY;
	    } else if (*p == 'c') {
		format = CHAR;
		length = 1;
	    } else if (*p == 's') {
		if (isReg) {
		    wsprintf(salidas, "Cannot examine \"%.50s\" ", argumentos[1]);
		    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		    machPtr->result[0]='\0';
		    return TCL_ERROR;
		}
		format = STRING;
		length = 1;
	    } else if (*p == 'i') {
		format = INS;
	    } else if (isdigit(*p)) {
		char *end;

		count = strtoul((char far *)p, &end, 10);
		p = end-1;
	    } else if (*p == 'v') {
		addrIsValue = 1;
	    } else {
		wsprintf(salidas, "Incorrect operation \"%c\" in command \"%.50s\"", *p, argumentos[0]);
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_ERROR;
	    }
	}
    }

    /*
     * Entra en un bucle para formar pares <posicion valor>.  Redondea la
     * direccion hacia abajo para que tenga un valor apropiado segun el
     * tipo de resultado.
     */

    if (!isReg && !addrIsValue) {
	address &= ~(length-1);
    }

    if (count > MAXLINES) {count = MAXLINES;}
    if (count<= 0) {count = 1;}

    p = machPtr->result;
    totalSpace = TCL_RESULT_SIZE;
    while (1) {
	long int value;

	/*
	 * Coge el valor a imprimir.
	 */

	if (addrIsValue) {
	    value = address;
	} else if (isReg) {
	    if (address == PC_REG) {
		value = Sim_GetPC(machPtr);
	    } else if (address == NEXT_PC_REG) {
		value = INDEX_TO_ADDR(value);
	    } else {
		value = machPtr->regs[(unsigned int)address];
	    }
	} else {
	    if (ADDR_TO_INDEX(address) < machPtr->numWords) {
		value = machPtr->memPtr[ADDR_TO_INDEX(address)].value;
	    } else if (Io_Read(machPtr, (address & ~0x3), &(value)) == 0) {
		wsprintf(salidas, "Posicion \"%.50s\" not existing in memory", argumentos[1]);
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_ERROR;
	    }
	    if (length == 1) {
		switch (address & 0x3) {
		    case 0:
			value >>= 24;
			break;
		    case 1:
			value >>= 16;
			break;
		    case 2:
			value >>= 8;
			break;
		}
	    } else if ((length == 2) && !(address & 0x2)) {
		value >>= 16;
	    }
	}
	if (length == 1) {
	    value &= 0xff;
	    if ((format == DECIMAL) && (value & 0x80)) {
		value |= 0xffffff00L;
	    }
	}
	if (length == 2) {
	    value &= 0xffff;
	    if ((format == DECIMAL) && (value & 0x8000)) {
		value |= 0xffff0000L;
	    }
	}

	/*
	 * Formatea la direccion del valor, dejando un puntero al mismo en
	 * locString.
	 */

	locString = locBuf;
	if (addrIsValue) {
	    locString = "";
	} else if (isReg) {
	    (void) sprintf(locBuf, "%s:\t", Asm_RegNames[(unsigned int)address]);
	} else {
	    (void) sprintf(locBuf, "%s:\t", Sym_GetString(machPtr, address));
	}

	/*
	 * Formatea el valor mismo, dejando un puntero en valueString.
	 */

	valueString = valueBuf;
	switch (format) {
	    case HEX:
		(void) sprintf(valueBuf, "0x%08lx", value);
		break;
	    case DECIMAL:
		(void) sprintf(valueBuf, "%ld", value);
		break;
	    case BINARY: {
		long int i;
		for (i = 31; i >= 0; i--) {
		    if (value & (1<<i)) {
			*valueString = '1';
			valueString++;
		    } else if (valueString != valueBuf) {
			*valueString = '0';
			valueString++;
		    }
		}
		if (valueString == valueBuf) {
		    *valueString = '0';
		    valueString++;
		}
		*valueString = 0;
		valueString = valueBuf;
		break;
	    }
	    case CHAR:
		if (isprint(value)) {
		    (void) sprintf(valueBuf, "%c (0x%lx)", (char) value,
			    value);
		} else {
		    (void) sprintf(valueBuf, "0x%lx", value);
		}
		break;
	    case STRING:
		valueString = GetString(machPtr, address, &length);
		break;
	    case INS:
		valueString = Asm_Disassemble(machPtr, value, address);
		break;
	}

	/*
	 * Nos aseguramos de que haya espacio suficiente para este valor en
	 * el buffer de resultados (aumenta el buffer de salidas si es
	 * necesario), entonces pone el valor en el buffer.
	 */

	spaceNeeded = (p - machPtr->result) + _fstrlen(locString)
		+ _fstrlen(valueString);
	while (totalSpace <= (spaceNeeded + 2)) {
	    char huge *newSpace;
	    long int spaceUsed;

	    totalSpace *= 2;
	    newSpace = farcalloc(1, (unsigned long int) totalSpace);
	    if (newSpace == NULL) {ErrMsg("Gp_GetCmd");
				   return TCL_ERROR;}
	    spaceUsed = p-machPtr->result;
	    _fstrncpy((char far *)newSpace, (char far *)machPtr->result,(unsigned int)spaceUsed);
	    if (machPtr->dynamic) {
		farfree(machPtr->result);
	    }
	    machPtr->dynamic = 1;
	    machPtr->result = newSpace;
	    p = newSpace + (unsigned int)spaceUsed;
	}
	strcpy((char far *)p, locString);
	_fstrcat((char far *)p, valueString);
	p += _fstrlen((char far *)p);

	/*
	 * Vamos al siguiente valor, si es que hay mas.
	 */

	count -= 1;
	if (count <= 0) {
	    break;
	}
	if (isReg) {
	    address += 1;
	    if (address >= TOTAL_REGS) {
		break;
	    }
	} else {
	    address += length;
	    if (ADDR_TO_INDEX(address) >= machPtr->numWords) {
		break;
	    }
	}
	*p = 13;
	p++;
	*p = 10;
	p++;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_FGetCmd --
 *
 *      Este procedimiento es invocado para procesar el comando "Examinar
 *      Punto Flotante".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores, o TCL_ERROR si hay problemas.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Gp_FGetCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{
    unsigned long int address;
    long int isReg, result, totalSpace, spaceNeeded;
    enum {FLOAT, DOUBLE} format = FLOAT;
    long int length = 4;                /* # de bytes en cada valor a
					 * imprimir. */
    long int count = 1;                 /* # de valores a imprimir. */
    char huge *p;                            /* Puntero al caracter NULL, al final
					 * del resultado actual. */
    char *locString, *valueString;
    char locBuf[100], valueBuf[100];

    if ((numarg != 2) && (numarg != 3)) {
	wsprintf(salidas, "Incorrect # argument: must be \"%.50s posicion [flags]\"", argumentos[0]);
	BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
    }
    result = GetAddress(machPtr, argumentos[1], &address, &isReg);
    if (result != TCL_OK) {
	return result;
    }

    /*
     * Analiza los flags.
     */

    if (numarg == 3) {
	for (p = argumentos[2]; *p != 0; p++) {
	    if (*p == 'f') {
		format = FLOAT;
		length = 4;
	    } else if (*p == 'd') {
		format = DOUBLE;
		length = 8;
	    } else if (isdigit(*p)) {
		char *end;

		count = strtoul((char far *)p, &end, 10);
		p = end-1;
	    } else {
		wsprintf(salidas, "Incorrect option \"%c\" in command \"%.50s\"", *p, argumentos[0]);
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_ERROR;
	    }
	}
    }

    /*
     * Entra en un bucle para formar pares <posicion valor>.  Redondea la
     * direccion hacia abajo para que tenga un valor apropiado segun el
     * tipo de resultado.
     */

    if (!isReg) {
	address &= ~3;
    }

    if (count > MAXLINES) {count = MAXLINES;}
    if (count<= 0) {count = 1;}

    p = machPtr->result;
    totalSpace = TCL_RESULT_SIZE;
    while (1) {
	long int value[2];
	double *pd = (double *)&(value[0]);
	float *pf = (float *)&(value[0]);

	/*
	 * Toma el valor a imprimir.
	 */

	if (isReg) {
	    switch (address) {
	    case PC_REG :
	    case NEXT_PC_REG :
		BWCCMessageBox(GetFocus(), "Cannot print floating point format", "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		return TCL_ERROR;
	    default :
		value[0] = machPtr->regs[(unsigned int)address];
		if (((format == FLOAT) && (address > 63)) ||
		    ((format == DOUBLE) && ((address < 32) || (address > 63)))) {
			BWCCMessageBox(GetFocus(), "Cannot print floating point format", "Error", MB_OK | MB_ICONEXCLAMATION);
			machPtr->result[0] = '\0';
			return TCL_ERROR;
		} else if ((format == DOUBLE) && (address % 2)) {
			BWCCMessageBox(GetFocus(), "Double precision register cannot be Odd", "Error", MB_OK | MB_ICONEXCLAMATION);
			machPtr->result[0] = '\0';
		       /*       "Misaligned double floating point register" */
			return TCL_ERROR;
		}
		if (format == DOUBLE)
		    value[1] = machPtr->regs[(unsigned int)(address + 1)];
		break;
	    }
	} else {
	    if (ADDR_TO_INDEX(address) < machPtr->numWords) {
		value[0] = machPtr->memPtr[ADDR_TO_INDEX(address)].value;
		if (format == DOUBLE) {
		    value[1] = machPtr->memPtr[ADDR_TO_INDEX((address+4))].value;
		}
	    } else if (Io_Read(machPtr, (address & ~0x3), &(value[0])) == 0) {
	    wsprintf(salidas, "Position \"%.50s\" not existing in memory", argumentos[1]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	    }
	}

	/*
	 * Formatea la direccion del valor, dejando un puntero al mismo en
	 * locString.
	 */

	locString = locBuf;
	if (isReg) {
	    (void) sprintf(locBuf, "%s:\t", Asm_RegNames[(unsigned int)address]);
	} else {
	    (void) sprintf(locBuf, "%s:\t", Sym_GetString(machPtr, address));
	}

	/*
	 * Formatea el valor mismo, dejando un puntero en valueString.
	 */

	valueString = valueBuf;
	switch (format) {
	    case FLOAT:
		(void) sprintf(valueBuf, "%f", *pf);
		break;
	    case DOUBLE:
		(void) sprintf(valueBuf, "%lf", *pd);
		break;
	}

	/*
	 * Nos aseguramos de que haya espacio suficiente para este valor en
	 * el buffer de resultados (aumenta el buffer de salida si es
	 * necesario), entonces pone el valor en el buffer.
	 */

	spaceNeeded = (p - machPtr->result) + _fstrlen(locString)
		+ _fstrlen(valueString);
	while (totalSpace <= (spaceNeeded + 2)) {
	    char huge *newSpace;
	    long int spaceUsed;

	    totalSpace *= 2;
	    newSpace = farcalloc(1, (unsigned long int) totalSpace);
	    if (newSpace == NULL) {ErrMsg("Gp_FGet");
				   return TCL_ERROR;}
	    spaceUsed = p-machPtr->result;
	    _fstrncpy((char far *)newSpace, (char far *)machPtr->result, (unsigned int)spaceUsed);
	    if (machPtr->dynamic) {
		farfree(machPtr->result);
	    }
	    machPtr->dynamic = 1;
	    machPtr->result = newSpace;
	    p = newSpace + (unsigned int)spaceUsed;
	}
	strcpy((char far *)p, locString);
	_fstrcat((char far *)p, valueString);
	p += _fstrlen((char far *)p);

	/*
	 * Vamos al siguiente valor, si es que hay mas.
	 */

	count -= 1;
	if (count <= 0) {
	    break;
	}
	if (isReg) {
	    address += length / 4;
	    if (address >= TOTAL_REGS) {
		break;
	    }
	} else {
	    address += length;
	    if (ADDR_TO_INDEX(address) >= machPtr->numWords) {
		break;
	    }
	}
	*p = 13;
	p++;
	*p = 10;
	p++;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PutCmd --
 *
 *      Este procedimiento es invocado para procesar el comando "Modificar
 *      Otros".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores, o TCL_ERROR si hay problemas.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Gp_PutCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{
    unsigned long int address;
    long int isReg, value, result;
    char *end;

    if (numarg != 3) {
    wsprintf(salidas, "Incorrect # de argument: must be \"%.50s posicion valor\"",argumentos[0]);
    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    return TCL_ERROR;
    }
    result = GetAddress(machPtr, argumentos[1], &address, &isReg);
    if (result != TCL_OK) {
	return result;
    }

    result = Sym_EvalExpr(machPtr, (char far *) NULL, argumentos[2], cero,
	    &value, &end);
    if (result != TCL_OK) {
	return result;
    }
    if (*end != 0) {
    wsprintf(salidas,"Value incorrect \"%.50s\" in command \"%.50s\"",
		argumentos[2], argumentos[0]);
    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    return TCL_ERROR;
    }
    if (isReg) {
	if (address == 0) {
	    wsprintf(salidas, "Cannot modify %s", Asm_RegNames[(unsigned int)address]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	if ((address == PC_REG) || (address == NEXT_PC_REG)) {
	    if (value & 0x3) {
	    sprintf(salidas,"Position 0x%lx not suitable aligned for pc or npc",
			value);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	    }
	    value = ADDR_TO_INDEX(value);
	}
	machPtr->regs[(unsigned int)address] = value;
    } else {
	if (ADDR_TO_INDEX(address) < machPtr->numWords) {
	    MemWord huge *wordPtr;

	    wordPtr = &machPtr->memPtr[ADDR_TO_INDEX(address)];
	    wordPtr->value = value;
	    wordPtr->opCode = OP_NOT_COMPILED;
	} else if (Io_Write(machPtr, (address & ~0x3), value, cuatro) == 0) {
	    wsprintf(salidas,
		    "Position \"%.50s\" not existing in memory",
		    argumentos[1]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_FPutCmd --
 *
 *      Este procedimiento es invocado para procesar el comando "Modificar
 *      Punto Flotante".
 *      Ver la documentacion de usuario para mas detalles.
 *
 * Resultados:
 *      TCL_OK si no hay errores, o TCL_ERROR si hay problemas.
 *
 * Efectos Laterales:
 *      Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

long int
Gp_FPutCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;                       /* Descripcion de la maquina. */
    int numarg;                         /* Numero de argumentos. */
    char argumentos[argus][largus];     /* Cadena de argumentos. */
{
    unsigned long int address;
    long int isReg, result, value[2];
    float *pf = (float *)&(value[0]);
    double *pd = (double *)&(value[0]), atof();
    enum {FLOAT, DOUBLE} format = FLOAT;

    if ((numarg != 3) && (numarg != 4)) {
	wsprintf(salidas,
		"Incorrect # argument:  must be \"%.50s value position [f|d]\"",
		argumentos[0]);
	BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
    }
    result = GetAddress(machPtr, argumentos[1], &address, &isReg);
    if (result != TCL_OK) {
	return result;
    }
    if (numarg == 4) {
	if (*argumentos[3] == 'f')
	    format = FLOAT;
	else if (*argumentos[3] == 'd')
	    format = DOUBLE;
	else {
	BWCCMessageBox(GetFocus(), "Incorrect Flag, must be f or d", "Error",
	MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
	}
    }

    if (!isReg) {
	address = address & ~3;
    } else if (format == FLOAT) {
	if (/* *** (address < 0) || *** */ (address > 63)) {
	    BWCCMessageBox(GetFocus(), "Illegal Register", "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
    } else {    /* format == DOUBLE */
	if ((address < 32) || (address > 63) || (address %2)) {
	    BWCCMessageBox(GetFocus(),"Illegal double precision format", "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
    }

    if (format == FLOAT) {
	*pd = atof(argumentos[2]);
	if ((*pd > MINFLOAT) && (*pd <MAXFLOAT)) { *pf = *pd;}
	else {*pf = 0;}
    } else {    /* format == DOUBLE */
	*pd = atof(argumentos[2]);
    }

    if (isReg) {
	if (address == 0) {
	    wsprintf(salidas, "Cannot modify %s", Asm_RegNames[(unsigned int)address]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
	machPtr->regs[(unsigned int)address] = value[0];
	if (format == DOUBLE) {
	    machPtr->regs[(unsigned int)(address+1)] = value[1];
	}
    } else {
	if (ADDR_TO_INDEX(address) < machPtr->numWords) {
	    MemWord huge *wordPtr;

	    wordPtr = &machPtr->memPtr[ADDR_TO_INDEX(address)];
	    wordPtr->value = value[0];
	    wordPtr->opCode = OP_NOT_COMPILED;
	    if (format == DOUBLE) {
		wordPtr = &machPtr->memPtr[ADDR_TO_INDEX((address+4))];
		wordPtr->value = value[1];
		wordPtr->opCode = OP_NOT_COMPILED;
	    }
	} else if (Io_Write(machPtr, (address & ~0x3), value[0], cuatro) == 0) {
	    wsprintf(salidas,
		    "Position \"%.50s\" not existing in memory",
		    argumentos[1]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetAddress --
 *
 *      Convierte un string de una direccion al valor de esa direccion,
 *      para que pueda ser utilizado para leer o escribir esa posicion.
 *
 * Resultados:
 *      Devuelve TCL_OK en el caso de que no haya errores, TCL_ERROR
 *      si ha habido problemas.
 *
 * Efectos Laterales:
 *      Genera un mensaje de error si el string no puede ser convertido
 *      en direccion.
 *
 *----------------------------------------------------------------------
 */

static long int
GetAddress(machPtr, string, addressPtr, regPtr)
    DLXV *machPtr;                       /* Maquina a la que pertenece la
					 * direccion.
					 * Tambien usada para comunicar
					 * errores. */
    char *string;                       /* Especifica la direccion. */
    unsigned long int *addressPtr;      /* Almacena aqui la direccion del
					 * valor. */
    long int *regPtr;                   /* Almacena 1 aqui si la direccion
					 * se refiere a un registro, 0 para
					 * podsicion de memoria. */
{
    unsigned long int result;
    char *end;

    result = Sym_GetSym(machPtr, (char far *) NULL, string, SYM_PSEUDO_OK,
	    addressPtr);
    if (result == SYM_FREG_FOUND) {
	*regPtr = 1;
	*addressPtr += 32;
	return TCL_OK;
    }
    if (result == SYM_REGISTER) {
	*regPtr = 1;
	return TCL_OK;
    }
    if (result == SYM_FOUND) {
	*regPtr = 0;
	return TCL_OK;
    }

    result = Sym_EvalExpr(machPtr, (char far *) NULL, string, cero,
	    (long int *) addressPtr, &end);
    if (result != TCL_OK) {
	return result;
    }
    if (*end != 0) {
	wsprintf(salidas,
		"Expresion \"%.50s\" unknown", string);
	BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
	return TCL_ERROR;
    }
    *regPtr = 0;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetString --
 *
 *      Dada una direccion, devuelve la cadena ASCII que esta almacenada
 *      en ella.
 *
 * Resultados:
 *      El valor devuelto es un puntero a una cadena ASCII que esta en
 *      "address" en la memoria de machPtr.  El entero de *countPtr es
 *      sobreescrito con el numero de bytes que ocupa el string, incluyendo
 *      el caracter NULL de fin de string. Si el string es muy largo, el valor
 *      devuleto puede ser truncado para contener unicamente los primeros
 *      caracteres del string (*countPtr tambien es truncado). El valor de
 *      retorno es almacenado en un buffer estatico que es sobreescrito en la
 *      siguiente llamada a este procedimiento.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

static char *
GetString(machPtr, address, countPtr)
    register DLXV *machPtr;      /* Maquina cuya memoria va a ser examinada. */
    unsigned long int address;  /* Direccion del string. */
    long int *countPtr;         /* Aqui almacena la longitud del string
				 * (incluyendo el caracter NULL, si existe). */
{
#define MAX_LENGTH 200L
    static char buffer[(4*MAX_LENGTH)+20];
    char *p;
    unsigned long int index;
    long int value, count;

    buffer[0] = '"';
    p = buffer+1;
    for (count = 0; ; count++, address++) {
	index = ADDR_TO_INDEX(address);
	if (index >= machPtr->numWords) {
	    break;
	}
	value = machPtr->memPtr[index].value;
	switch (address & 0x3) {
	    case 0:
		value >>= 24;
		break;
	    case 1:
		value >>= 16;
		break;
	    case 2:
		value >>= 8;
		break;
	}
	value &= 0xff;
	if (value == 0) {
	    *countPtr = count+1;
	    break;
	}
	if (count == MAX_LENGTH) {
	    *countPtr = MAX_LENGTH;
	    strcpy(p, "...");
	    p += 3;
	    break;
	}
	if (value == '\\') {
	    strcpy(p, "\\\\");
	    p += 2;
	} else if (value == '"') {
	    strcpy(p, "\\\"");
	    p += 2;
	} else if (isascii(value) && isprint(value)) {
	    *p = value;
	    p++;
	} else if (value == '\n') {
	    strcpy(p, "\\n");
	    p += 2;
	} else if (value == '\t') {
	    strcpy(p, "\\t");
	    p += 2;
	} else if (value == '\b') {
	    strcpy(p, "\\b");
	    p += 2;
	} else if (value == '\r') {
	    strcpy(p, "\\r");
	    p += 2;
	} else {
	    sprintf(p, "\\x%02lx", value);
	    p += 4;
	}
    }
    *p = '"';
    p[1] = 0;
    return buffer;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PutString --
 *
 *      Dado una cadena ASCII, la almacena en la memoria de la maquina.
 *
 * Resultados:
 *      El valor de retorno es el numero de bytes realmente almacenados
 *      en la memoria de machPtr. Si endPtr no es NULL, en *endPtr se
 *      almacena la direccion del caracter que finalizaba la cadena
 *      (o term o 0).
 *
 * Efectos Laterales:
 *      Los bytes del "string" son almacenados en la memoria de machPtr
 *      comenzando en "address".
 *
 *----------------------------------------------------------------------
 */

long int
Gp_PutString(machPtr, string, term, address, addNull, endPtr)
    register DLXV *machPtr;              /* Maquina cuya memoria va a ser
					 * modificada. */
    register char *string;              /* Cadena a almacenar. */
    char term;                          /* Caracter con el que finaliza la
					 * cadena. */
    unsigned long int address;          /* En que lugar de la memoria de
					 * machPtr hay que almacenar. */
    long int addNull;                   /* Si no-cero, aadir un caracter NULL
					 * de fin de cadena en la memoria
					 * despues del string. */
    char **endPtr;                      /* Si no-NULL, contiene la direccion
					 * del ultimo caracter de la cadena. */
{
    long int backslashCount;
    long int size;

    size = 0;
    while (1) {
	if (*string == '\\') {
	    Gp_PutByte(machPtr, address,
		    Tcl_Backslash(string, &backslashCount));
	    string += (unsigned int)backslashCount;
	} else if ((*string == 0) || (*string == term)) {
	    break;
	} else {
	    Gp_PutByte(machPtr, address, *string);
	    string++;
	}
	address += 1;
	size += 1;
    }
    if (addNull) {
	Gp_PutByte(machPtr, address, cero);
	size += 1;
    }
    if (endPtr != 0) {
	*endPtr = string;
    }
    return size;
}

/*
 *----------------------------------------------------------------------
 *
 * Gp_PutByte --
 *
 *      Almacena un byte en una determinada posicion de memoria.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      La memoria de machPtr es modificada.
 *
 *----------------------------------------------------------------------
 */

void
Gp_PutByte(machPtr, address, value)
    register DLXV *machPtr;              /* Maquina cuya memoria va a ser
					 * modificada. */
    unsigned long int address;          /* Donde almacenar el valor. */
    long int value;                     /* Valor a almacenar como byte. */
{
    MemWord huge *wordPtr;
    long int index;

    index = ADDR_TO_INDEX(address);
    if (index < machPtr->numWords) {
	wordPtr = &machPtr->memPtr[index];
	switch (address & 0x3) {
	    case 0:
		wordPtr->value = (wordPtr->value & 0xffffffL)
			| (value << 24);
		break;
	    case 1:
		wordPtr->value = (wordPtr->value & 0xff00ffffL)
			| ((value << 16) & 0xff0000L);
		break;
	    case 2:
		wordPtr->value = (wordPtr->value & 0xffff00ffL)
			| ((value << 8) & 0xff00);
		break;
	    case 3:
		wordPtr->value = (wordPtr->value & 0xffffff00L)
			| (value & 0xff);
		break;
	}
	wordPtr->opCode = OP_NOT_COMPILED;
    }
}

static void
ErrMsg(mensaje)
char *mensaje;
{
wsprintf(salidas, "Insufficient memory while calling: %s", mensaje);
BWCCMessageBox(GetFocus(), salidas, "Memory error", MB_OK | MB_ICONEXCLAMATION);
}