/*
 * sim.c --
 *
 *      Este fichero contiene el cuerpo principal del simulador
 *      para el DLXV.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:                                   *
 *      Simulador Vectorial para PC bajo entorno Windows        *
 *                                                              *
 *      Autor: Roberto Calpe Lorenzo                            *
 *                                                              *
 ****************************************************************/

#include <alloc.h>
#include <stdio.h>
#include <bwcc.h>
#include <values.h>
#include "tcl.h"
#include "asm.h"
#include "dlx.h"
#include "sym.h"
#include "sim.h"

/*
 * La siguiente tabla es utilizada para transformar los bits 31:26 de la
 * instruccion en un valor adecuado para el campo "opCode" de la estructura
 * MemWord, o en un valor especial para una posterior decodificacion.
 */

extern BOOL encadenado;
extern BOOL iniciar;
extern HWND easy;

#define SPECIAL 120L
#define FPARITH 121L

#define IFMT 1L
#define JFMT 2L
#define RFMT 3L

typedef struct {
    long int opCode;            /* Codigo de operacion traducido. */
    long int format;            /* Formato de la instruccion
				 * (IFMT o JFMT o RFMT) */
} OpInfo;

OpInfo far opTable[] = {
    {SPECIAL, RFMT}, {FPARITH, RFMT}, {OP_J, JFMT}, {OP_JAL, JFMT},
    {OP_BEQZ, IFMT}, {OP_BNEZ, IFMT}, {OP_BFPF, IFMT}, {OP_BFPT, IFMT},
    {OP_ADDI, IFMT}, {OP_ADDUI, IFMT}, {OP_SUBI, IFMT}, {OP_SUBUI, IFMT},
    {OP_ANDI, IFMT}, {OP_ORI, IFMT}, {OP_XORI, IFMT}, {OP_LHI, IFMT},
    {OP_RFE, IFMT}, {OP_TRAP, IFMT}, {OP_JR, IFMT}, {OP_JALR, IFMT},
    {OP_SLLI, IFMT}, {OP_RES, IFMT}, {OP_SRLI, IFMT}, {OP_SRAI, IFMT},
    {OP_SEQI, IFMT}, {OP_SNEI, IFMT}, {OP_SLTI, IFMT}, {OP_SGTI, IFMT},
    {OP_SLEI, IFMT}, {OP_SGEI, IFMT}, {OP_RES, IFMT}, {OP_RES, IFMT},
    {OP_LB, IFMT}, {OP_LH, IFMT}, {OP_LV, IFMT}, {OP_LW, IFMT},
    {OP_LBU, IFMT}, {OP_LHU, IFMT}, {OP_LF, IFMT}, {OP_LD, IFMT},
    {OP_SB, IFMT}, {OP_SH, IFMT}, {OP_SV, IFMT}, {OP_SW, IFMT},
    {OP_RES, IFMT}, {OP_RES, IFMT}, {OP_SF, IFMT}, {OP_SD, IFMT},
    {OP_SEQUI, IFMT}, {OP_SNEUI, IFMT}, {OP_SLTUI, IFMT}, {OP_SGTUI, IFMT},
    {OP_SLEUI, IFMT}, {OP_SGEUI, IFMT}, {OP_RES, IFMT}, {OP_RES, IFMT},
    {OP_RES, IFMT}, {OP_RES, IFMT}, {OP_RES, IFMT}, {OP_RES, IFMT},
    {OP_RES, IFMT}, {OP_RES, IFMT}, {OP_RES, IFMT}, {OP_RES, IFMT}
};

/*
 * La siguiente tabla es utilizada para convertir el campo "funct" de las
 * instrucciones SPECIAL en el campo "opCode" de la estructura MemWord.
 */

long int far specialTable[] = {
    OP_NOP, OP_RES, OP_RES, OP_RES, OP_SLL, OP_RES, OP_SRL, OP_SRA,
    OP_RES, OP_RES, OP_RES, OP_RES, OP_TRAP, OP_RES, OP_RES, OP_RES,
    OP_SEQU, OP_SNEU, OP_SLTU, OP_SGTU, OP_SLEU, OP_SGEU, OP_RES, OP_RES,
    OP_RES, OP_RES, OP_RES, OP_RES, OP_RES, OP_RES, OP_RES, OP_RES,
    OP_ADD, OP_ADDU, OP_SUB, OP_SUBU, OP_AND, OP_OR, OP_XOR, OP_RES,
    OP_SEQ, OP_SNE, OP_SLT, OP_SGT, OP_SLE, OP_SGE, OP_RES, OP_RES,
    OP_MOVI2S, OP_MOVS2I, OP_MOVF, OP_MOVD, OP_MOVFP2I, OP_MOVI2FP,
    OP_MOVF2S, OP_MOVS2F,
    OP_POP, OP_CVM, OP_CVI, OP_LVWS, OP_SVWS, OP_LVI, OP_SVI, OP_RES
};

/*
 * La siguiente tabla es usada para convertir el campo "funct" de las
 * instrucciones FPARITH en el campo "opCode" de la estructura MemWord.
 */
long int far FParithTable[] = {
  OP_ADDF, OP_SUBF, OP_MULTF, OP_DIVF, OP_ADDD, OP_SUBD, OP_MULTD, OP_DIVD,
  OP_CVTF2D, OP_CVTF2I, OP_CVTD2F, OP_CVTD2I, OP_CVTI2F, OP_CVTI2D,
  OP_MULT, OP_DIV,
  OP_EQF, OP_NEF, OP_LTF, OP_GTF, OP_LEF, OP_GEF, OP_MULTU, OP_DIVU,
  OP_EQD, OP_NED, OP_LTD, OP_GTD, OP_LED, OP_GED, OP_ADDV, OP_ADDSV,
  OP_SUBV, OP_SUBVS, OP_SUBSV, OP_MULTV, OP_MULTSV, OP_DIVV, OP_DIVVS, OP_DIVSV,
  OP_SEQV, OP_SNEV, OP_SGTV, OP_SLTV, OP_SGEV, OP_SLEV, OP_SEQSV, OP_SNESV,
  OP_SGTSV, OP_SLTSV, OP_SGESV, OP_SLESV, OP_RES, OP_RES, OP_RES, OP_RES,
  OP_RES, OP_RES, OP_RES, OP_RES, OP_RES, OP_RES, OP_RES, OP_RES
  };


static long int indices[MVL];
long int strid = 0;
extern char far salidas[199];

/*
 * Declaraciones de procedimientos definidos mas tarde en el fichero:
 */

static long int AddressError();
static long int BusError();
static long int vlrError();
static void     Compile();
static long int      Mult();
static long int Overflow();
static long int ReadMem();
static long int ReadVMem();
static long int WriteVMem();
static long int WriteMem();
static long int      FPissue ();
static void     FPwriteBack();
static void     Vwrite();
static void     FPwait();
static FPop     *mallocFPop();
static void     freeFPop();
static void     Terminar();

extern long int Handle_Trap();


/*
 *----------------------------------------------------------------------
 *
 * ReadMem --
 *
 *      Lee una palabra de la memoria del DLXV.
 *
 * Resultados:
 *      Bajo circunstancias normales, el resultado es 1 y la palabra en
 *      *valuePtr es modificada para que contenga la palabra del DLXV de esa
 *      direccion. Si no existe esa direccion de memoria, o si existe un stop
 *      en esa posicion, entonces 0 es devuelto para comunicar que la
 *      simulacion debe parar.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

static long int
ReadMem(machPtr, address, valuePtr, numarg, argumentos)
    register DLXV *machPtr;      /* Maquina cuya memoria va a ser leida. */
    unsigned long int address;  /* Direccion. */
    long int *valuePtr;         /* Almacena el contenido de esa palabra
				 * en esta variable. */
    int numarg;
    char argumentos[argus][largus];
{
    unsigned long int index;
    long int result;
    register MemWord huge *wordPtr;

    index = ADDR_TO_INDEX(address);
    if (index < machPtr->numWords) {
	wordPtr = &machPtr->memPtr[index];
	if ((wordPtr->stopList != NULL)
		&& (machPtr->insCount != machPtr->firstIns)) {
	    result = Stop_Execute(machPtr, wordPtr->stopList, numarg, argumentos);
	    if ((result != TCL_OK) || (machPtr->flags & STOP_REQUESTED)) {
		return 0;
	    }
	}
	if (machPtr->refTraceFile)
	  fprintf( machPtr->refTraceFile, "0 %lx\n", address );
	*valuePtr = wordPtr->value;
	return 1;
    }

    /*
     * La palabra no esta en la memoria principal.  Miramos si es un
     * registro de E/S.
     */

    if (Io_Read(machPtr, (address & ~0x3), valuePtr) == 1) {
	return 1;
    }

    /*
     * La palabra no existe.  Registramos un error de bus.
     */

    (void) BusError(machPtr, address, cero);
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * ReadVMem --
 *      Lee una parte de los datos de un vector.
 *
 * Resultados:
 *      Bajo circunstancias normales, el resultado es 1 y la palabra en
 *      *valuePtr es modificada para que contenga la palabra del DLXV de esa
 *      direccion. Si no existe esa direccion de memoria, entonces 0 es
 *      devuelto para comunicar que la simulacion debe parar.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

static long int
ReadVMem(machPtr, address, valuePtr)
    register DLXV *machPtr;      /* Maquina cuya memoria va a ser leida. */
    unsigned long int address;  /* Direccion deseada. */
    long int *valuePtr;         /* Almacena el contenido de esa palabra en
				 * esta variable. */
{
    unsigned long int index;
    register MemWord huge *wordPtr;

    index = ADDR_TO_INDEX(address);
    if (index < machPtr->numWords) {
	wordPtr = &machPtr->memPtr[index];

	*valuePtr = wordPtr->value;
	return 1;
    }

    /*
     * La palabra no esta en la memoria principal. Miramos si es un registro
     * de E/S.
     */

    if (Io_Read(machPtr, (address & ~0x3), valuePtr) == 1) {
	return 1;
    }

    /*
     * La palabra no existe.  Registramos un error de bus.
     */

    (void) BusError(machPtr, address, cero);
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * WriteVMem --
 *
 *      Escribe un valor en la memoria del DLXV que es parte de un vector.
 *
 * Resultados:
 *      Si la escritura es completada satisfactoriamente entonces 1 es
 *      devuelto. Si cualquier clase de problema ocurre se devuelve 0;
 *      el que hace la llamada deberia parar la simulacion.
 *
 * Efectos Laterales:
 *      La memoria del DLXV es actualizada con el nuevo valor de la
 *      palabra.
 *
 *----------------------------------------------------------------------
 */
static long int
WriteVMem(machPtr, address, size, value)
    register DLXV *machPtr;      /* Maquina cuya memoria va a ser escrita. */
    unsigned long int address;  /* Direccion. */
    long int size;              /* Tamao del dato a ser escrito (4 bytes). */
    long int value;             /* Nuevo valor a escribir en la memoria. */
{
    unsigned long int index;
    register MemWord huge *wordPtr;

    if ((size == 4) && (address & 0x3)) {
	(void) AddressError(machPtr, address, cero);
	return 0;
    }
    index = ADDR_TO_INDEX(address);
    if (index < machPtr->numWords) {
	wordPtr = &machPtr->memPtr[index];

	if (machPtr->refTraceFile)
	  fprintf(machPtr->refTraceFile, "1 %lx\n", address);

	wordPtr->value = value;
	wordPtr->opCode = OP_NOT_COMPILED;
	return 1L;
    }

    /*
     * No esta en la memoria principal.  Miramos si es un registro de E/S.
     */

    if (Io_Write(machPtr, address, value, size) == 1) {
	return 1;
    }

    (void) BusError(machPtr, address, cero);
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * WriteMem --
 *
 *      Escribe un valor en la memoria del DLXV.
 *
 * Resultados:
 *      Si la escritura es completada satisfactoriamente entonces 1 es
 *      devuelto. Si cualquier clase de problema ocurre ( tal como un
 *      error de direccion o un stop) entonces se devuelve 0; el que
 *      hace la llamada deberia parar la simulacion.
 *
 * Efectos Laterales:
 *      La memoria del DLXV es actualizada con el nuevo valor del byte,
 *      halfword, o word.
 *
 *----------------------------------------------------------------------
 */
static long int
WriteMem(machPtr, address, size, value, numarg, argumentos)
    register DLXV *machPtr;      /* Maquina cuya memoria va a ser escrita. */
    unsigned long int address;  /* Direccion. */
    long int size;              /* Tamao del dato a ser escrito (1, 2, o 4 bytes). */
    long int value;             /* Nuevo valor a escribir en la memoria. */
    int numarg;
    char argumentos[argus][largus];
{
    unsigned long int index;
    long int result;
    register MemWord huge *wordPtr;

    if (((size == 4) && (address & 0x3)) || ((size == 2) && (address & 0x1))) {
	(void) AddressError(machPtr, address, cero);
	return 0;
    }
    index = ADDR_TO_INDEX(address);
    if (index < machPtr->numWords) {
	wordPtr = &machPtr->memPtr[index];
	if ((wordPtr->stopList != NULL)
		&& (machPtr->insCount != machPtr->firstIns)) {
	    result = Stop_Execute(machPtr, wordPtr->stopList, numarg, argumentos);
	    if ((result != TCL_OK) || (machPtr->flags & STOP_REQUESTED)) {
		return 0;
	    }
	}
	if (machPtr->refTraceFile)
	  fprintf(machPtr->refTraceFile, "1 %lx\n", address);
	if (size == 4) {
	    wordPtr->value = value;
	} else if (size == 2) {
	    if (address & 0x2) {
		wordPtr->value = (wordPtr->value & 0xffff0000L)
			| (value & 0xffff);
	    } else {
		wordPtr->value = (wordPtr->value & 0xffff)
			| (value << 16);
	    }
	} else {
	    switch (address & 0x3) {
		case 0:
		    wordPtr->value = (wordPtr->value & 0x00ffffffL)
			    | (value << 24);
		    break;
		case 1:
		    wordPtr->value = (wordPtr->value & 0xff00ffffL)
			    | ((value & 0xff) << 16);
			    break;
		case 2:
		    wordPtr->value = (wordPtr->value & 0xffff00ffL)
			    | ((value & 0xff) << 8);
		    break;
		case 3:
		    wordPtr->value = (wordPtr->value & 0xffffff00L)
			    | (value & 0xff);
		    break;
	    }
	}
	wordPtr->opCode = OP_NOT_COMPILED;
	return 1L;
    }

    /*
     * No esta en la memoria principal.  Miramos si es un registro del E/S.
     */

    if (Io_Write(machPtr, address, value, size) == 1) {
	return 1;
    }

    (void) BusError(machPtr, address, cero);
    return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Simulate --
 *
 *      Este procedimiento forma el corazon del simulador.  Ejecuta
 *      instrucciones hasta que se produce un break o si ocurre un
 *      error (o hasta que se ha ejecutado una sola instruccion, si
 *      se ha pedido un Paso a Paso).
 *
 * Resultados:
 *      TCL_OK si no hay problemas, TCL_ERROR si se produce algun error.
 *
 * Efectos Laterales:
 *      El estado de *machPtr cambia en respuesta a la simulacion.
 *      Informacion de retorno puede ser dejada en machPtr->result.
 *
 *----------------------------------------------------------------------
 */

#define Checker(x) ((lastLoad1 && (lastLoad1 == (x))) || (lastLoad2 && (lastLoad2 == (x))))
#define CheckerF(x) Checker((x) + 32)
#define CheckerD(x) (Checker((x) + 32) || Checker((x) + 33))
#define Check   if (stalled) {machPtr->loadStalls++; machPtr->cycleCount++;}
#define CheckS1 if (!stalled && Checker(wordPtr->rs1)) stalled = 1;
#define CheckS2 if (!stalled && Checker(wordPtr->rs2)) stalled = 1;
#define CheckD  if (!stalled && Checker(wordPtr->rd)) stalled = 1;
#define CheckS1F        if (!stalled && CheckerF(wordPtr->rs1)) stalled = 1;
#define CheckS2F        if (!stalled && CheckerF(wordPtr->rs2)) stalled = 1;
#define CheckDF if (!stalled && CheckerF(wordPtr->rd)) stalled = 1;
#define CheckS1D        if (!stalled && CheckerD(wordPtr->rs1)) stalled = 1;
#define CheckS2D        if (!stalled && CheckerD(wordPtr->rs2)) stalled = 1;
#define CheckDD if (!stalled && CheckerD(wordPtr->rd)) stalled = 1;
/* Estos sirven para detectar riesgos RAW en la que la que R es una instruccion de salto */
#define CheckerRdst(x) (machPtr->lastRdst && (machPtr->lastRdst == (x)))
#define CheckRdst   if (stalled) {machPtr->loadStalls++; machPtr->cycleCount++;}
#define CheckRdstS1 if (CheckerRdst(wordPtr->rs1)) stalled = 1;

/*
 *----------------------------------------------------------------------
 *
 * Sim2 --
 *
 *      Este procedimiento es llamado desde el interior de Simulate para
 *      simular las operaciones vectoriales.
 *
 *----------------------------------------------------------------------
 */

static long int
Sim2(machPtr, wordPtr, errorMessage, flagstop, stalled, lastLoad1, lastLoad2)
    DLXV *machPtr;
    MemWord huge *wordPtr;
    long int *flagstop;
    char **errorMessage;
    long int stalled, lastLoad1, lastLoad2;
{
 long int flagis;
 long int errorValue;
 long int result;
 unsigned long int tmp;
 unsigned char opco;
 *flagstop = 0;

	opco = wordPtr->opCode;
	if ((machPtr->regs[(int)VLR_REG] > MVL) && (opco != OP_POP) && (opco != OP_CVM)
	     && (opco != OP_MOVF2S) && (opco != OP_MOVS2F))
					       {result = vlrError(machPtr);
						if (result != TCL_OK) {
						  *flagstop = 1;
						  return TCL_ERROR;
						}
						}

	switch (opco) {
	 case OP_SEQSV: flagis = FP_SVEQ_OP; wordPtr->opCode = OP_SEQSV; break;
	 case OP_SEQV: flagis = FP_VEQ_OP; wordPtr->opCode = OP_SEQV;  break;
	 case OP_SGESV: flagis =FP_SVGE_OP; wordPtr->opCode = OP_SEQSV; break;
	 case OP_SGEV: flagis = FP_VGE_OP; wordPtr->opCode = OP_SEQV; break;
	 case OP_SGTSV: flagis = FP_SVGT_OP; wordPtr->opCode = OP_SEQSV; break;
	 case OP_SGTV: flagis = FP_VGT_OP; wordPtr->opCode = OP_SEQV; break;
	 case OP_SLESV: flagis = FP_SVLE_OP; wordPtr->opCode = OP_SEQSV; break;
	 case OP_SLEV:flagis = FP_VLE_OP; wordPtr->opCode = OP_SEQV; break;
	 case OP_SLTSV:flagis = FP_SVLT_OP; wordPtr->opCode = OP_SEQSV; break;
	 case OP_SLTV: flagis = FP_VLT_OP; wordPtr->opCode = OP_SEQV; break;
	 case OP_SNESV: flagis = FP_SVNE_OP; wordPtr->opCode = OP_SEQSV; break;
	 case OP_SNEV:flagis = FP_VNE_OP; wordPtr->opCode = OP_SEQV; break;
	 case OP_SUBV: flagis = V_SUB;  wordPtr->opCode = OP_ADDV; break;
	 case OP_ADDV: flagis = V_ADD;  wordPtr->opCode = OP_ADDV; break;
	 case OP_MULTV: flagis = V_MUL;  wordPtr->opCode = OP_ADDV; break;
	 case OP_DIVV: flagis = V_DIV;  wordPtr->opCode = OP_ADDV; break;
	 }

	switch (wordPtr->opCode) {
	 /* Operaciones vectoriales */

	    case OP_ADDV: {
	      long int stalls;
	      while ((stalls = FPissue (machPtr, flagis, FP_VECT_OP,
				       wordPtr->rs1 + 32, wordPtr->rs2 + 32,
				       wordPtr->rd + 32, &errorValue)) != 0) {
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if ((errorValue) && (flagis==V_DIV)) {
		*errorMessage = "Division by zero\n";
	        *flagstop = 1;
	      }
	      break;
	    }

	    case OP_ADDSV: {
	      long int stalls;
	      CheckS1D Check
	      while ((stalls = FPissue (machPtr, V_ADD, FP_SV_OP,
				       wordPtr->rs1, wordPtr->rs2 + 32,
				       wordPtr->rd + 32, &errorValue)) != 0) {
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_SUBSV: {
	      long int stalls;
	      CheckS1D Check
	      while ((stalls = FPissue (machPtr, V_SUB, FP_SV_OP,
				       wordPtr->rs1, wordPtr->rs2 + 32,
				       wordPtr->rd + 32, &errorValue)) != 0) {
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_SUBVS: {
	      long int stalls;
	      CheckS2D Check
	      while ((stalls = FPissue (machPtr, V_SUB, FP_VS_OP,
				       wordPtr->rs1 + 32, wordPtr->rs2,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_MULTSV: {
	      long int stalls;
	      CheckS1D Check
	      while ((stalls = FPissue (machPtr, V_MUL, FP_SV_OP,
				       wordPtr->rs1, wordPtr->rs2 + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_DIVSV: {
	      long int stalls;
	      CheckS1D Check
	      while ((stalls = FPissue (machPtr, V_DIV, FP_SV_OP,
				       wordPtr->rs1, wordPtr->rs2 + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if (errorValue)  {
		*errorMessage = "Division by zero\n";
	        *flagstop = 1;
	      }
	      break;
	    }

	    case OP_DIVVS: {
	      long int stalls;
	      CheckS2D Check
	      while ((stalls = FPissue (machPtr, V_DIV, FP_VS_OP,
				       wordPtr->rs1 + 32, wordPtr->rs2,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if (errorValue)  {
		*errorMessage = "Division by zero\n";
	      *flagstop = 1;
	      }
	      break;
	    }

	    case OP_SEQV: {
	      long int stalls;
	      while ((stalls = FPissue (machPtr, V_LOGIC, flagis,
				       wordPtr->rs1 + 32, wordPtr->rs2 + 32,
				       (int)VMREG, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_SEQSV: {
	      long int stalls;
	      CheckS1D Check
// printf("Antes de FPissue \n");
	      while ((stalls = FPissue (machPtr, V_LOGIC, flagis,
				       wordPtr->rs1, wordPtr->rs2 + 32,
				       (int)VMREG, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
// printf("Antes de FPwriteback \n");
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_POP: {
	    /* Esta instruccion no se simula como vectorial */
	    /* En versiones posteriores del simulador debera
            corregirse */
	      long int cuenta=0;
	      unsigned long int num1, num2;
	      CheckD Check
	      FPwait(machPtr, uno, (int)VMREG, cero, cero, cero);
	      num1 = (unsigned long int)(machPtr->regs[(int)(VLR_REG + 1)]);
	      num2 = (unsigned long int)(machPtr->regs[(int)(VLR_REG + 2)]);
	      while ((num1 != 0) || (num2 != 0))
	       {
		 cuenta += num1 % 2;
		 num1 /= 2;
		 cuenta += num2 % 2;
		 num2 /= 2;
	       }
	       machPtr->regs[wordPtr->rd] = cuenta;
	      break;
	      }

	    case OP_CVM: {
	      FPwait(machPtr, uno, (int)VMREG, cero, cero, cero);
	      machPtr->regs[(int)(VLR_REG + 1)] = -1;
	      machPtr->regs[(int)(VLR_REG + 2)] = -1;
	      break;
	      }

	    case OP_CVI: {
	     long int stalls;
	     CheckS1 Check
	      FPwait(machPtr, uno, (int)VMREG, cero, cero, cero);
	      errorValue = wordPtr->rs1; /* El registro se pasa en errorValue */
	      while ((stalls = FPissue (machPtr, V_MUL, FP_MASK_OP,
				       wordPtr->rd + 32, wordPtr->rd + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_MOVF2S: {
		CheckS1D Check
		FPwait(machPtr, tres, wordPtr->rs1, wordPtr->rs1 + 1, (int)VMREG, cero);
		machPtr->regs[(int)(VLR_REG + 1)] = machPtr->regs[wordPtr->rs1 + 32];
		machPtr->regs[(int)(VLR_REG + 2)] = machPtr->regs[wordPtr->rs1 + 33];
		break;
		}

	    case OP_MOVS2F: {
		CheckDD Check
		FPwait(machPtr, tres, wordPtr->rd, wordPtr->rd+1, (int)VMREG, cero);
		machPtr->regs[wordPtr->rd+32] = machPtr->regs[(int)(VLR_REG+1)];
		machPtr->regs[wordPtr->rd+33] = machPtr->regs[(int)(VLR_REG+2)];
		break;
		}

	    case OP_LV: {
	      long int stalls;
	      CheckS1 Check
	      tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (tmp & 0x3) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			*errorMessage="Error. Acces not aligned\n";
			*flagstop = 1;
			return TCL_ERROR;
		    }
		}
	      errorValue = tmp; /* La direccion va en errorValue! */
	      while ((stalls = FPissue (machPtr, V_LOAD, FP_LOAD_OP,
				       wordPtr->rd + 32, wordPtr->rd + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		errorValue = tmp;
	      }
	      if (errorValue) {
		*errorMessage="Error in LV\n";
		*flagstop = 1;}
	      break;
	    }

	    case OP_SV: {
	      long int stalls;
	      CheckS1 Check
	      tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
	      errorValue = tmp;  /* La direccion va en errorValue! */
	      while ((stalls = FPissue (machPtr, V_LOAD, FP_STORE_OP,
				       wordPtr->rd + 32, wordPtr->rd + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		errorValue = tmp;
	      }
	      if (errorValue) {
		*errorMessage="Error in SV\n";
		*flagstop = 1;}
	      break;
	    }

	    case OP_LVWS: {
	      long int stalls;
	      CheckS1 CheckS2 Check
	      tmp = machPtr->regs[wordPtr->rs1];
		if (tmp & 0x3) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			*flagstop = 1;
			return TCL_ERROR;
		    }
		}
	      errorValue = tmp;  /* La direccion va en errorValue! */
              /* Y el stride en una variable global! */
	      strid = machPtr->regs[wordPtr->rs2];
	      while ((stalls = FPissue (machPtr, V_LOAD, FP_LSTD_OP,
				       wordPtr->rd + 32, wordPtr->rd + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		errorValue = tmp;
              }
	      if (errorValue) {
		*errorMessage="Error in LVWS\n";
		*flagstop = 1;}
	      break;
	      }
//	    break;
//	    }

	    case OP_SVWS: {
	      long int stalls;
	      CheckS1 CheckD Check
	      tmp = machPtr->regs[wordPtr->rd];
	      errorValue = tmp;  /* La direccion va en errorValue! */
	      /* Y el stride en una variable global! */
	      strid = machPtr->regs[wordPtr->rs1];
	      while ((stalls = FPissue (machPtr, V_LOAD, FP_SSTD_OP,
				       wordPtr->rs2 + 32, wordPtr->rs2 + 32,
				       wordPtr->rs2 + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		errorValue = tmp;
	      }
	      if (errorValue) {
		*errorMessage="Error in SVWS\n";
		*flagstop = 1;}
	      break;
	    }

	    case OP_LVI: {
	      long int stalls;
	      CheckS1 Check
	      tmp = machPtr->regs[wordPtr->rs1];
		if (tmp & 0x3) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			*flagstop = 1;
			return TCL_ERROR;
		    }
		}
	      errorValue = tmp;  /* La direccion va en errorValue! */
	      while ((stalls = FPissue (machPtr, V_LOAD, FP_LINDEX_OP,
				       wordPtr->rs2 + 32, wordPtr->rs2 + 32,
				       wordPtr->rd + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		errorValue = tmp;
	      }
	      if (errorValue) {
		*errorMessage="Error in LVI\n";
		*flagstop = 1;}
	      break;
	    }

	    case OP_SVI: {
	      long int stalls;
	      CheckD Check
	      tmp = machPtr->regs[wordPtr->rd];
	      errorValue = tmp;  /* La direccion va en errorValue! */
	      while ((stalls = FPissue (machPtr, V_LOAD, FP_SINDEX_OP,
				       wordPtr->rs1 + 32, wordPtr->rs2 + 32,
				       wordPtr->rs2 + 32, &errorValue)) != 0){
		machPtr->Vstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		errorValue = tmp;
	      }
	      if (errorValue) {
		*errorMessage="Error in SVI\n";
		*flagstop = 1;}
	      break;
	    }

	}
 wordPtr->opCode = opco;
 return TCL_OK;
}

long int
Simulate(machPtr, singleStep, numarg, argumentos)
    register DLXV *machPtr;              /* Descripcion de la maquina. */
    long int singleStep;                /* No-cero significa que se ha de
					 * ejecutar una sola instruccion,
					 * a pesar de los breaks que se
					 * puedan encontrar. */
    int numarg;
    char argumentos[argus][largus];
{
    register MemWord huge *wordPtr;      /* Palabra de memoria de la
					  * instruccion. */
    register unsigned long int pc;
    unsigned long int tmp;
    long int i, result;
    char *errMsg;
    long int errorValue, flag;

    /*
     * No podemos continuar si se produce un error de direccionamiento
     * en el contador de programa.
     */
    if ((machPtr->badPC != 0) && (machPtr->addrErrNum == machPtr->insCount)) {
    wsprintf(salidas, "Error in finding inctruction, pc = 0x%lx",
		machPtr->badPC);
    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
	return TCL_ERROR;
    }

    machPtr->flags &= ~STOP_REQUESTED;
    machPtr->firstIns = machPtr->insCount;
/*    Io_BeginSim(machPtr); */

    if (iniciar) {_InitEasyWin();
		  easy = GetActiveWindow();
		  iniciar = FALSE;
		 }

    setCheckNum:
    if (machPtr->callBackList != NULL) {
	checkNum = machPtr->callBackList->serialNum;
    } else {
	checkNum = machPtr->insCount+100000L;
    }
    if ((machPtr->badPC != 0) && (machPtr->addrErrNum > machPtr->insCount)) {
	if (checkNum > machPtr->addrErrNum) {
	    checkNum = machPtr->addrErrNum;
	}
    } else {
	machPtr->badPC = 0;
    }
    if (singleStep) {
	checkNum = machPtr->insCount+1;
    }
    while (1) {
	long int last_pc, lastLoad1, lastLoad2, stalled, trapCaught;

	stalled = 0;
	trapCaught = 0;
	/* First handle a delayed load */
	if ((lastLoad1 = machPtr->loadReg1) != 0) {
	    machPtr->regs[(int)lastLoad1] = machPtr->loadValue1;
	    machPtr->loadReg1 = 0;
	    if ((lastLoad2 = machPtr->loadReg2) != 0) {
		machPtr->regs[(int)lastLoad2] = machPtr->loadValue2;
		machPtr->loadReg2 = 0;
	    }
	} else lastLoad2 = 0;

	/*
	 * Hacemos la busqueda de la instruccion, Calculamos el nuevo npc
	 * (pero no lo almacenamos todavia, por si acaso la instruccion no
	 * se puede completar).
	 */
	pc = machPtr->regs[(int)PC_REG];
	if (pc >= machPtr->numWords) {
	    result = BusError(machPtr, INDEX_TO_ADDR(pc), uno);
	    if (result != TCL_OK) {
		goto stopSimulation;
	    } else {
		goto endOfIns;
	    }
	}
	if (machPtr->refTraceFile)
	  fprintf(machPtr->refTraceFile, "2 %lx\n", INDEX_TO_ADDR(pc));
	wordPtr = &machPtr->memPtr[pc];
	last_pc = pc;
	pc = machPtr->regs[(int)NEXT_PC_REG]+1;

	/*
	 * Maneja las paradas de la simulacion, si no es la primera instruccion
	 * ejecutada.
	 */

	if ((wordPtr->stopList != NULL)
		&& (machPtr->insCount != machPtr->firstIns)) {
	    result = Stop_Execute(machPtr, wordPtr->stopList, numarg, argumentos);
	    if ((result != TCL_OK) || (machPtr->flags & STOP_REQUESTED)) {
		goto stopSimulation;
	    }
	}

	/*
	 * Ejecuta la instruccion.
	 */

	execute:
	machPtr->operationCount[wordPtr->opCode]++;
	if ((wordPtr->opCode >= OP_ADDSV) && (wordPtr->opCode <= OP_SVWS))
	{ Sim2(machPtr, wordPtr, &errMsg, &flag, stalled, lastLoad1, lastLoad2);
//	  if (flag) goto stopSimulation;
          if (flag) goto error;
	}
	else {

	switch (wordPtr->opCode) {

	    case OP_ADD: {
		long int sum;
		CheckS1 CheckS2 Check
		sum = machPtr->regs[wordPtr->rs1] + machPtr->regs[wordPtr->rs2];
		if (!((machPtr->regs[wordPtr->rs1] ^ machPtr->regs[wordPtr->rs2])
			& SIGN_BIT) && ((machPtr->regs[wordPtr->rs1]
			^ sum) & SIGN_BIT)) {
		    result = Overflow(machPtr);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		machPtr->regs[wordPtr->rd] = sum;
		break;
	    }

	    case OP_ADDI: {
		long int sum;
		CheckS1 Check
		sum = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (!((machPtr->regs[wordPtr->rs1] ^ wordPtr->extra)
			& SIGN_BIT) && ((wordPtr->extra ^ sum) & SIGN_BIT)) {
		    result = Overflow(machPtr);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		machPtr->regs[wordPtr->rd] = sum;
		break;
	    }

	    case OP_ADDU:
		CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			+ machPtr->regs[wordPtr->rs2];
		break;

	    case OP_ADDUI:
		CheckS1 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			+ (wordPtr->extra & 0xffff);
		break;

	    case OP_AND:
		CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			& machPtr->regs[wordPtr->rs2];
		break;

	    case OP_ANDI:
		CheckS1 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			& (wordPtr->extra & 0xffff);
		break;

	    case OP_BEQZ:
	      CheckS1 Check Check stalled=0; CheckRdstS1 CheckRdst /* 1/12/98 */
	      if (machPtr->regs[wordPtr->rs1] == 0) {
		pc = machPtr->regs[(int)NEXT_PC_REG] + ADDR_TO_INDEX(wordPtr->extra);
		machPtr->branchYes++;
	      }
	      else machPtr->branchNo++;
	      machPtr->branchSerial = machPtr->insCount;
	      machPtr->branchPC = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
	      break;

	    case OP_BFPF:
//	    printf ("BFPF \n %d ",machPtr->FPstatusReg);
	      if ((machPtr->FPstatusReg)!=1) {
		pc = machPtr->regs[(int)NEXT_PC_REG] + ADDR_TO_INDEX(wordPtr->extra);
		machPtr->branchYes++;
	      }
	      else machPtr->branchNo++;
	      machPtr->branchSerial = machPtr->insCount;
	      machPtr->branchPC = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
	      break;

	    case OP_BFPT:
//	    printf ("BFPT \n %d ",machPtr->FPstatusReg);
	      if (machPtr->FPstatusReg==1) {
		pc = machPtr->regs[(int)NEXT_PC_REG] + ADDR_TO_INDEX(wordPtr->extra);
		machPtr->branchYes++;
	      }
	      else machPtr->branchNo++;
	      machPtr->branchSerial = machPtr->insCount;
	      machPtr->branchPC = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
	      break;

	    case OP_BNEZ:
	      CheckS1 Check Check stalled=0; CheckRdstS1 CheckRdst /* 1/12/98 */
	      if (machPtr->regs[wordPtr->rs1] != 0) {
		pc = machPtr->regs[(int)NEXT_PC_REG] + ADDR_TO_INDEX(wordPtr->extra);
		machPtr->branchYes++;
	      }
	      else machPtr->branchNo++;
	      machPtr->branchSerial = machPtr->insCount;
	      machPtr->branchPC = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
	      break;

	    case OP_JAL:
		machPtr->regs[(int)R31] =
			INDEX_TO_ADDR((machPtr->regs[(int)NEXT_PC_REG] + 1));
	    case OP_J:
		pc = last_pc + 1 + ADDR_TO_INDEX(wordPtr->extra);
		machPtr->branchSerial = machPtr->insCount;
		machPtr->branchPC = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
		break;

	    case OP_JALR:
		machPtr->regs[(int)R31] =
			INDEX_TO_ADDR((machPtr->regs[(int)NEXT_PC_REG] + 1));
	    case OP_JR:
		CheckS1 Check
		tmp = machPtr->regs[wordPtr->rs1];
		pc = ADDR_TO_INDEX(tmp);
		if ((tmp & 0x3) && (machPtr->badPC == 0)) {
		    machPtr->badPC = tmp;
		    machPtr->addrErrNum = machPtr->insCount + 2;
		    if (checkNum > machPtr->addrErrNum) {
			checkNum = machPtr->addrErrNum;
		    }
		}
		machPtr->branchSerial = machPtr->insCount;
		machPtr->branchPC = INDEX_TO_ADDR(machPtr->regs[(int)PC_REG]);
		break;

	    case OP_LB:
	    case OP_LBU: {
		long int value;
		CheckS1 Check
		tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (ReadMem(machPtr, tmp, &value, numarg, argumentos) == 0) {
			goto stopSimulation;
		}

		switch (tmp & 0x3) {
		    case 0:
			value >>= 24;
			break;
		    case 1:
			value >>= 16;
			break;
		    case 2:
			value >>= 8;
		}
		if ((value & 0x80) && (wordPtr->opCode == OP_LB)) {
		    value |= 0xffffff00L;
		} else {
		    value &= 0xff;
		}
		machPtr->loadValue1 = value;
		machPtr->loadReg1 = wordPtr->rd;
		break;
	    }

	    case OP_LD: {
		long int valueHi, valueLo;
		CheckS1 Check
		FPwait (machPtr, dos, wordPtr->rd, wordPtr->rd + 1, cero, cero);
		tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (tmp & 0x3) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		if (ReadMem(machPtr, tmp, &valueHi, numarg, argumentos) == 0) {
		  goto stopSimulation;
		}
		if (ReadMem(machPtr, tmp+4, &valueLo, numarg, argumentos) == 0) {
		  goto stopSimulation;
		}
		machPtr->loadValue1 = valueHi;
		machPtr->loadValue2 = valueLo;
		machPtr->loadReg1 = wordPtr->rd + 32;
		machPtr->loadReg2 = wordPtr->rd + 32 + 1;
		break;
	      }

	    case OP_LF: {
		long int value;
		CheckS1 Check
		FPwait (machPtr, uno, wordPtr->rd, cero, cero, cero);
		tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (tmp & 0x3) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		if (ReadMem(machPtr, tmp, &value, numarg, argumentos) == 0) {
		    goto stopSimulation;
		}
		machPtr->loadValue1 = value;
		machPtr->loadReg1 = wordPtr->rd + 32;
		break;
	      }

	    case OP_LH:
	    case OP_LHU: {
		long int value;
		CheckS1 Check
		tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (tmp & 0x1) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		if (ReadMem(machPtr, tmp, &value, numarg, argumentos) == 0) {
		    goto stopSimulation;
		}
		if (!(tmp & 0x2)) {
		    value >>= 16;
		}
		if ((value & 0x8000) && (wordPtr->opCode == OP_LH)) {
		    value |= 0xffff0000L;
		} else {
		    value &= 0xffff;
		}
		machPtr->loadValue1 = value;
		machPtr->loadReg1 = wordPtr->rd;
		break;
	    }

	    case OP_LHI:
		machPtr->regs[wordPtr->rd] = wordPtr->extra << 16;
		break;

	    case OP_LW: {
		long int value;
		CheckS1 Check
		tmp = machPtr->regs[wordPtr->rs1] + wordPtr->extra;
		if (tmp & 0x3) {
		    result = AddressError(machPtr, tmp, uno);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		if (ReadMem(machPtr, tmp, &value, numarg, argumentos) == 0) {
		    goto stopSimulation;
		}
		machPtr->loadValue1 = value;
		machPtr->loadReg1 = wordPtr->rd;
		break;
	      }

	    case OP_MOVD: {
	      double *source, *dest;
	      CheckS1D Check
	      FPwait (machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		      wordPtr->rd, wordPtr->rd + 1);
	      source = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (double *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = *source;
	      break;
	    }

	    case OP_MOVF: {
	      float *source, *dest;
	      CheckS1F Check
	      FPwait (machPtr, dos, wordPtr->rs1, wordPtr->rd, cero, cero);
	      source = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (float *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = *source;
	      break;
	    }

	    case OP_MOVFP2I:
	      CheckS1F Check
	      FPwait (machPtr, uno, wordPtr->rs1, cero, cero, cero);
	      machPtr->regs[wordPtr->rd] = machPtr->regs[32 + wordPtr->rs1];
	      break;

	    case OP_MOVI2FP:
	      CheckS1 Check
	      FPwait (machPtr, uno, wordPtr->rd, cero, cero, cero);
	      machPtr->regs[32 + wordPtr->rd] = machPtr->regs[wordPtr->rs1];
	      break;

	    case OP_MOVI2S:
	      CheckS1 Check
	      machPtr->regs[(int)VLR_REG] = machPtr->regs[wordPtr->rs1];
	      break;

	    case OP_MOVS2I:
	      CheckD Check
	      machPtr->regs[wordPtr->rd] = machPtr->regs[(int)VLR_REG];
	      break;

	    case OP_OR:
	      CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			| machPtr->regs[wordPtr->rs2];
		break;

	    case OP_ORI:
		CheckS1 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			| (wordPtr->extra & 0xffff);
		break;

	    case OP_RFE:
		errMsg = "Instruction RFE not implemented\n";
		goto error;

	    case OP_SB:
		CheckS1 CheckD Check
		if (WriteMem(machPtr, (unsigned long int) (machPtr->regs[wordPtr->rs1]
				+ wordPtr->extra),
			uno, machPtr->regs[wordPtr->rd], numarg, argumentos) == 0) {
		    goto stopSimulation;
		}
		break;

	    case OP_SD:
		CheckS1 CheckDD Check
	      FPwait (machPtr, dos, wordPtr->rd, wordPtr->rd + 1, cero, cero);
	      if (WriteMem(machPtr, (unsigned long int) (machPtr->regs[wordPtr->rs1]
						+ wordPtr->extra),
			   cuatro, machPtr->regs[wordPtr->rd + 32], numarg, argumentos) == 0)
		goto stopSimulation;
	      if (WriteMem(machPtr, (unsigned long int) (machPtr->regs[wordPtr->rs1]
						+ wordPtr->extra + 4),
			   cuatro, machPtr->regs[wordPtr->rd + 32 + 1], numarg, argumentos) == 0)
		goto stopSimulation;
	      break;


	    case OP_SEQ:
		CheckS1 CheckS2 Check
		if (machPtr->regs[wordPtr->rs1] == machPtr->regs[wordPtr->rs2]) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SEQI:
		CheckS1 Check
		if (machPtr->regs[wordPtr->rs1] == wordPtr->extra) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SEQU:
		CheckS1 CheckS2 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) ==
		    ((unsigned long int) machPtr->regs[wordPtr->rs2])) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SEQUI:
		CheckS1 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) ==
		    ((unsigned long int) wordPtr->extra)) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SF:
	      CheckS1 CheckDF Check
	      FPwait (machPtr, uno, wordPtr->rd, cero, cero, cero);
	      if (WriteMem(machPtr, (unsigned long int) (machPtr->regs[wordPtr->rs1]
						+ wordPtr->extra),
			   cuatro, machPtr->regs[wordPtr->rd + 32], numarg, argumentos) == 0)
		goto stopSimulation;
	      break;

	    case OP_SGE:
		CheckS1 CheckS2 Check
		if (machPtr->regs[wordPtr->rs1] >= machPtr->regs[wordPtr->rs2]) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGEI:
		CheckS1 Check
		if (machPtr->regs[wordPtr->rs1] >= wordPtr->extra) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGEU:
		CheckS1 CheckS2 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) >=
		    ((unsigned long int) machPtr->regs[wordPtr->rs2])) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGEUI:
		CheckS1 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) >=
		    ((unsigned long int) wordPtr->extra)) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGT:
		CheckS1 CheckS2 Check
		if (machPtr->regs[wordPtr->rs1] > machPtr->regs[wordPtr->rs2]) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGTI:
		CheckS1 Check
		if (machPtr->regs[wordPtr->rs1] > wordPtr->extra) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGTU:
		CheckS1 CheckS2 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) >
		    ((unsigned long int) machPtr->regs[wordPtr->rs2])) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SGTUI:
		CheckS1 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) >
		    ((unsigned long int) wordPtr->extra)) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SH:
		CheckS1 CheckD Check
		if (WriteMem(machPtr, (unsigned long int) (machPtr->regs[wordPtr->rs1]
				+ wordPtr->extra),
			dos, machPtr->regs[wordPtr->rd], numarg, argumentos) == 0) {
		    goto stopSimulation;
		}
		break;

	    case OP_SLE:
		CheckS1 CheckS2 Check
		if (machPtr->regs[wordPtr->rs1] <= machPtr->regs[wordPtr->rs2]) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLEI:
		CheckS1 Check
		if (machPtr->regs[wordPtr->rs1] <= wordPtr->extra) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLEU:
		CheckS1 CheckS2 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) <=
		    ((unsigned long int) machPtr->regs[wordPtr->rs2])) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLEUI:
		CheckS1 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) <=
		    ((unsigned long int) wordPtr->extra)) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLL:
		CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			<< (machPtr->regs[wordPtr->rs2] & 0x1f);
		break;

	    case OP_SLLI:
		CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			<< (wordPtr->extra & 0x1f);
		break;

	    case OP_SLT:
		CheckS1 CheckS2 Check
		if (machPtr->regs[wordPtr->rs1] < machPtr->regs[wordPtr->rs2]) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLTI:
		CheckS1 Check
		if (machPtr->regs[wordPtr->rs1] < wordPtr->extra) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLTU:
		CheckS1 CheckS2 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) <
		    ((unsigned long int) machPtr->regs[wordPtr->rs2])) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SLTUI:
		CheckS1 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) <
		    ((unsigned long int) wordPtr->extra)) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SNE:
		CheckS1 CheckS2 Check
		if (machPtr->regs[wordPtr->rs1] != machPtr->regs[wordPtr->rs2]) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SNEI:
		CheckS1 Check
		if (machPtr->regs[wordPtr->rs1] != wordPtr->extra) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SNEU:
		CheckS1 CheckS2 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) !=
		    ((unsigned long int) machPtr->regs[wordPtr->rs2])) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SNEUI:
		CheckS1 Check
		if (((unsigned long int) machPtr->regs[wordPtr->rs1]) !=
		    ((unsigned long int) wordPtr->extra)) {
		    machPtr->regs[wordPtr->rd] = 1;
		} else {
		    machPtr->regs[wordPtr->rd] = 0;
		}
		break;

	    case OP_SRA:
		CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			>> (machPtr->regs[wordPtr->rs2] & 0x1f);
		break;

	    case OP_SRAI:
	      CheckS2 Check
	      machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
		>> (wordPtr->extra & 0x1f);
	      break;

	    case OP_SRL:
	      CheckS1 CheckS2 Check
	      tmp = machPtr->regs[wordPtr->rs1];
	      tmp >>= (machPtr->regs[wordPtr->rs2] & 0x1f);
	      machPtr->regs[wordPtr->rd] = tmp;
	      break;

	    case OP_SRLI:
		CheckS2 Check
		tmp = machPtr->regs[wordPtr->rs1];
		tmp >>= (wordPtr->extra & 0x1f);
		machPtr->regs[wordPtr->rd] = tmp;
		break;

	    case OP_SUB: {
		long int diff;

		CheckS1 CheckS2 Check
		diff = machPtr->regs[wordPtr->rs1] - machPtr->regs[wordPtr->rs2];
		if (((machPtr->regs[wordPtr->rs1] ^ machPtr->regs[wordPtr->rs2])
			& SIGN_BIT) && ((machPtr->regs[wordPtr->rs1]
			^ diff) & SIGN_BIT)) {
		    result = Overflow(machPtr);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		machPtr->regs[wordPtr->rd] = diff;
		break;
	    }

	    case OP_SUBI: {
		long int diff;

		CheckS1 Check
		diff = machPtr->regs[wordPtr->rs1] - wordPtr->extra;
		if (((machPtr->regs[wordPtr->rs1] ^ wordPtr->extra)
			& SIGN_BIT) && ((machPtr->regs[wordPtr->rs1]
			^ diff) & SIGN_BIT)) {
		    result = Overflow(machPtr);
		    if (result != TCL_OK) {
			goto stopSimulation;
		    } else {
			goto endOfIns;
		    }
		}
		machPtr->regs[wordPtr->rd] = diff;
		break;
	    }

	    case OP_SUBU:
		CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			- machPtr->regs[wordPtr->rs2];
		break;

	    case OP_SUBUI:
		CheckS1 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			- (wordPtr->extra & 0xffff);
		break;

	    case OP_SW:
		CheckS1 CheckD Check
		if (WriteMem(machPtr, (unsigned long int) (machPtr->regs[wordPtr->rs1]
				+ wordPtr->extra),
			cuatro, machPtr->regs[wordPtr->rd], numarg, argumentos) == 0) {
		    goto stopSimulation;
		}
		break;

	    case OP_TRAP:
		switch (Handle_Trap(machPtr, (wordPtr->extra & 0x3ffffffL))) {
		case 1 :
		  wsprintf(salidas,"TRAP #%ld failed", wordPtr->extra & 0x3ffffffL);
		  BWCCMessageBox(GetFocus(), salidas, "TRAPS", MB_OK | MB_ICONINFORMATION);
		  trapCaught = 1;
		  break;
		case 2 :
		  wsprintf(salidas,"Reached TRAP #%ld", wordPtr->extra & 0x3ffffffL);
		  BWCCMessageBox(GetFocus(), salidas, "TRAPS", MB_OK | MB_ICONINFORMATION);
		  trapCaught = 1;
		  break;
		case 3 :
		  wsprintf(salidas,"Reached TRAP #%ld.\nFinalizing operations.", wordPtr->extra & 0x3ffffffL);
		  BWCCMessageBox(GetFocus(), salidas, "TRAPS", MB_OK | MB_ICONINFORMATION);
		  Terminar(machPtr);
		  trapCaught = 1;
		  break;
		}
		break;

	    case OP_XOR:
		CheckS1 CheckS2 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			^ machPtr->regs[wordPtr->rs2];
		break;

	    case OP_XORI:
		CheckS1 Check
		machPtr->regs[wordPtr->rd] = machPtr->regs[wordPtr->rs1]
			^ (wordPtr->extra & 0xffff);
		break;

	case OP_NOP:
	  break; 

/* Operaciones en coma flotante */

	    case OP_DIV: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, INT_DIV, FP_INT_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if (errorValue) {
		errMsg = "Division by zero\n";
		goto error;
	      }
	      break;
	    }

	    case OP_DIVU: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, INT_UDIV, FP_INT_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if (errorValue) {
		errMsg = "Division by zero\n";
		goto error;
	      }
	      break;
	    }

	    case OP_MULT: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, INT_MUL, FP_INT_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if (errorValue) {
		errMsg = "Multiplication overflow\n";
		goto error;
	      }
	      break;
	    }

	    case OP_MULTU: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, INT_UMUL, FP_INT_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      if (errorValue) {
		errMsg = "Multiplication overflow\n";
		goto error;
	      }
	      break;
	    }

	    case OP_ADDD: {
	      long int stalls;
	      CheckS1D CheckS2D Check
	      while ((stalls = FPissue (machPtr, FP_ADD, FP_DOUBLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_ADDF: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, FP_ADD, FP_SINGLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_CVTD2F: {
	      double *source;
	      float *dest;
	      CheckS1D Check
	      FPwait(machPtr, tres, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rd, cero);
	      source = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (float *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = (float) *source;
	      break;
	    }

	    case OP_CVTD2I: {
	      double *source;
	      long int *dest;
	      CheckS1D Check
	      FPwait(machPtr, tres, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rd, cero);
	      source = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (long int *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = (long int) *source;
	      break;
	    }

	    case OP_CVTF2D: {
	      float *source;
	      double *dest;
	      CheckS1F Check
	      FPwait(machPtr, tres, wordPtr->rs1,
		     wordPtr->rd, wordPtr->rd + 1, cero);
	      source = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (double *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = (double) *source;
	      break;
	    }

	    case OP_CVTF2I: {
	      float *source;
	      long int *dest;
	      CheckS1F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rd, cero, cero);
	      source = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (long int *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = (long int) *source;
	      break;
	    }

	    case OP_CVTI2D: {
	      long int *source;
	      double *dest;
	      CheckS1F Check
	      FPwait(machPtr, tres, wordPtr->rs1, wordPtr->rd, wordPtr->rd + 1, cero);
	      source = (long int *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (double *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = (double) *source;
	      break;
	    }
	    case OP_CVTI2F: {
	      long int *source;
	      float *dest;
	      CheckS1F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rd, cero, cero);
	      source = (long int *) & (machPtr->regs[32 + wordPtr->rs1]);
	      dest = (float *) & (machPtr->regs[32 + wordPtr->rd]);
	      *dest = (float) *source;
	      break;
	    }

	    case OP_DIVD: {
	      long int stalls;
	      CheckS1D CheckS2D Check
	      while ((stalls = FPissue (machPtr, FP_DIV, FP_DOUBLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		}
		if (errorValue) {
		  errMsg = "Division by zero\n";
		goto error;
	        }
	      break;
	    }

	    case OP_DIVF: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, FP_DIV, FP_SINGLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
		}
		if (errorValue) {
		  errMsg = "Division by zero\n";
		goto error;
	        }
	      break;
	    }

	    case OP_EQD: {
	      double *source1, *source2;
	      CheckS1D CheckS2D Check
	      FPwait(machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rs2, wordPtr->rs2 + 1);
	      source1 = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (double *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 == *source2);
	      break;
	    }

	    case OP_EQF: {
	      float *source1, *source2;
	      CheckS1F CheckS2F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rs2, cero, cero);
	      source1 = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (float *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 == *source2);
	      break;
	    }

	    case OP_GED: {
	      double *source1, *source2;
	      CheckS1D CheckS2D Check
	      FPwait(machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rs2, wordPtr->rs2 + 1);
	      source1 = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (double *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 >= *source2);
//	      printf("GED \n %f %f %d",*source1,*source2,machPtr->FPstatusReg);
	      break;
	    }

	    case OP_GEF: {
	      float *source1, *source2;
	      CheckS1F CheckS2F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rs2, cero, cero);
	      source1 = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (float *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 >= *source2);
	      break;
	    }

	    case OP_GTD: {
	      double *source1, *source2;
	      CheckS1D CheckS2D Check
	      FPwait(machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rs2, wordPtr->rs2 + 1);
	      source1 = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (double *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 > *source2);
	      break;
	    }

	    case OP_GTF: {
	      float *source1, *source2;
	      CheckS1F CheckS2F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rs2, cero, cero);
	      source1 = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (float *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 > *source2);
	      break;
	    }

	    case OP_LED: {
	      double *source1, *source2;
	      CheckS1D CheckS2D Check
	      FPwait(machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rs2, wordPtr->rs2 + 1);
	      source1 = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (double *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 <= *source2);
	      break;
	    }

	    case OP_LEF: {
	      float *source1, *source2;
	      CheckS1F CheckS2F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rs2, cero, cero);
	      source1 = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (float *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 <= *source2);
	      break;
	    }

	    case OP_LTD: {
	      double *source1, *source2;
	      CheckS1D CheckS2D Check
	      FPwait(machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rs2, wordPtr->rs2 + 1);
	      source1 = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (double *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 < *source2);
	      break;
	    }

	    case OP_LTF: {
	      float *source1, *source2;
	      CheckS1F CheckS2F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rs2, cero, cero);
	      source1 = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (float *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 < *source2);
	      break;
	    }

	    case OP_MULTD: {
	      long int stalls;
	      CheckS1D CheckS2D Check
	      while ((stalls = FPissue (machPtr, FP_MUL, FP_DOUBLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_MULTF: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, FP_MUL, FP_SINGLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_NED: {
	      double *source1, *source2;
	      CheckS1D CheckS2D Check
	      FPwait(machPtr, cuatro, wordPtr->rs1, wordPtr->rs1 + 1,
		     wordPtr->rs2, wordPtr->rs2 + 1);
	      source1 = (double *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (double *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 != *source2);
	      break;
	    }

	    case OP_NEF: {
	      float *source1, *source2;
	      CheckS1F CheckS2F Check
	      FPwait(machPtr, dos, wordPtr->rs1, wordPtr->rs2, cero, cero);
	      source1 = (float *) & (machPtr->regs[32 + wordPtr->rs1]);
	      source2 = (float *) & (machPtr->regs[32 + wordPtr->rs2]);
	      machPtr->FPstatusReg = (*source1 != *source2);
	      break;
	    }

	    case OP_SUBD: {
	      long int stalls;
	      CheckS1D CheckS2D Check
	      while ((stalls = FPissue (machPtr, FP_SUB, FP_DOUBLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }

	    case OP_SUBF: {
	      long int stalls;
	      CheckS1F CheckS2F Check
	      while ((stalls = FPissue (machPtr, FP_SUB, FP_SINGLE_FP_OP,
				       wordPtr->rs1, wordPtr->rs2,
				       wordPtr->rd, &errorValue)) != 0){
		machPtr->FPstalls += stalls;
		machPtr->cycleCount += stalls;
		FPwriteBack (machPtr);
	      }
	      break;
	    }


	    case OP_NOT_COMPILED:
		Compile(wordPtr);
		goto execute;
/* ***          break; *** */

	    case OP_RES:
		errMsg = "Operacion reserved\n";
		goto error;

	    case OP_UNIMP:
		errMsg = "Instruction not implemented\n";
		goto error;

	    default:
		i = wordPtr->opCode;
		wsprintf(salidas,
			"Internal error in Simulate():  Operation code error %ld, pc = %.100s",
			i, Sym_GetString(machPtr, Sim_GetPC(machPtr)));
		BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
		machPtr->result[0] = '\0';
		result = TCL_ERROR;
		goto stopSimulation;
	    }
	/* 1/12/98 Actualizamos la variable que aloja el ultimo registro destino, solo si la instruccion
	ejecutada ha sido entera, ha escrito algo en el registro destino y no es una carga, ya que estas
        se gestionan por separado */
	if ((wordPtr->opCode <= 74) && (wordPtr->opCode != OP_LF) && (wordPtr->opCode != OP_LD)
	    && (wordPtr->opCode != OP_LB) && (wordPtr->opCode != OP_LBU)
	    && (wordPtr->opCode != OP_LH) && (wordPtr->opCode != OP_LHU) && (wordPtr->opCode != OP_LW)
	    && (wordPtr->opCode != OP_SB) && (wordPtr->opCode != OP_SW) && (wordPtr->opCode != OP_SF)
	    && (wordPtr->opCode != OP_SD) && (wordPtr->opCode != OP_MOVD) && (wordPtr->opCode != OP_MOVF)
	    && (wordPtr->opCode != OP_MOVFP2I)  && (wordPtr->opCode != OP_MOVI2FP)
	    && (wordPtr->opCode != OP_MOVI2S)  && (wordPtr->opCode != OP_MOVS2I)) {
	    machPtr->lastRdst = wordPtr->rd; /* 1/12/98 */
	   }
	else {
	    machPtr->lastRdst = 0; /* 1/12/98 */
           }
	}

	/*
	 * Monitoriza el numero de ciclos y las unidades de proceso.
	 */

	machPtr->cycleCount++;
	if (machPtr->checkFP && machPtr->cycleCount >= machPtr->checkFP)
	  FPwriteBack (machPtr);

	/* lanza, si puede, los datos pendientes de las operaciones
	 * vectoriales */
	Vwrite(machPtr);

	/* pone a cero el contador de ciclos periodicamente, para evitar
	 * problemas de de wrap around */
	if (machPtr->cycleCount >= CYC_CNT_RESET) {
	  int i;
	  FPop *opPtr = machPtr->FPopsList;
	  for (i = 0; i < (32 + NUM_V_REGS + 1); i++){
	    if (machPtr->waitingE[i])
	      machPtr->waitingE[i] -= CYC_CNT_RESET - 5;
	    if (machPtr->waiting_FPRs[i])
	      machPtr->waiting_FPRs[i] -= CYC_CNT_RESET - 5;
	      }
	  for (i = 0; i< 8; i++)
	    if (machPtr->waitS[i])
		machPtr->waitS[i] -= CYC_CNT_RESET - 5;
	  machPtr->cycleCount -= CYC_CNT_RESET - 5;
	  if (machPtr->checkFP)
	    machPtr->checkFP -= CYC_CNT_RESET - 5;
	  while (opPtr != NULL) {
	    opPtr->ready -= CYC_CNT_RESET - 5;
	    opPtr->ready2 -= CYC_CNT_RESET - 5;
	    opPtr = opPtr->nextPtr;
	  }
	}

	/*
	 * Nos aseguramos de que R0 permanece con el valor a cero.
	 */

	machPtr->regs[0] = 0;

	/*
	 * Avanza el contador de programa.
	 */

	machPtr->regs[(int)PC_REG] = machPtr->regs[(int)NEXT_PC_REG];
	machPtr->regs[(int)NEXT_PC_REG] = pc;

	/*
	 * Comprueba los flags para acciones especiales a realizar
	 * despues de la ejecucion de la instruccion.
	 */

	if (trapCaught) {
	    result = TCL_OK;
	    goto stopSimulation;
	}
	if ((machPtr->insCount += 1) >= checkNum) {
	    while (machPtr->callBackList != NULL) {
		register CallBack *cbPtr;

		cbPtr = machPtr->callBackList;
		if (machPtr->insCount < cbPtr->serialNum) {
		    break;
		}
		machPtr->callBackList = cbPtr->nextPtr;
		(*cbPtr->proc)(cbPtr->clientData, machPtr);
		farfree((char far *) cbPtr);
	    }
	    if ((machPtr->badPC != 0)
		    && (machPtr->insCount == machPtr->addrErrNum)) {
		result = AddressError(machPtr, machPtr->badPC, uno);
		if (result != TCL_OK) {
		    goto stopSimulation;
		}
	    }
	    if (singleStep) {
		tmp = Sim_GetPC(machPtr);
		sprintf((char far *)machPtr->result,
			"Pause in Step by Step, pc = %.100s: %.50s\n",
			Sym_GetString(machPtr, tmp),
			Asm_Disassemble(machPtr,
				machPtr->memPtr[machPtr->regs[(int)PC_REG]].value,
				tmp & ~0x3));
		result = TCL_OK;
		goto stopSimulation;
	    }
	    if (machPtr->flags & STOP_REQUESTED) {
		errMsg = "Execution halted\n";
		goto error;
	    }
	    goto setCheckNum;
	}

	endOfIns:
	continue;
    }

    error:
    tmp = Sim_GetPC(machPtr);
    wsprintf(salidas, "%spc = %.50s => %.60s", errMsg,
	    Sym_GetString(machPtr, tmp),
	    Asm_Disassemble(machPtr,
	    machPtr->memPtr[machPtr->regs[(int)PC_REG]].value, tmp & ~0x3));
    BWCCMessageBox(GetFocus(), salidas, "Simulacion", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    result = TCL_ERROR;


    stopSimulation:
    Io_EndSim(machPtr);
/*    sprintf(msg, "%ld", machPtr->insCount); */
/*    Tcl_SetVar(machPtr->interp, "insCount", msg, 1); */
    if (machPtr->refTraceFile)
      fflush(machPtr->refTraceFile);

    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * Compile --
 *
 *      Dada una palabra de memoria, la decodifica en un conjunto de campos
 *      que permiten una interpretacion mucho mas rapida.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      El contenido de *wordPtr es modificado.
 *
 *----------------------------------------------------------------------
 */

static void
Compile(wordPtr)
    register MemWord *wordPtr;     /* Posicion de memoria a ser compilada. */
{
    register OpInfo *opPtr;

    wordPtr->rs1 = (wordPtr->value >> 21) & 0x1f;
    wordPtr->rs2 = (wordPtr->value >> 16) & 0x1f;
    wordPtr->rd = (wordPtr->value >> 11) & 0x1f;
    opPtr = &opTable[(unsigned int)((wordPtr->value >> 26) & 0x3f)];
    wordPtr->opCode = opPtr->opCode;
    if (opPtr->format == IFMT) {
	wordPtr->rd = wordPtr->rs2;
	wordPtr->extra = wordPtr->value & 0xffff;
	if (wordPtr->extra & 0x8000) {
	    wordPtr->extra |= 0xffff0000L;
	}
    } else if (opPtr->format == RFMT) {
	wordPtr->extra = (wordPtr->value >> 6) & 0x1f;
    } else {
	wordPtr->extra = wordPtr->value & 0x3ffffffL;
	if (wordPtr->extra & 0x2000000L) {
	    wordPtr->extra |= 0xfc000000L;
	}
    }
    if (wordPtr->opCode == SPECIAL) {
      wordPtr->opCode = specialTable[(unsigned int)(wordPtr->value & 0x3f)];
    }
    else if (wordPtr->opCode == FPARITH) {
      wordPtr->opCode = FParithTable[(unsigned int)(wordPtr->value & 0x3f)];
    }
  }

/*
 *----------------------------------------------------------------------
 *
 * BusError --
 *
 *      Maneja las excepciones debidas a errores de bus.
 *
 * Resultados:
 *      TCL_OK ( sin string de retorno y se puede seguir la simulacion)
 *      TCL_ERROR ( debe pararse la simulacion)
 *
 * Efectos Laterales:
 *      Puede simular una trap de la maquina.
 *
 *----------------------------------------------------------------------
 */

/* ARGSUSED */
static long int
BusError(machPtr, address, iFetch)
    DLXV *machPtr;               /* Descripcion de la maquina. */
    unsigned long int address;  /* Posicion que fue referenciada pero que
				 * no existe. */
    long int iFetch;            /* 1 significa que un error ocurrio durante
				 * la busqueda de la instruccion. 0
				 * significa que se produjo durante una carga
				 * o un almacenamiento. */
{
    unsigned long int pcAddr;

    pcAddr = Sim_GetPC(machPtr);
    if (iFetch) {
	wsprintf(salidas,
		"It's recommanded searching from position 0x%lx",
		address);
	BWCCMessageBox(GetFocus(), salidas, "Error de Bus", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
    } else {
	wsprintf(salidas,
		"Position reference 0x%lx during instruction Check/Store, pc = %.100s: %s",
		address, Sym_GetString(machPtr, pcAddr),
		Asm_Disassemble(machPtr,
			machPtr->memPtr[machPtr->regs[(int)PC_REG]].value,
			pcAddr & ~0x3));
	BWCCMessageBox(GetFocus(), salidas, "Bus error", MB_OK | MB_ICONEXCLAMATION);
	machPtr->result[0] = '\0';
    }
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * AddressError --
 *
 *      Maneja las excepciones producidas por errores de direccionamiento.
 *
 * Resultados:
 *      TCL_OK ( sin string de retorno y se puede seguir la simulacion)
 *      TCL_ERROR ( debe pararse la simulacion)
 *
 * Efectos Laterales:
 *      Puede simular una trap de la maquina.
 *
 *----------------------------------------------------------------------
 */

#pragma argsused
static long int
AddressError(machPtr, address, load)
    DLXV *machPtr;               /* Descripcion de la maquina. */
    unsigned long int address;  /* Posicion que fue referenciada pero que no
				 * existe. */
    long int load;              /* 1 significa que un error ocurrio durante
				 * la busqueda de la instruccion o durante una
				 * carga. 0 significa que se produjo durante
				 * un almacenamiento. */
{
    wsprintf(salidas,
	    "Reference 0x%lx, pc = %.100s", address,
	    Sym_GetString(machPtr, Sim_GetPC(machPtr)));
    BWCCMessageBox(GetFocus(), salidas, "Position error", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * vlrError --
 *
 *      Maneja el error de que el registro vlr contenga un valor mayor de MVL.
 *
 * Resultados:
 *      TCL_OK, entonces no se devuelve ningun string y se puede continuar
 *      con la simulacion.
 *      TCL_ERROR, entonces se debe parar la simulacion.
 *
 * Efectos Laterales:
 *      Puede simular una trap de la maquina.
 *
 *----------------------------------------------------------------------
 */

static long int
vlrError(machPtr)
    DLXV *machPtr;               /* Descripcion de la maquina. */
{
    wsprintf(salidas,
	    "Register content \"vlr\" greater then maximum %ld", MVL);
    BWCCMessageBox(GetFocus(), salidas, "Register error", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * Overflow --
 *
 *      Maneja las excepciones aritmeticas de overflow
 *
 * Resultados:
 *      TCL_OK, entonces no se devuelve ningun string y se puede continuar
 *      con la simulacion.
 *      TCL_ERROR, entonces se debe parar la simulacion.
 *
 * Efectos Laterales:
 *      Puede simular una trap de la maquina.
 *
 *----------------------------------------------------------------------
 */

/* ARGSUSED */
static long int
Overflow(machPtr)
    DLXV *machPtr;               /* Descripcion de la maquina. */
{
    unsigned long int pcAddr;

    pcAddr = Sim_GetPC(machPtr);
    wsprintf(salidas, "Overflow, pc = %.100s: %s\n",
	    Sym_GetString(machPtr, pcAddr),  Asm_Disassemble(machPtr,
		    machPtr->memPtr[machPtr->regs[(int)PC_REG]].value,
		    pcAddr & ~0x3));
    BWCCMessageBox(GetFocus(), salidas, "Simulacion", MB_OK | MB_ICONEXCLAMATION);
    machPtr->result[0] = '\0';
    return TCL_ERROR;
}


/*
 *----------------------------------------------------------------------
 *
 * Mult --
 *
 *      Simula la multiplicacion en el DLXV
 *
 * Resultados:
 *      La palabra en *resultPtr es sobreescrita con el resultado de la
 *      multiplicacion y cero es devuelto si no hay overflow. Si existe
 *      overflow, un entero negativo es devuelto y *resultPtr no es
 *      cambiado.
 *
 * Efectos Laterales:
 *      Ninguno.
 *
 *----------------------------------------------------------------------
 */

static long int
Mult(a, b, signedArith, resultPtr)
    long int a, b;              /* Los dos operandos a multiplicar. */
    long int signedArith;       /* Cero significa hacer la operacion sin
				 * signo. Uno significa hacer la operacion
				 * con signo. */
     long int *resultPtr;       /* Aqui se pone el resultado. */
{
    unsigned long int aHi, aLo, bHi, bLo, res1, res2, res3, res4, res5;
    unsigned long int *resPtr;
    long int negative;

    if ((a == 0) || (b == 0)) {
	*resultPtr = 0;
	return 0;
    }

    /*
     * Calcula el signo del resultado, entonces lo hace todo positivo
     * de manera que se pueda realizar el calculo principal sin signo.
     */

    negative = 0;
    if (signedArith) {
	if (a < 0) {
	    negative = !negative;
	    a = -a;
	}
	if (b < 0) {
	    negative = !negative;
	    b = -b;
	}
    }

    /*
     * Calcula el resultado en aritmetica sin signo (Comprueba los bits de a
     * uno por uno y suma el valor desplazado de b).
     */

    aLo = (unsigned long int) (a & 0xffff);
    aHi = ((unsigned long int) a) >> 16;
    bLo = (unsigned long int) (b & 0xffff);
    bHi = ((unsigned long int) b) >> 16;

    /* printf ("\naHi %ld aLo %ld bHi %ld bLo %ld\n", aHi, aLo, bHi, bLo); */

    if (aHi && bHi) return -1;

    res1 = aLo * bLo;
    res2 = aHi * bLo;
    res3 = aLo * bHi;
    if ((res2 | res3) & 0xffff0000L) return -2;

    res2 <<= 16;
    res3 <<= 16;
    if ((res2 & res3) & 0x80000000L) return -3;

    res4 = res2 + res3;
    if (((res2 | res3) & 0x80000000L) && (!(res4 & 0x80000000L))) return -4;

    if ((res1 & res4) & 0x80000000L) return -5;
    res5 = res1 + res4;
    if (((res1 | res4) & 0x80000000L) && (!(res5 & 0x80000000L))) return -6;

    /*
     * Si se supone que el resultado ha de ser negativo, calcula el
     * complemento a dos de la doble palabra resultante.
     */

    if (negative) {
	if ((res5 & 0x80000000L) && (res5 & 0x7fffffffL)) return -7;
	else res5 = (~res5) + 1;
      }

    resPtr = (unsigned long int *) resultPtr;
    *resPtr = res5;
    return 0;
}

/* Funcion que calcula el maximo comun divisor de dos numeros*/
long int mcd(num1, num2)
 long int num1;
 long int num2;
 { long int i;
 for (i = min(num1, num2); i >0; i--)
 { if (((num1 % i) == 0) && ((num2 % i) == 0)) return i;}
 return TCL_OK;
 }

/*
 *----------------------------------------------------------------------
 *
 * FPissue --
 *
 *      Inicia una operacion que necesita una unidad de proceso.
 *
 * Resultados:
 *      Devuelve cero si la instruccion fue correctamente lanzada.
 *      Los siguientes problemas son detectados y resueltos:
 *           RAW    (operando fuente es destino de una operacion pendiente)
 *           WAW    (operando destino es destino de una operacion pendiente)
 *           estructurales (No hay disponible una unidad del tipo requerido).
 *      En cualquiera de estos casos, un entero es devuelto el cual indica
 *      cuantos ciclos de parada han de pasar antes de que el problema
 *      desaparezca. Si existen multiples problemas, el entero que se devuelve
 *      indica cuando se resuelve el primero de los problemas; otros problemas
 *      pueden estar todavia presentes.
 *
 * Efectos Laterales:
 *      En el caso de un lanzamiento satisfactorio, la lista de operaciones
 *      pendientes es actualizada. Despues del tiempo adecuado indicado por la
 *      latencia (mas el numero de elementos en el caso de operaciones
 *      vectoriales) FPWriteBack es invocado.
 *
 *----------------------------------------------------------------------
 */

static long int
FPissue (machPtr, type, resultType, source1, source2, dest, error)
     DLXV *machPtr;              /* Maquina simulada. */
     long int type;             /* Tipo de operacion. */
     long int resultType;       /* Tipo de resultado (ver dlx.h) */
     int source1, source2;      /* Numero de registro fuente. */
     int dest;                  /* numero de registro destino. */
     long int *error;           /* Se pone un valor no-cero si ocurre un error,
				 * 0 en cualquier otro caso. */
{
  register FPop *opPtr, *opPtr1, *opPtr2;
  long int i,j;
  long int soonest;
  long int wait;
  long int num_units;
  long int latency, latency2 = 0;
  long int unitType;
  long int direccion;
  long int maximo;
  long int result;
  long int stride;
  long int fuente1 = 0, fuente2 = 0;
  double *ds1, *ds2, *dd;
  float *fs1, *fs2, *fd;
  /* Estas variables se utilizan para calcular el nmero de ciclos que
  se necesitan para hacer un acceso a memoria con stride o indices */
  unsigned long int ndir;
  long int ciclo;
  unsigned long int modulo;
  long int mm_units[MAX_MM];  /* 0 indica unidad disponible, cualquier otro numero
				 indica el numero de ciclos tras los cuales la unidad
				 logica (de comparacion de vectores) completara la
				 operacion actual */


  opPtr = mallocFPop();
  opPtr->longitud = maximo = machPtr->regs[(int)VLR_REG];

  wait = 0;

  /* Detecta si hay un riesgo estructural motivado por el
  nmero limitado de puertos de lectura y escrituta de los
  registros vectoriales */

  /* VxV->V */
  if (resultType == FP_VECT_OP)
     {
     if (machPtr->rports[source1-32] >= machPtr->read_ports)
	{wait=machPtr->waitS[source1-32];}
     if (machPtr->rports[source2-32] >= machPtr->read_ports)
     	{wait=max(wait,machPtr->waitS[source2-32]);}
     if (machPtr->wports[dest-32] >= machPtr->write_ports)
	{wait=max(wait,machPtr->waitS[dest-32]);}
//     printf("FP_VECT %d %d \n",source1-32,source2-32);
     }
  /* S_V, SVI */
  if ((resultType == FP_SINDEX_OP) ||
     ((resultType >= FP_VEQ_OP) && (resultType <= FP_VLE_OP)))
     {            
     if (machPtr->rports[source1-32] >= machPtr->read_ports)
	{wait=machPtr->waitS[source1-32];}
     if (machPtr->rports[source2-32] >= machPtr->read_ports)
     	{wait=max(wait,machPtr->waitS[source2-32]);}
//     printf("S_V o SVI %d \n",resultType);
     }
  /* VxS -> V */
  if (resultType == FP_VS_OP)
     {
     if (machPtr->rports[source1-32] >= machPtr->read_ports)
     {wait = machPtr->waitS[source1-32];}
     if (machPtr->wports[dest-32] >= machPtr->write_ports)
	{wait=max(wait,machPtr->waitS[dest-32]);}
//     printf("xxVS %d \n",resultType);
     }
  /* SxV -> V, LVI */
  if ((resultType == FP_SV_OP) || (resultType == FP_LINDEX_OP))
     { 
     if (machPtr->rports[source2-32] >= machPtr->read_ports)
     {wait = machPtr->waitS[source2-32];}
     if (machPtr->wports[dest-32] >= machPtr->write_ports)
     {wait = max(wait,machPtr->waitS[dest-32]);}
//     printf("xxSV o LVI %d \n",resultType);
     }
  /* S_SV */
  if ((resultType >= FP_SVEQ_OP) && (resultType <= FP_SVLE_OP))
     { 
     if (machPtr->rports[source2-32] >= machPtr->read_ports)
     {wait = machPtr->waitS[source2-32];}
//     printf("S_SV %d \n",resultType);
     }
  /* SV, SVWS */
  if ((resultType == FP_STORE_OP) || (resultType == FP_SSTD_OP))
     {
     if (machPtr->rports[source1-32] >= machPtr->read_ports)
     {wait = machPtr->waitS[source1-32];}
//     printf("SV o SVWS %d \n",resultType);
     }
  /* LV */
  if ((resultType == FP_LOAD_OP) || (resultType == FP_LSTD_OP))
     {
     if (machPtr->wports[dest-32] >= machPtr->write_ports)
     {wait = machPtr->waitS[dest-32];}
//     printf("LV %d \n",resultType);
     }


  /* Detecta los problemas RAW y WAW */

  /* Caso de encadenamiento de operaciones vectoriales */
  /* Es una operacin vectorial encadenada? */
  /* OJO: No se puede encadenar si  LVI, LVWS, SVI o SVWS */
  if (encadenado && (resultType != FP_INT_OP) &&
    (resultType != FP_DOUBLE_FP_OP) && (resultType != FP_SINGLE_FP_OP) &&
    (resultType != FP_LSTD_OP) && (resultType != FP_SSTD_OP) &&
    (resultType != FP_LINDEX_OP) && (resultType != FP_SINDEX_OP))
    {
    /* LV no se puede encadenar, pero puede estar inmersa en un riesgo WAW */ 
      if (resultType == FP_LOAD_OP)
	 {
	 if ((machPtr->waitingE[dest] > machPtr->cycleCount) &&
	     (!wait || machPtr->waitingE[dest] < wait)){
	   wait = machPtr->waitingE[dest];
	   }
	 }
      /* No es LV */
      else {

  /*    if ((machPtr->waiting_FPRs[dest]) &&
	  (!wait || machPtr->waiting_FPRs[dest] < wait)){
	wait = machPtr->waiting_FPRs[dest];
	if (dest > 31) {wait += V_PEN_HZRD;}
	}
  */
	 if ((machPtr->waitingE[dest] > machPtr->cycleCount) &&
	     (!wait || machPtr->waitingE[dest] < wait)){
	   wait = machPtr->waitingE[dest];
	   }
	 if ((machPtr->waitingE[source1] > machPtr->cycleCount) &&
	    (!wait || machPtr->waitingE[source1] < wait)){
	   wait = machPtr->waitingE[source1];
	  }
	 if ((machPtr->waitingE[source2] > machPtr->cycleCount)  &&
	     (!wait || machPtr->waitingE[source2] < wait)){
	   wait = machPtr->waitingE[source2];
	   }
      }
    }
  /* No se encadena, o es una instruccion escalar */
  else {
//printf ("dest source1 source2 %d %d %d \n",dest,source1,source2);
  if ((machPtr->waiting_FPRs[dest])  &&
      (!wait || machPtr->waiting_FPRs[dest] < wait)){
    wait = machPtr->waiting_FPRs[dest];
    if (dest > 31) {wait += V_PEN_HZRD;}
    }
  if (machPtr->waiting_FPRs[source1] &&
      (!wait || machPtr->waiting_FPRs[source1] < wait)){
    wait = machPtr->waiting_FPRs[source1];
    if (source1 > 31) {wait += V_PEN_HZRD;}
    }
  if (machPtr->waiting_FPRs[source2] &&
      (!wait || machPtr->waiting_FPRs[source2] < wait)){
    wait = machPtr->waiting_FPRs[source2];
    if (source2 > 31) {wait += V_PEN_HZRD;}
    }
  }
//printf("FPIssue, tras encadenamiento \n");

  /* Los casos especiales:
    CVI le pasa el registro de uso general en error.
    Las operaciones de carga/almacenamiento vectorial
    pasan la direccin en error, y LVWS/SVWS el stride
    en la variable global strid */   
  if (resultType == FP_MASK_OP) {source1 = (int)*error;}
  if (type == V_LOAD) {direccion = *error;
		       stride = strid;}

  *error = 0;
//printf("type %d\n", type);

/* Si VLR=0, hay un error en el clculo de los ciclos, ya que
contabiliza el tiempo de evaluacin. */
  unitType = type;
  switch (type) {
  case FP_SUB:
    unitType = FP_ADD;
  case FP_ADD:            
    num_units = machPtr->num_add_units;
    latency = machPtr->fp_add_latency;
    break;
  case INT_DIV:
  case INT_UDIV:
    unitType = FP_DIV;
  case FP_DIV:
    num_units = machPtr->num_div_units;
    latency = machPtr->fp_div_latency;
    break;
  case INT_MUL:
  case INT_UMUL:
    unitType = FP_MUL;
  case FP_MUL:
    num_units = machPtr->num_mul_units;
    latency = machPtr->fp_mul_latency;
    break;
  case V_SUB:
    unitType = V_ADD;
  case V_ADD:
    num_units = machPtr->num_addv_units;
    latency = machPtr->v_add_latency + maximo;
    latency2 = machPtr->v_add_latency;
    break;
  case V_DIV:
    num_units = machPtr->num_divv_units;
    latency = machPtr->v_div_latency + maximo;
    latency2 = machPtr->v_div_latency;
    break;
  case V_MUL:
    num_units = machPtr->num_mulv_units;
    latency = machPtr->v_mul_latency + maximo;
    latency2 = machPtr->v_mul_latency;
    break;
  case V_LOGIC:
//  printf("V_LOGIC \n");
    num_units = machPtr->num_logic_units;
    latency = machPtr->v_logic_latency + maximo;
    latency2 = machPtr->v_logic_latency + maximo;
    break;
  case V_LOAD:
    num_units = machPtr->num_loadv_units;
    latency = machPtr->v_load_latency + maximo;
    latency2 = machPtr->v_load_latency;
    break;
  }
// printf("FPIssue, tras unidades \n");

  for (i = 0, soonest = MAXLONG; i < num_units; i++) {
    if (!(j = machPtr->fp_units[(int)unitType][(int)i])) break;
    else if (j < soonest) soonest = j;
  }
  if (i == num_units && (soonest < wait || !wait)) wait = soonest;
  if (wait) {   freeFPop(opPtr);
		return (wait - machPtr->cycleCount);
		}

  opPtr->unit = i;
  opPtr->dest = dest;
  opPtr->ready = machPtr->cycleCount + latency;
  opPtr->ready2 = machPtr->cycleCount + latency2;
  opPtr->latcarga = latency2;
  machPtr->fp_units[(int)unitType][(int)i] = opPtr->ready;
  machPtr->waiting_FPRs[dest] = opPtr->ready;

  if (encadenado && (resultType != FP_INT_OP) &&
    (resultType != FP_DOUBLE_FP_OP) && (resultType != FP_SINGLE_FP_OP) &&
    (resultType != FP_LSTD_OP) && (resultType != FP_SSTD_OP) &&
    (resultType != FP_LINDEX_OP) && (resultType != FP_SINDEX_OP) && (resultType != FP_LOAD_OP))
    {
/*  if (encadenado && (type != V_LOAD) && (resultType != FP_INT_OP) &&
  (resultType != FP_DOUBLE_FP_OP) && (resultType != FP_SINGLE_FP_OP))
  {
  */
  /* Caso de que exista encadenamiento de operaciones vectoriales */
  switch (resultType) {
   case FP_VS_OP:
   case FP_STORE_OP:  /* SV tambien se encadena */
	if ((fuente1 = machPtr->waitingE[source1]) != 0)
	{
	opPtr1 = machPtr->FPopsList;
	while ((opPtr1 != NULL) && (opPtr1->dest != source1))
		opPtr1 = opPtr1->nextPtr;
	}
	break;
   case FP_VECT_OP:
   case FP_VEQ_OP:
   case FP_VNE_OP:
   case FP_VGT_OP:
   case FP_VLT_OP:
   case FP_VGE_OP:
   case FP_VLE_OP:
	if ((fuente1 = machPtr->waitingE[source1]) != 0)
	{
	opPtr1 = machPtr->FPopsList;
	while ((opPtr1 != NULL) && (opPtr1->dest != source1))
		opPtr1 = opPtr1->nextPtr;
	}
   case FP_SV_OP:
   case FP_SVEQ_OP:
   case FP_SVNE_OP:
   case FP_SVGT_OP:
   case FP_SVLT_OP:
   case FP_SVGE_OP:
   case FP_SVLE_OP:
   /* Este cdigo es comun para VxV y SxV! */
	if ((fuente2 = machPtr->waitingE[source2]) != 0)
	{
	opPtr2 = machPtr->FPopsList;
	while ((opPtr2 != NULL) && (opPtr2->dest != source2))
		opPtr2 = opPtr2->nextPtr;
	}
  }
  }


  /* En las operaciones de Carga/Almacenamiento con stride o indices no es posible
  siempre obtener un dato de la memoria a cada ciclo de reloj.
  Este cdigo calcula cuantos ciclos seran necesarios. */

  switch (resultType) {
  case FP_LSTD_OP:    /* LVWS */
  case FP_SSTD_OP: {  /* SVWS */

   opPtr->direccion = direccion;
   opPtr->stride = stride;
   opPtr->k=0;
   if ((opPtr->stride <= 0)||((opPtr->stride % 8)!=0)) {BWCCMessageBox(GetFocus(), "Valor de stride incorrecto.\nSe utilizar stride unitario", "Error", MB_OK | MB_ICONEXCLAMATION);
			    opPtr->stride = 8;
			    }
   ndir = direccion;
   ciclo = 0;
   for (i = 0 ; i<machPtr->bancos ; i++){
     mm_units[(int)i]=0;
   }
   for (i = 0 ; i<maximo ; i++){
     modulo = (ndir >> 3) % machPtr->bancos;
//     printf("Ciclo %ld; Direccion %ld; Modulo %ld \n",ciclo,ndir,modulo);
     if (ciclo < mm_units[(int)modulo]) {
	ciclo = mm_units[(int)modulo];}
     mm_units [(int)modulo] = latency2 + ciclo; 
     ciclo ++;
     ndir = ndir + stride;
   }
   opPtr->ready = machPtr->cycleCount + ciclo + latency2;
//   printf("Ciclos: %ld \n",ciclo);
   if (maximo>0){ 
     opPtr->cada = ciclo / maximo;} 
   break;
   }     
       
/*   long int num;
   opPtr->direccion = direccion;
   opPtr->stride = stride;
   if ((opPtr->stride <= 0)||((opPtr->stride % 8)!=0)) {BWCCMessageBox(GetFocus(), "Valor de stride incorrecto.\nSe utilizar stride unitario", "Error", MB_OK | MB_ICONEXCLAMATION);
			    opPtr->stride = 8;
			    }
   opPtr->k = 0;
   num = machPtr->bancos / mcd(machPtr->bancos, (opPtr->stride/8));
   if (num > latency2) {opPtr->cada = maximo;}
   else {opPtr->cada = num;}
   num = maximo / opPtr->cada;
   if (num == maximo) {opPtr->ready = machPtr->cycleCount + num*latency2;}
   else {if ((num != 0) && (num != 1)) { opPtr->ready = machPtr->cycleCount + (num+1)*latency2 + maximo - num;}
	 else { opPtr->ready = machPtr->cycleCount + latency2 + maximo;}
	 }
   break;
   }
*/            
  case FP_LINDEX_OP:    /* LVI */
  case FP_SINDEX_OP: {  /* SVI */

   opPtr->direccion = direccion;
   opPtr->k = 0;
   ciclo = 0;
   for (i = 0 ; i<machPtr->bancos ; i++){
     mm_units[(int)i]=0;
   }
   for (i = 0 ; i<maximo ; i++){
     j = 2*i;
     indices[(int)i] = *(double *)&machPtr->vecregs[source2-32][(int)j] + direccion;
   }
   for (i = 0 ; i<maximo ; i++){
     ndir = indices[(int)i];
     modulo = (ndir >> 3) % machPtr->bancos;
//     printf("Ciclo %ld; Direccion %ld; Modulo %ld \n",ciclo,ndir,modulo);
     if (ciclo < mm_units[(int)modulo]) {
	ciclo = mm_units[(int)modulo];}
     mm_units [(int)modulo] = latency2 + ciclo; 
     ciclo ++;
   }
   opPtr->ready = machPtr->cycleCount + ciclo + latency2;
   if (maximo>0){
     opPtr->cada = ciclo / maximo;} 
//   printf("Ciclos: %ld \n",ciclo);
   break;
   }
 }                                       
/*   long int num;
   opPtr->direccion = direccion;
   opPtr->k = 0;
   if (resultType == FP_SINDEX_OP) {
   opPtr->stride = (*(double *)&(machPtr->vecregs[source1-32][2]) - *(double *)&(machPtr->vecregs[source1-32][0]))/8;}
   else { opPtr->stride = (*(double *)&(machPtr->vecregs[source2-32][2]) - *(double *)&(machPtr->vecregs[source2-32][0]))/8;}
   if (opPtr->stride <= 0) {BWCCMessageBox(GetFocus(), "Valor de stride incorrecto.\nSe utilizar stride unitario", "Error", MB_OK | MB_ICONEXCLAMATION);
			    opPtr->stride = 8;
			    }
   num = machPtr->bancos / mcd(machPtr->bancos, opPtr->stride);
   if (num > latency2) {opPtr->cada = maximo;}
   else {opPtr->cada = num;}
   num = maximo / opPtr->cada;
   if (num == maximo) {opPtr->ready = machPtr->cycleCount + num*latency2;}
   else {if ((num != 0) && (num != 1)) { opPtr->ready = machPtr->cycleCount + (num+1)*latency2 + maximo - num;}
	 else { opPtr->ready = machPtr->cycleCount  + latency2 + maximo;}
	 }
   break;
   }
  }  */

  /* Hacemos la operacin: */
  switch (resultType) {

  case FP_SINGLE_FP_OP:
    fs1 = (float *)&(machPtr->regs[source1 + 32]);
    fs2 = (float *)&(machPtr->regs[source2 + 32]);
    fd = (float *)&(opPtr->result[0]);
    switch (type) {
    case FP_ADD: *fd = *fs1 + *fs2; break;
    case FP_DIV:{
       if (*fs1 == 0) {
         /* division por 0 */
	 *fd = 0;
	 *error = 1;
       } else
	 *fd = *fs1 / *fs2;
       break;
    }
    case FP_MUL: *fd = *fs1 * *fs2; break;
    case FP_SUB: *fd = *fs1 - *fs2; break;
    }
    break;

  case FP_DOUBLE_FP_OP:
    ds1 = (double *)&(machPtr->regs[source1 + 32]);
    ds2 = (double *)&(machPtr->regs[source2 + 32]);
    dd = (double *)&(opPtr->result[0]);
    switch (type) {
    case FP_ADD: *dd = *ds1 + *ds2; break;
    case FP_DIV: {
       if (*ds1 == 0) {
         /* division por 0 */
	 *dd = 0;
	 *error = 1;
       } else
	 *dd = *ds1 / *ds2;
       break;
    }
    case FP_MUL: *dd = *ds1 * *ds2; break;
    case FP_SUB: *dd = *ds1 - *ds2; break;
    }
    machPtr->waiting_FPRs[dest+1] = opPtr->ready;
    break;

  case FP_VECT_OP: /* VxV -> V */
   opPtr->ultimo = 0;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->f1=source1-32;opPtr->f2=source2-32;
   opPtr->d=dest-32;

   if (machPtr->rports[opPtr->f1])
    {machPtr->waitS[opPtr->f1] = min(machPtr->waitS[opPtr->f1], opPtr->ready);}
    else {machPtr->waitS[opPtr->f1] = opPtr->ready;}
   if (machPtr->rports[opPtr->f2])
    {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}
   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}

   machPtr->rports[opPtr->f1]++;
   machPtr->rports[opPtr->f2]++;
   machPtr->wports[opPtr->d]++;
  

   for (i = 0; i < maximo; i++)
   {
    j = i*2;
    if (fuente1) { ds1 = (double *)&(opPtr1->result[(int)j]);}
    else {ds1 = (double *)&(machPtr->vecregs[source1 - 32][(int)j]);}
    if (fuente2) { ds2 = (double *)&(opPtr2->result[(int)j]);}
    else {ds2 = (double *)&(machPtr->vecregs[source2 - 32][(int)j]);}
    dd = (double *)&(opPtr->result[(int)j]);
    switch (type) {
    case V_ADD: *dd = *ds1 + *ds2; break;
    case V_DIV: {
       if (*ds1 == 0) {
         /* division por 0 */
	 *dd = 0;
	 *error = 1;
       } else
	 *dd = *ds1 / *ds2;
       break;
    }
    case V_MUL: *dd = *ds1 * *ds2; break;
    case V_SUB: *dd = *ds1 - *ds2; break;
    }
    }
    break;

  case FP_SV_OP:  /* S x V -> V */
   opPtr->ultimo = 0;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->f2=source2-32;
   opPtr->d=dest-32;

   if (machPtr->rports[opPtr->f2])
    {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}
   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}

   machPtr->rports[opPtr->f2]++;
   machPtr->wports[opPtr->d]++;

   ds1 = (double *)&(machPtr->regs[source1 + 32]);
   for (i = 0; i < maximo; i++)
   {
    j = i*2;
    if (fuente2) { ds2 = (double *)&(opPtr2->result[(int)j]);}
    else {ds2 = (double *)&(machPtr->vecregs[source2 - 32][(int)j]);}
    dd = (double *)&(opPtr->result[(int)j]);
    switch (type) {
    case V_ADD: *dd = *ds1 + *ds2; break;
    case V_DIV: {
       if (*ds1 == 0) {
         /* division por 0 */
	 *dd = 0;
	 *error = 1;
       } else
	 *dd = *ds1 / *ds2;
       break;
    }
    case V_MUL: *dd = *ds1 * *ds2; break;
    case V_SUB: *dd = *ds1 - *ds2; break;
    }
    }
    break;

  case FP_VS_OP: /* V x S -> V */
   opPtr->ultimo = 0;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->f1=source1-32;
   opPtr->d=dest-32;

   if (machPtr->rports[opPtr->f1])
    {machPtr->waitS[opPtr->f1] = min(machPtr->waitS[opPtr->f1], opPtr->ready);}
    else {machPtr->waitS[opPtr->f1] = opPtr->ready;}
   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}

   machPtr->rports[opPtr->f1]++;
   machPtr->wports[opPtr->d]++;

   ds2 = (double *)&(machPtr->regs[source2 + 32]);
   for (i = 0; i < maximo; i++)
   {
    j = i*2;
    if (fuente1) { ds1 = (double *)&(opPtr1->result[(int)j]);}
    else {ds1 = (double *)&(machPtr->vecregs[source1 - 32][(int)j]);}
    dd = (double *)&(opPtr->result[(int)j]);
    switch (type) {
    case V_ADD: *dd = *ds1 + *ds2; break;
    case V_DIV: {
       if (*ds1 == 0) {
         /* division por 0 */
	 *dd = 0;
	 *error = 1;
       } else
	 *dd = *ds1 / *ds2;
       break;
    }
    case V_MUL: *dd = *ds1 * *ds2; break;
    case V_SUB: *dd = *ds1 - *ds2; break;
    }
    }
    break;

  case FP_VEQ_OP: /* S_V */
  case FP_VNE_OP:
  case FP_VGT_OP:
  case FP_VLT_OP:
  case FP_VGE_OP:
  case FP_VLE_OP:
   opPtr->ultimo = 0;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->f1=source1-32;opPtr->f2=source2-32;
//   opPtr->d=dest-32;

   if (machPtr->rports[opPtr->f1])
    {machPtr->waitS[opPtr->f1] = min(machPtr->waitS[opPtr->f1], opPtr->ready);}
    else {machPtr->waitS[opPtr->f1] = opPtr->ready;}
   if (machPtr->rports[opPtr->f2])
    {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}

   machPtr->rports[opPtr->f1]++;
   machPtr->rports[opPtr->f2]++;

   for (i = 0; i < maximo; i++)
   {
    j = i*2;
    if (fuente1) { ds1 = (double *)&(opPtr1->result[(int)j]);}
    else {ds1 = (double *)&(machPtr->vecregs[source1 - 32][(int)j]);}
    if (fuente2) { ds2 = (double *)&(opPtr2->result[(int)j]);}
    else {ds2 = (double *)&(machPtr->vecregs[source2 - 32][(int)j]);}
    dd = (double *)&(opPtr->result[(int)j]);
    switch (resultType) {
    case FP_VEQ_OP: *dd = (*ds1 == *ds2);break;
    case FP_VNE_OP: *dd = (*ds1 != *ds2);break;
    case FP_VGT_OP: *dd = (*ds1 > *ds2);break;
    case FP_VLT_OP: *dd = (*ds1 < *ds2);break;
    case FP_VGE_OP: *dd = (*ds1 >= *ds2);break;
    case FP_VLE_OP: *dd = (*ds1 <= *ds2);break;
    }
    }
    break;

  case FP_SVEQ_OP: /* S_SV */
  case FP_SVNE_OP:
  case FP_SVGT_OP:
  case FP_SVLT_OP:
  case FP_SVGE_OP:
  case FP_SVLE_OP:

//     printf("S_SV ejecucion \n");
   opPtr->ultimo = 0;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->f2=source2-32;
//   opPtr->d=dest-32;

   if (machPtr->rports[opPtr->f2])
    {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}

   machPtr->rports[opPtr->f2]++;

   ds1 = (double *)&(machPtr->regs[source1 + 32]);
   for (i = 0; i < maximo; i++)
   {
    j = i*2;
    if (fuente2) { ds2 = (double *)&(opPtr2->result[(int)j]);}
    else {ds2 = (double *)&(machPtr->vecregs[source2 - 32][(int)j]);}
    dd = (double *)&(opPtr->result[(int)j]);
    switch (resultType) {
    case FP_SVEQ_OP: *dd = (*ds1 == *ds2);break;
    case FP_SVNE_OP: *dd = (*ds1 != *ds2);break;
    case FP_SVGT_OP: *dd = (*ds1 > *ds2);break;
    case FP_SVLT_OP: *dd = (*ds1 < *ds2);break;
    case FP_SVGE_OP: *dd = (*ds1 >= *ds2);break;
    case FP_SVLE_OP: *dd = (*ds1 <= *ds2);break;
    }
    }
    break;

  case FP_MASK_OP: /* CVI */
  {
   unsigned long int num1, num2, auxi;

   opPtr->ultimo = 0;
   machPtr->waitingE[dest] = opPtr->ready2;
   opPtr->longitud = maximo = MVL;
   latency = machPtr->v_mul_latency + maximo;
   opPtr->ready = machPtr->cycleCount + latency;
   machPtr->fp_units[(int)unitType][(int)i] = opPtr->ready;
   machPtr->waiting_FPRs[dest] = opPtr->ready;

   opPtr->d=dest-32;

   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}

   machPtr->wports[opPtr->d]++;

   num2 = (unsigned long int)(machPtr->regs[(int)(VLR_REG + 1)]);
   num1 = (unsigned long int)(machPtr->regs[(int)(VLR_REG + 2)]);
   i = 0;auxi = 0;
   while (num1 != 0)
	       {
		 if ((num1 % 2) == 1) {
			j = i*2;
			dd = (double *)&(opPtr->result[(int)j]);
			*dd = machPtr->regs[source1] * auxi;
			i++;
			}
		 auxi++;
		 num1 /= 2;
	       }
   auxi = 32;
   while (num2 != 0)
	       {
		 if ((num2 % 2) == 1) {
                 	j = i*2;
		 	dd = (double *)&(opPtr->result[(int)j]);
			*dd = machPtr->regs[source1] * auxi;
			i++;
			}
		 auxi++;
		 num2 /= 2;
	       }
  for (auxi = i; auxi < MVL; auxi++)
  {
   j = auxi*2;
   dd = (double *)&(opPtr->result[(int)j]);
   *dd = 0;
  }

  }
   break;

  case FP_LOAD_OP:  /* LV */
   opPtr->ultimo = 0;
   opPtr->cada = direccion;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->d=dest-32;

   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}

   machPtr->wports[opPtr->d]++;


   for (i = 0; i < maximo; i++)
   {
    j=i*2;
    result = ReadVMem(machPtr, direccion, &opPtr->result[(int)j]);
    if (result == 0) { *error = 1; break;}
    direccion += 4;
    result = ReadVMem(machPtr, direccion, &opPtr->result[(int)(j+1)]);
    if (result == 0) { *error = 1; break;}
    direccion += 4;
   }
   break;

  case FP_STORE_OP: /* SV */
    opPtr->ultimo = 0;
    machPtr->waiting_FPRs[dest] = 0;
    machPtr->waitingE[dest] = 0;

    opPtr->f1=source1-32;

    if (machPtr->rports[opPtr->f1])
    {machPtr->waitS[opPtr->f1] = min(machPtr->waitS[opPtr->f1], opPtr->ready);}
    else {machPtr->waitS[opPtr->f1] = opPtr->ready;}

    machPtr->rports[opPtr->f1]++;

    opPtr->dest = direccion;
    for (i = 0; i < maximo; i++)
    {
     j = i*2;
     if (fuente1) { ds1 = (double *)&(opPtr1->result[(int)j]);}
     else {ds1 = (double *)&(machPtr->vecregs[source1 - 32][(int)j]);}
/*     ds1 = (double *)&(machPtr->vecregs[source1 - 32][(int)j]);  */
     dd = (double *)&(opPtr->result[(int)j]);
     *dd = *ds1;
     }
    break;

  case FP_LSTD_OP:   /* LVWS */
   opPtr->ultimo = 0;
   machPtr->fp_units[(int)unitType][(int)opPtr->unit] = opPtr->ready;
   machPtr->waiting_FPRs[dest] = opPtr->ready;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->d=dest-32;

   machPtr->wports[opPtr->d]++;

   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}


   for (i = 0; i < maximo; i++)
   {
    j=i*2;
// Si el stride va en palabras:
//    direccion = opPtr->direccion + opPtr->stride*i*8;
// Si el stride va en bytes:
    direccion = opPtr->direccion + opPtr->stride*i;
    result = ReadVMem(machPtr, direccion, &opPtr->result[(int)j]);
    if (result == 0) { *error = 1; break;}
    direccion += 4;
    result = ReadVMem(machPtr, direccion, &opPtr->result[(int)(j+1)]);
    if (result == 0) { *error = 1; break;}
   }
   break;

  case FP_SSTD_OP: /* SVWS */
    opPtr->ultimo = 0;
    machPtr->fp_units[(int)unitType][(int)opPtr->unit] = opPtr->ready;
    machPtr->waiting_FPRs[dest] = 0;
    machPtr->waitingE[dest] = 0;

    opPtr->f2=source2-32;

    if (machPtr->rports[opPtr->f2])
     {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}

    machPtr->rports[opPtr->f2]++;

    opPtr->dest = -1;

    for (i = 0; i < maximo; i++)
    {
     j = i*2;
     ds1 = (double *)&(machPtr->vecregs[source2 - 32][(int)j]);
     dd = (double *)&(opPtr->result[(int)j]);
     *dd = *ds1;
     }
    break;

  case FP_LINDEX_OP: /* LVI */
   opPtr->ultimo = 0;
   machPtr->fp_units[(int)unitType][(int)opPtr->unit] = opPtr->ready;
   machPtr->waiting_FPRs[dest] = opPtr->ready;
   machPtr->waitingE[dest] = opPtr->ready2;

   opPtr->f2=source2-32;
   opPtr->d=dest-32;

   if (machPtr->rports[opPtr->f2])
    {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}
   if (machPtr->wports[opPtr->d])
    {machPtr->waitS[opPtr->d] = min(machPtr->waitS[opPtr->d], opPtr->ready);}
    else {machPtr->waitS[opPtr->d] = opPtr->ready;}

   machPtr->rports[opPtr->f2]++;
   machPtr->wports[opPtr->d]++;

   for (i = 0; i < maximo; i++)
   {
    j=i*2;
    indices[(int)i] = *(double *)&machPtr->vecregs[source2-32][(int)j] + direccion;
    result = ReadVMem(machPtr, indices[(int)i], &opPtr->result[(int)j]);
    if (result == 0) { *error = 1; break;}
    result = ReadVMem(machPtr, indices[(int)i]+4, &opPtr->result[(int)(j+1)]);
    if (result == 0) { *error = 1; break;}
   }
   break;

  case FP_SINDEX_OP: /* SVI */
    opPtr->ultimo = 0;
    opPtr->dest = -1;
    machPtr->waiting_FPRs[dest] = 0;
    machPtr->waitingE[dest] = 0;
    machPtr->fp_units[(int)unitType][(int)opPtr->unit] = opPtr->ready;

    opPtr->f1=source1-32;opPtr->f2=source2-32;

    if (machPtr->rports[opPtr->f1])
    {machPtr->waitS[opPtr->f1] = min(machPtr->waitS[opPtr->f1], opPtr->ready);}
    else {machPtr->waitS[opPtr->f1] = opPtr->ready;}
    if (machPtr->rports[opPtr->f2])
     {machPtr->waitS[opPtr->f2] = min(machPtr->waitS[opPtr->f2], opPtr->ready);}
    else {machPtr->waitS[opPtr->f2] = opPtr->ready;}

    machPtr->rports[opPtr->f1]++;
    machPtr->rports[opPtr->f2]++;

    for (i = 0; i < maximo; i++)
    {
     j = i*2;
     ds2 = (double *)&(machPtr->vecregs[source1 - 32][(int)j]);
     indices[(int)i] = *ds2 + direccion;
     ds1 = (double *)&(machPtr->vecregs[source2 - 32][(int)j]);
     dd = (double *)&(opPtr->result[(int)j]);
     *dd = *ds1;

     }
    break;

  case FP_INT_OP:
    switch (type) {
    case INT_MUL:
      *error = Mult (machPtr->regs[source1 + 32],
		     machPtr->regs[source2 + 32], uno,
		     (long int *)&(opPtr->result[0]));
      break;
    case INT_UMUL:
      *error = Mult (machPtr->regs[source1 + 32],
		     machPtr->regs[source2 + 32], cero,
		     (long int *)&(opPtr->result[0]));
      break;
    case INT_DIV:
      if (machPtr->regs[source2 + 32] == 0) {
	/* divisin por 0 */
	*((long int *)&(opPtr->result[0])) = 0;
	*error = 1;
      } else
	*((long int *)&(opPtr->result[0])) = machPtr->regs[source1 + 32] /
	  machPtr->regs[source2 + 32];
      break;
    case INT_UDIV:
      {
	unsigned long int rs1, rs2, tmp;

	rs1 = machPtr->regs[source1 + 32];
	rs2 = machPtr->regs[source2 + 32];
	if (rs2 == 0) {
	  /* divisin por 0 */
	  *((unsigned long int *)&(opPtr->result[0])) = 0;
	  *error = 1;
	} else {
	  tmp = rs1 / rs2;
	  *((unsigned long int *)&(opPtr->result[0])) = tmp;
	}
	break;
      }
    }
    break;
  }


//      printf("Operandos: V%d y V%d; Resultado: V%d \n",opPtr->f1,opPtr->f2,opPtr->d);      printf("Puertos:\n");
/*      { int i;
      for (i=0; i<5; i++){
      printf("Registro %d: %d lectura/%descritura\n)",
	     i,machPtr->rports[i], machPtr->wports[i]);}
      }
*/

  opPtr->type = type;
  opPtr->resultType = resultType;

  if (machPtr->FPopsList == NULL ||
      opPtr->ready < machPtr->FPopsList->ready) {
    opPtr->nextPtr = machPtr->FPopsList;
    machPtr->FPopsList = opPtr;
  } else {
    register FPop *opPtr2;
    for (opPtr2 = machPtr->FPopsList; opPtr2->nextPtr != NULL;
	 opPtr2 = opPtr2->nextPtr)
      if (opPtr->ready < opPtr2->nextPtr->ready) break;
    opPtr->nextPtr = opPtr2->nextPtr;
    opPtr2->nextPtr = opPtr;
  }

  if (!machPtr->checkFP || (opPtr->ready < machPtr->checkFP))
    machPtr->checkFP = opPtr->ready;
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * FPwriteBack --
 *
 *      Completa una operacion pendiente, y escribe el resultado en el destino
 *      correspondiente.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Todas las operaciones que han sido completadas en el ciclo actual
 *      escriben sus resultados en el destino adecuado.
 *      Todas las estructuras asociadas con las operaciones pendientes
 *      son actualizadas adecuadamente.
 *
 *----------------------------------------------------------------------
 */

static void
FPwriteBack (machPtr)
     DLXV *machPtr;              /* Descripcion de la maquina. */
{
  register FPop *opPtr;
  long int i,j,k, direccion;

  while (machPtr->FPopsList != NULL &&
	 machPtr->FPopsList->ready <= machPtr->cycleCount) {
    opPtr = machPtr->FPopsList;
    machPtr->FPopsList = opPtr->nextPtr;
    if (machPtr->FPopsList != NULL)
      machPtr->checkFP = machPtr->FPopsList->ready;
    else machPtr->checkFP = 0;

    switch (opPtr->type) {
    case FP_ADD:
    case FP_SUB: machPtr->fp_add_units[(int)opPtr->unit] = 0; break;
    case INT_DIV:
    case INT_UDIV:
    case FP_DIV: machPtr->fp_div_units[(int)opPtr->unit] = 0; break;
    case INT_MUL:
    case INT_UMUL:
    case FP_MUL: machPtr->fp_mul_units[(int)opPtr->unit] = 0; break;
    case V_SUB:
    case V_ADD: machPtr->v_add_units[(int)opPtr->unit] = 0; break;
    case V_MUL: machPtr->v_mul_units[(int)opPtr->unit] = 0; break;
    case V_DIV: machPtr->v_div_units[(int)opPtr->unit] = 0; break;
    case V_LOGIC: {
	long int ind, tempo, hi = 2;
	machPtr->regs[(int)(VLR_REG + 1)] = 0;
	machPtr->regs[(int)(VLR_REG + 2)] = 0;
	for (i = opPtr->ultimo; i < opPtr->longitud; i++){
	j = i*2;
	if (i > 31) {hi = 1;}
	ind = (i % 32);
	for (k=0,tempo=1;k<ind;k++) { tempo = 2 * tempo;}
	if (*(double *)&(opPtr->result[(int)j]) == 1) {
	machPtr->regs[(int)(VLR_REG + hi)] += tempo;
	}
	}
	machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
	machPtr->waitingE[(int)opPtr->dest] = 0;
	machPtr->v_logic_units[(int)opPtr->unit] = 0;
	break;
	}
    case V_LOAD: machPtr->v_load_units[(int)opPtr->unit] = 0;
    }


    /* Que operacion es? */
    switch (opPtr->resultType) {

    case FP_VEQ_OP: /* S_V */
    case FP_VNE_OP:
    case FP_VGT_OP:
    case FP_VLT_OP:
    case FP_VGE_OP:
    case FP_VLE_OP:
	machPtr->rports[opPtr->f1]--;
    case FP_SVEQ_OP: /* S_SV */
    case FP_SVNE_OP:
    case FP_SVGT_OP:
    case FP_SVLT_OP:
    case FP_SVGE_OP:
    case FP_SVLE_OP:{
	/* Este codigo es comun a S_V y S_SV */
	machPtr->rports[opPtr->f2]--;
	break;
    }
    case FP_INT_OP: {
      long int *source, *dest;
      source = (long int *)&(opPtr->result[0]);
      dest = (long int *)&(machPtr->regs[(int)(32 + opPtr->dest)]);
      *dest = *source;
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }
    case FP_SINGLE_FP_OP: {
      float *source, *dest;
      source = (float *)&(opPtr->result[0]);
      dest = (float *)&(machPtr->regs[(int)(32 + opPtr->dest)]);
      *dest = *source;
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }
    case FP_DOUBLE_FP_OP: {
      double *source, *dest;
      source = (double *)&(opPtr->result[0]);
      dest = (double *)&(machPtr->regs[(int)(32 + opPtr->dest)]);
      *dest = *source;
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waiting_FPRs[(int)(opPtr->dest + 1)] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)(opPtr->dest + 1)] = 0;
      break;
    }
    case FP_VECT_OP: /* V x V -> V*/
	 machPtr->rports[opPtr->f1]--;

    case FP_SV_OP:  /* S x V -> V */
    /* este codigo es comun a ambos casos */
    	 machPtr->rports[opPtr->f2]--;

    case FP_MASK_OP:{ /* CVI */
      double *source, *dest;
    /* este codigo es comun a los tres casos anteriores */

      machPtr->wports[opPtr->d]--;

      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      source = (double *)&(opPtr->result[(int)j]);
      dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
      *dest = *source;
      }
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }
    case FP_VS_OP:{ /* VxS -> V */
      double *source, *dest;
      machPtr->rports[opPtr->f1]--;
      machPtr->wports[opPtr->d]--;

      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      source = (double *)&(opPtr->result[(int)j]);
      dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
      *dest = *source;
      }
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }
    case FP_LOAD_OP: { /* LV */
      double *source, *dest;
 
      machPtr->wports[opPtr->d]--;

      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      source = (double *)&(opPtr->result[(int)j]);
      dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
      *dest = *source;
      if (machPtr->refTraceFile) {
	  fprintf( machPtr->refTraceFile, "0 %lx\n", opPtr->cada);
	  opPtr->cada += 4;
	  fprintf( machPtr->refTraceFile, "0 %lx\n", opPtr->cada);
	  opPtr->cada += 4;
	  }
      }
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }

    case FP_LSTD_OP: { /* LVWS */
      double *source, *dest;

      machPtr->wports[opPtr->d]--;

      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      source = (double *)&(opPtr->result[(int)j]);
      dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
      *dest = *source;
      if (machPtr->refTraceFile) {
// Si el stride va en palabras
//      	  direccion = opPtr->direccion + opPtr->stride*i*8;
// Si el stride va en bytes
 	  direccion = opPtr->direccion + opPtr->stride*i;
	  fprintf( machPtr->refTraceFile, "0 %lx\n", direccion);
	  direccion += 4;
	  fprintf( machPtr->refTraceFile, "0 %lx\n", direccion);
	  }
      }
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }

    case FP_STORE_OP: { /* SV */
      machPtr->rports[opPtr->f1]--;
      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      direccion = opPtr->dest + i*8;
      if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)j]) == 0)
			   break;
      direccion += 4;
      if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)(j+1)]) == 0)
			   break;
      }
      break;
    }

    case FP_SSTD_OP: { /* SVWS */
      machPtr->rports[opPtr->f2]--;
      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
// Si el stride va en palabras
//      	  direccion = opPtr->direccion + opPtr->stride*i*8;
// Si el stride va en bytes
      direccion = opPtr->direccion + opPtr->stride*i;
      if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)j]) == 0)
			   break;
      direccion += 4;
      if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)(j+1)]) == 0)
			   break;
      }
      break;
    }

    case FP_LINDEX_OP: { /* LVI */
      double *source, *dest;
      machPtr->rports[opPtr->f2]--;
      machPtr->wports[opPtr->d]--;

      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      source = (double *)&(opPtr->result[(int)j]);
      dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
      *dest = *source;
      if (machPtr->refTraceFile) {
	  fprintf( machPtr->refTraceFile, "0 %lx\n", indices[(int)i]);
	  fprintf( machPtr->refTraceFile, "0 %lx\n", indices[(int)i]+4);
	  }
      }
      machPtr->waiting_FPRs[(int)opPtr->dest] = 0;
      machPtr->waitingE[(int)opPtr->dest] = 0;
      break;
    }
    case FP_SINDEX_OP: { /* SVI */
      machPtr->rports[opPtr->f1]--;
      machPtr->rports[opPtr->f2]--;
      for (i = opPtr->ultimo; i < opPtr->longitud; i++){
      j = i*2;
      if (WriteVMem(machPtr, (unsigned long int) (indices[(int)i]),
			   cuatro, opPtr->result[(int)j]) == 0)
			   break;
      if (WriteVMem(machPtr, (unsigned long int) (indices[(int)i]+4),
			   cuatro, opPtr->result[(int)(j+1)]) == 0)
			   break;
      }
      break;
    }


    } /* switch */
    freeFPop (opPtr);
  }
}

/*
 *----------------------------------------------------------------------
 *
 * Vwrite --
 *
 *   Lanza, en cada ciclo de reloj los datos de aquellas operaciones
 *   vectoriales que hayan cumplido su penalizacion de comienzo.
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      Todas las operaciones vectoriales pendientes realizan la
 *      escritura de resultados en los registros correspondientes.
 *
 * Nota: Este procedimiento se ha de ejecutar ya que todas las operaciones
 * vectoriales, una vez superada la penalizacion de arranque lanzan un
 * resultado por cada ciclo de reloj.
 *
 *
 *----------------------------------------------------------------------
 */

static void
Vwrite (machPtr)
     DLXV *machPtr;
{
  long int i, j, diferencia, direccion;
  register FPop *opPtr = machPtr->FPopsList;

  while (opPtr != NULL) {
    if ((opPtr->type >= V_ADD) && (opPtr->type <= V_SUB) &&
    (opPtr->ready2 < machPtr->cycleCount))
    {
    diferencia = (machPtr->cycleCount - opPtr->ready2);
    switch (opPtr->type) {
    case V_ADD:
    case V_SUB:
    case V_DIV:
    case V_MUL: {
      double *source, *dest;
      for (i = opPtr->ultimo; ((i < opPtr->longitud) && (i < diferencia)); i++){
      j = i*2;
      source = (double *)&(opPtr->result[(int)j]);
      dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
      *dest = *source;
      }
      opPtr->ultimo = i;
      break;
    }
    case V_LOGIC: break;
    case V_LOAD: {
     switch (opPtr->resultType) {
       case FP_LOAD_OP: {
	double *source, *dest;
	for (i = opPtr->ultimo; ((i < opPtr->longitud) && (i < diferencia)); i++){
	 j = i*2;
	 source = (double *)&(opPtr->result[(int)j]);
	 dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
	 *dest = *source;
	 if (machPtr->refTraceFile) {
	  fprintf( machPtr->refTraceFile, "0 %lx\n", opPtr->cada);
	  opPtr->cada += 4;
	  fprintf( machPtr->refTraceFile, "0 %lx\n", opPtr->cada);
	  opPtr->cada += 4;
	 }
	}
	opPtr->ultimo = i;
	break;
	}
	case FP_STORE_OP: {
	for (i = opPtr->ultimo; ((i < opPtr->longitud) && (i < diferencia)); i++){
	 j = i*2;
	 direccion = opPtr->dest + i*8;
	 if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)j]) == 0)
			   break;
	 direccion += 4;
	 if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)(j+1)]) == 0)
			   break;
	}
	opPtr->ultimo = i;
	break;
	}

	case FP_LSTD_OP: {
	double *source, *dest;
	for (i = opPtr->ultimo; ((i < opPtr->longitud) && (opPtr->k < (machPtr->cycleCount - opPtr->ready2))); i++){
	 j = i*2;
// Si el stride va en palabras:
//	 direccion = opPtr->direccion + opPtr->stride*i*8;
// Si el stride va en bytes:
	 direccion = opPtr->direccion + opPtr->stride*i;
	 source = (double *)&(opPtr->result[(int)j]);
	 dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
	 *dest = *source;
	 if (machPtr->refTraceFile) {
	  fprintf( machPtr->refTraceFile, "0 %lx\n", direccion);
	  direccion  += 4;
	  fprintf( machPtr->refTraceFile, "0 %lx\n", direccion);
	 }
	 if ((((i+1) % opPtr->cada) == 0) && (i != (opPtr->longitud-1))) { opPtr->ready2 = machPtr->cycleCount - 1 + opPtr->latcarga;
									   opPtr->k = -1;}
	 opPtr->ultimo = i+1;
	 opPtr->k++;
	}
	break;
	}

	case FP_SSTD_OP: {
	for (i = opPtr->ultimo; ((i < opPtr->longitud) && (opPtr->k < (machPtr->cycleCount - opPtr->ready2))); i++){
	 j = i*2;
// Si el stride va en palabras:
//	 direccion = opPtr->direccion + opPtr->stride*i*8;
// Si el stride va en bytes:
	 direccion = opPtr->direccion + opPtr->stride*i;

	 if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)j]) == 0)
			   break;
	 direccion += 4;
	 if (WriteVMem(machPtr, (unsigned long int) (direccion),
			   cuatro, opPtr->result[(int)(j+1)]) == 0)
			   break;
	 if ((((i+1) % opPtr->cada) == 0) && (i != (opPtr->longitud-1))) { opPtr->ready2 = machPtr->cycleCount - 1 + opPtr->latcarga;
									   opPtr->k = -1;}
	 opPtr->ultimo = i+1;
	 opPtr->k++;
	}
	break;
	}

	case FP_LINDEX_OP: {
	double *source, *dest;
	for (i = opPtr->ultimo; ((i < opPtr->longitud) && (opPtr->k < (machPtr->cycleCount - opPtr->ready2))); i++){
	 j = i*2;
	 source = (double *)&(opPtr->result[(int)j]);
	 dest = (double *)&(machPtr->vecregs[(int)(opPtr->dest-32)][(int)j]);
	 *dest = *source;
	 if (machPtr->refTraceFile) {
	  fprintf( machPtr->refTraceFile, "0 %lx\n", indices[(int)i]);
	  fprintf( machPtr->refTraceFile, "0 %lx\n", indices[(int)i]+4);
	 }
	 if ((((i+1) % opPtr->cada) == 0) && (i != (opPtr->longitud-1))) { opPtr->ready2 = machPtr->cycleCount - 1 + opPtr->latcarga;
									   opPtr->k = -1;}
	 opPtr->ultimo = i+1;
	 opPtr->k++;
	}
	break;
	}

	case FP_SINDEX_OP: {
	for (i = opPtr->ultimo; ((i < opPtr->longitud) && (opPtr->k < (machPtr->cycleCount - opPtr->ready2))); i++){
	 j = i*2;
	 if (WriteVMem(machPtr, (unsigned long int) (indices[(int)i]),
			   cuatro, opPtr->result[(int)j]) == 0)
			   break;
	 if (WriteVMem(machPtr, (unsigned long int) (indices[(int)i]+4),
			   cuatro, opPtr->result[(int)(j+1)]) == 0)
			   break;
	 if ((((i+1) % opPtr->cada) == 0) && (i != (opPtr->longitud-1))) { opPtr->ready2 = machPtr->cycleCount - 1 + opPtr->latcarga;
									   opPtr->k = -1;}
	 opPtr->ultimo = i+1;
	 opPtr->k++;
	}
	break;
	}
       }
       break;
      }
    } /* switch */
    }
    opPtr = opPtr->nextPtr;
  }
}

/*
 *----------------------------------------------------------------------
 *
 * FPwait --
 *
 *      Comprueba hasta cuatro registros para ver si son destino de
 *      operaciones pendientes. Es usado para comprobar los problemas
 *      de RAW y WAW ( por ej en la instruccion MOVD).
 *
 * Resultados:
 *      Ninguno.
 *
 * Efectos Laterales:
 *      La cuenta de ciclos de parada por dependencia entre operaciones
 *      es incrementada con la cantidad adecuada de manera que los
 *      registros especificados reciban el resultado que estan esperando
 *      de las operaciones pendientes.
 *
 *----------------------------------------------------------------------
 */

static void
FPwait (machPtr, numChks, fpr1, fpr2, fpr3, fpr4)
     DLXV *machPtr;               /* Descripcion de la maquina. */
     long int numChks;            /* Numero de registros a comprobar. */
     int fpr1, fpr2, fpr3, fpr4;  /* Registros a comprobar.  Solamente los
				   * primeros n valores son examinados, donde
				   * n es el numero pasado en numChks. */
{
  register FPop *opPtr;
  long int latest, dest, tipo=0;

  opPtr = machPtr->FPopsList;
  latest = 0;
  while (opPtr != NULL) {
    dest = opPtr->dest;

    switch (opPtr->type) {
    case V_ADD:
    case V_DIV:
    case V_MUL:
    case V_LOGIC:
    case V_LOAD:
    case V_SUB:
      switch (numChks) {
      case 4: if (dest == fpr4) {latest = opPtr->ready; break;}
      case 3: if (dest == fpr3) {latest = opPtr->ready; break;}
      case 2: if (dest == fpr2) {latest = opPtr->ready; break;}
      case 1: if (dest == fpr1) latest = opPtr->ready;
      }
      tipo = opPtr->type;
      break;
    }

    switch (opPtr->resultType) {
    case FP_DOUBLE_FP_OP:
      switch (numChks) {
      case 4: if ((dest == fpr4) || ((dest+1) == fpr4)) {
	latest = opPtr->ready; break;}
      case 3: if ((dest == fpr3) || ((dest+1) == fpr3)) {
	latest = opPtr->ready; break;}
      case 2: if ((dest == fpr2) || ((dest+1) == fpr2)) {
	latest = opPtr->ready; break;}
      case 1: if ((dest == fpr1) || ((dest+1) == fpr1))
	latest = opPtr->ready;
      }
      tipo = opPtr->type;
      break;
    case FP_SINGLE_FP_OP:
    case FP_INT_OP:
      switch (numChks) {
      case 4: if (dest == fpr4) {latest = opPtr->ready; break;}
      case 3: if (dest == fpr3) {latest = opPtr->ready; break;}
      case 2: if (dest == fpr2) {latest = opPtr->ready; break;}
      case 1: if (dest == fpr1) latest = opPtr->ready;
      }
      tipo = opPtr->type;
      break;
    }

    opPtr = opPtr->nextPtr;
  }

  if (latest) {
    while (machPtr->cycleCount < latest) {
      if ((tipo >= V_ADD) && (tipo <= V_SUB))
       {
	machPtr->Vstalls += (machPtr->checkFP - machPtr->cycleCount);
       }
       else
	{
	 machPtr->FPstalls += (machPtr->checkFP - machPtr->cycleCount);
	}
      machPtr->cycleCount = machPtr->checkFP;
      FPwriteBack(machPtr);
    }
  }
}

static FPop *FPfree = NULL;

/*--------------------------------------------------------------------
 *
 * Terminar--
 *      Procedimiento llamado cuando se produce la trap #6
 *
 * Resultados:
 *      Todas las operaciones pendientes son finalizadas y
 *      el numero de ciclos se incrementa en consecuencia.
 *
 * Efectos Laterales:
 *      Se actualizan todas las estructuras asociadas con las
 *      operaciones pendientes.
 *
 *---------------------------------------------------------------
 */
static void
Terminar(machPtr)
     DLXV *machPtr;                /* Maquina a simular. */
{
  register FPop *opPtr;
  long int latest;

  opPtr = machPtr->FPopsList;
  latest = 0;
  while (opPtr != NULL) {
    latest = opPtr->ready;
    opPtr = opPtr->nextPtr;
  }
  if (latest) {
      machPtr->Terminar = latest - machPtr->cycleCount - 1;
      machPtr->cycleCount = latest - 1;
    }
}


/*
 *----------------------------------------------------------------------
 *
 * mallocFPop --
 *
 *      Devuelve un puntero a una estructura FPop no utilizada.
 *
 * Resultados:
 *      Puntero a estructura FPop.
 *
 * Efectos laterales:
 *
 *----------------------------------------------------------------------
 */

static FPop *
mallocFPop()
{
    FPop *p;

    if (!FPfree)
	return (FPop far *)farmalloc(sizeof(FPop));
    p = FPfree;
    FPfree = FPfree->nextPtr;
    return p;
}

/*
 *----------------------------------------------------------------------
 *
 * freeFPop --
 *
 *      Libera lo que esta apuntando a
 *
 * Resultados:
 *      Ver arriba
 *
 * Efectos Laterales:
 *
 *----------------------------------------------------------------------
 */

static void
freeFPop(p)
FPop *p;
{
    p->nextPtr = FPfree;
/*    FPfree = p;*/
    farfree(p);
/*    FPfree = NULL;*/
}
