/*
 * cop0.c --
 *
 *	Este fichero implementa un subconjunto de las facilidades
 *	del coprocesador 0 del DLXV para el DLXVsim.
 *	Por ahora solo se ha realizado la implementacion para manejar
 *	interrupciones simples de E/S.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <stdio.h>
#include <bwcc.h>
#include "dlx.h"


#define VECTOR_ADDRESS		0x80L

/*
 * Declaraciones de procedimientos definidos mas adelante en este fichero:
 */

static void	Interrupt();

/*
 *----------------------------------------------------------------------
 *
 * Cop0_Init --
 *
 *	Este procedimiento es llamado cuando se crea la maquina para
 *	inicializar el estado del coprocesador 0.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	Todo aquello que conlleva la inicializacion del estado.
 *
 *----------------------------------------------------------------------
 */

void
Cop0_Init(machPtr)
    register DLXV *machPtr;		/* Maquina que esta siendo creada. */
{
    int i;

    machPtr->cop0.cause = 0;
    machPtr->cop0.status = 0;
    machPtr->cop0.epc = 0;
    for (i = 0; i < COP0_NUM_LEVELS; i++) {
	machPtr->cop0.pending[i] = 0;
    }
    machPtr->cop0.pendingMask = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * Cop0_IntPending --
 *
 *	Este procedimiento es llamado para registrar cambios en el estado
 *	de interrupciones pendientes.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	El numero de interrupciones pendientes en el nivel "level" es
 *	incrementado segun "change". Si el nuevo numero es mayor que cero
 *	y el correspondiente nivel de interrupcion esta habilitado, entonces
 *	se produce una interrupcion de E/S.
 *
 *----------------------------------------------------------------------
 */

void
Cop0_IntPending(machPtr, level, change)
    register DLXV *machPtr;		/* Maquina a manipular. */
    long int level;			/* Nivel de interrupcion (0-5). */
    long int change;			/* 1 significa que una nueva fuente de
					 * interrupcion esta preparada; -1
					 * significa que la fuente de interrupcion
					 * no esta preperada. */
{
    long int old, new;
    extern char far salidas[199];

    old = machPtr->cop0.pending[(unsigned int)level];
    new = old + change;
    machPtr->cop0.pending[(unsigned int)level] = new;

    if (new < 0) {
	wsprintf(salidas, "Hey, level %ld just went negative!\n", level);
	BWCCMessageBox(GetFocus(), salidas, "Mensaje: Cop0",MB_OK | MB_ICONINFORMATION);
/*	printf("Hey, level %ld just went negative!\n", level);*/
    }
    if ((old <= 0) && (new > 0)) {
	machPtr->cop0.pendingMask |= COP0_STATUS_IMASK0 << level;
	if (machPtr->cop0.status & machPtr->cop0.pendingMask) {
	    Sim_CallBack(machPtr, cero, Interrupt, (ClientData) machPtr);
	}
    } else if ((old > 0) && (new <= 0)) {
	machPtr->cop0.pendingMask &= ~(COP0_STATUS_IMASK0 << level);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Interrupt --
 *
 *	Este es un procedimiento de llamada retardada (Sim_CallBack) que
 *	es invocado a veces cuando PUEDE QUE haya una interrupcion pendiente.
 *	Este procedimiento comprueba si una interrupcion debe ser atendida en
 *	ese momento; si es asi, la atiende.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	El estado de la maquina puede ser actualizado para reflejar que se ha
 *	atendido una interrupcion.
 *
 *----------------------------------------------------------------------
 */

static void
Interrupt(machPtr)
    register DLXV *machPtr;	/* Maquina en la que comprobar las
				 * interrupciones. */
{
    if (!(machPtr->cop0.status & COP0_STATUS_IEC)) {
	return;
    }
    if ((machPtr->cop0.status & machPtr->cop0.pendingMask) == 0) {
	return;
    }
    machPtr->cop0.cause = (machPtr->cop0.status & machPtr->cop0.pendingMask)
	    | COP0_CAUSE_CODE_INT;
    if (machPtr->branchSerial == (machPtr->insCount-1)) {
	machPtr->cop0.epc = machPtr->branchPC;
	machPtr->cop0.cause |= COP0_CAUSE_BD;
    } else {
	machPtr->cop0.epc = Sim_GetPC(machPtr);
    }
    if (machPtr->cop0.status & COP0_STATUS_IEP) {
	machPtr->cop0.status |= COP0_STATUS_IE0;
    } else {
	machPtr->cop0.status &= ~COP0_STATUS_IE0;
    }
    if (machPtr->cop0.status & COP0_STATUS_IEC) {
	machPtr->cop0.status |= COP0_STATUS_IEP;
    } else {
	machPtr->cop0.status &= ~COP0_STATUS_IEP;
    }
    machPtr->cop0.status &= ~COP0_STATUS_IEC;
    machPtr->regs[(int)PC_REG] = ADDR_TO_INDEX(VECTOR_ADDRESS);
    machPtr->regs[(int)NEXT_PC_REG] = machPtr->regs[(int)PC_REG] + 1;
}
