/*
 * cop0.h --
 *
 *	Declaraciones de aquellas cosas que son exportadas por el modulo
 *	cop0 al resto del simulador.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#ifndef _COP0
#define _COP0

/*
 * La siguiente estructura encapsula el estado del coprocesador 0 de la
 * mauiqna DLXV. Nota: En la actualidad solamente unas pocas facilidades
 * del coprocesador 0 estan realmente implementadas.
 */

#define COP0_NUM_LEVELS 6L

typedef struct Cop0 {
    long int cause;		/* Registro de origen, identifica cual ha sido
				 * la ultima interrupcion. */
    long int status;		/* Registro de estado, usado junto con otras
				 * cosas para habilitar las interrupciones. */
    long int epc;		/* Registro EPC :  Donde ocurrio la
				 * interrupcion. */
    long int pending[COP0_NUM_LEVELS];
				/*  Cuantas interrupciones estan pendientes en
				 * cada nivel definido? */
    long int pendingMask;	/* Mascara de los niveles con pending > 0. */
    long int flags;		/* Varios flags:  Ver abajo. */
} Cop0;

/*
 * Definiciones de varios bits en los registros cause y status. Ver paginas
 * 5-4 y siguientes del Kane para mas detalles.
 */

#define COP0_STATUS_REGNUM	12L
#define COP0_STATUS_IMASK0	0x400L
#define COP0_STATUS_IE0		0x10L
#define COP0_STATUS_IEP		0x4L
#define COP0_STATUS_IEC		0x1L
#define COP0_STATUS_IGNORE	0xffff03eaL

#define COP0_CAUSE_REGNUM	13L
#define COP0_CAUSE_BD		0x80000000L
#define COP0_CAUSE_IP0		0x400L
#define COP0_CAUSE_CODE_INT	0x0L
#define COP0_CAUSE_IGNORE	0x7fff03c3L

#define COP0_EPC_REGNUM		14L

/*
 * Procedimientos exportados:
 */

extern void	Cop0_Init();
extern void	Cop0_IntPending();
extern long int	Cop0_ReadReg();
extern void	Cop0_Rfe();
extern void	Cop0_WriteReg();

#endif /* _COP0 */
