/*
 * dlx.h --
 *
 *	Declaraciones de estructuras usadas para simular la
 *	arquitectura DLXV.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 *
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#ifndef _MIPS
#define _MIPS

#define MEMSIZE 65536L

#ifndef _TCL
#include "tcl.h"
#endif
#ifndef _HASH
#include "hash.h"
#endif
#ifndef _MIPSIM_IO
#include "io.h"
#endif
#ifndef _COP0
#include "cop0.h"
#endif

#define cero	0L
#define uno	1L
#define muno   -1L
#define dos	2L
#define tres	3L
#define cuatro	4L

#define argus	15
#define largus	80
#define MVL	64L

/*
 * La siguiente estructura es utilizada para cada "stop" que ha sido
 * solicitado en la maquina simulada.
 */

typedef struct Stop {
    char *command;		/* Comando a ejecutar, o NULL si
				 * es un simple stop. */
    long int number;		/* Numero que identifica a este stop para
				 * cuando tenga que ser borrado. */
    unsigned long int address;	/* Direccion (en la memoria DLXV ) de la
				 * palabra de memoria asociada con este
				 * stop. */
    struct Stop *nextPtr;	/* Siguiente en la lista de stops
				 * asociados con la misma posicion de
				 * memoria (NULL para fin lista). */
    struct Stop *overallPtr;	/* Siguiente en la lista de todos
				 * los stops utilizados en la maquina
				 * (NULL para fin lista). */
} Stop;

/*
 * Cada palabra de memoria esta representada por una estructura con el
 * siguiente formato. Para interpretar las instrucciones eficientemente,
 * ellas son decodificadas en numerosos campos cuando se produce la primera
 * ejecucion tras haber sido cambiada la palabra de memoria.
 */

typedef struct {
    long int value;		/* Contenido de la posicion de memoria. */
    unsigned char opCode;	/* Tipo de instruccion.  Este no es el
				 * mismo campo opcode de la instruccion:
				 * Ver los #defines de abajo para mas
				 * detalles */
    char rs1, rs2, rd;		/* Los tres registros de la instruccion */
    long int extra;		/* Campo del valor inmediato o destino o
				 * o offset.  Inmediatos son sign-extended. */
    Stop *stopList;		/* Lista de stops a invocar siempre que esta
				 * posicion de memoria sea accedida. */
} MemWord;

/*
 * Para cada llamada retardada (Callback) registrada a traves de Sim_CallBack,
 * hay una estructura de la siguiente forma:
 */

typedef struct CallBack {
    long int serialNum;		/* Llamar al procedimiento despues de ejecutar
				 * la instruccion con este numero de serie. */
    void (*proc)();		/* Procedimiento a llamar. */
    ClientData clientData;	/* Argumento a pasar a proc. */
    struct CallBack *nextPtr;	/* Siguiente llamada retrasada(callback) en
				 * la lista de todas aquellas asociadas con
				 * esta maquina.  NULL significa final de la
				 * lista. */
} CallBack;

/*
 * Para cada operacion pendiente, hay una estructura de la siguiente forma:
 */

typedef struct FPop {
  long int type;               /* tipo de operacion */
  long int unit;               /* numero de unidad */
  long int dest;               /* registro destino */
  long int resultType;         /* Ver los valores *_OP */
  long int longitud;	       /* longitud del vector */
  long result[MVL*2];          /* result[0] es el resultado de una operacion
				* en simple precision, 2 elementos son
				* usados para doble precision, y el resto son
				* usados en operaciones vectoriales. */
  long int ready;              /* ciclo en el que el resultado estara
				* disponible */
  long int ready2;	       /* ciclo en el que se puede ir entregando un
				* resultado por ciclo de reloj. */
  long int latcarga;	       /* latencia de la unidad de carga y
				* almacenamiento vectorial */
  long int ultimo;	       /* Cuenta el numero de elementos del vector que
				* han sido entregados hasta este momento. */
  long int cada;	       /* Indica cada cuantos elementos se produce un
				* un stall en la carga o almacenamiento de
				* vectores. */
  long int direccion;	       /* Direccion de destino u origen de las
				* operaciones de carga/almacenamiento
				* vectoriales. */
  long int stride;	       /* Stride de la operacion vectorial */
  int k;		       /* variable auxiliar */
  int f1, f2;		       /* registros fuentes de las op. vectoriales. */
  int d;                       /* registro destino de las op. vectoriales */
  struct FPop *nextPtr;        /* Siguiente FPop en la lista todos aquellos
				* asociados con esta maquina.  NULL
				* significa final de la lista. */
} FPop;

/*
 * La siguiente estructura describe el estado de la maquina DLXV.
 */

#define R_OK	   04L
#define TOTAL_REGS 69L
#define NUM_GPRS 64L
#define PC_REG (NUM_GPRS)
#define NEXT_PC_REG (NUM_GPRS+1)
#define VLR_REG (NEXT_PC_REG+1)
#define VMREG 40L  /* El VMREG es tiene como ndice el que sigue a los 32 regs de coma flotante + 8 regs vectoriales = 40
		      ver waiting_FPRs por ejemplo */
                   /* Sin embargo, se almacena fisicamente en regs[VLR+1] y regs[VLR+2] */
#define NUM_V_REGS	8L

/* Valores por defecto para numero de unidades y latencias */

#define MAX_FP_UNITS  20L        /* Numero maximo de unidades permitido
				  * para cualquier tipo. */
#define MAX_LD_UNITS	1L	 /* Maximo numero de unidades de
				  * carga/almacenamiento vectorial */
#define MAX_LG_UNITS	20L	 /* Maximo numero de unidades logicas
				  * vectoriales */
#define MAX_MM		64L      /* Mximo nmero de mdulos de memoria */
#define NUM_BANCOS	16L	 /* Nmero de mdulos de memoria */
#define NUM_ADD_UNITS 1L         /* Numero de unidades de suma fp. */
#define NUM_DIV_UNITS 1L         /* Numero de unidades de div fp. */
#define NUM_MUL_UNITS 1L         /* Numero de unidades de mult fp. */
#define NUM_VADD_UNITS 	1L	 /* Numero de unidades vect. de suma. */
#define NUM_VDIV_UNITS	1L	 /* Numero de unidades vect. de div. */
#define	NUM_VMUL_UNITS	1L	 /* Numero de unidades vect. de mult. */
#define NUM_LD_UNITS	1L	 /* Numero de unidades vectoriales carg/alm */
#define NUM_LG_UNITS	1L	 /* Numero de unidades logicas vectoriales */
#define FP_ADD_LATENCY 2L        /* Latencia para la suma FP. */
#define FP_DIV_LATENCY 19L       /* Latencia para la div  FP. */
#define FP_MUL_LATENCY 5L        /* Latencia para la mult FP. */
#define V_ADD_PEN	6L	 /* Start up penalty de unidad vect. de suma */
#define V_DIV_PEN	20L	 /* Start up penalty de unidad vect. de
				  * division */
#define V_MUL_PEN	7L	 /* Start up penalty de unidad vect.
				  * multiplicacion */
#define V_LD_PEN	12L	 /* Start up penalty de unidad vect de
				  * carga/almacenamiento */
#define V_LG_PEN	1L	 /* Start up penalty de unidad logica
				  * vectorial */
#define V_PEN_HZRD	0L	 /* Penalizacion por dependencia entre
				  * instrucciones vectoriales*/
                                 // Hennessy: 4 ciclos
#define RPORTS		2	/* Numero de ports de lectura de los
				registros vectoriales. */
#define WPORTS		1	/* Numero de ports de escritura de los
				registros vectoriales. */

#define FP_ADD 0L
#define FP_DIV 1L
#define FP_MUL 2L
#define V_ADD  3L
#define V_DIV  4L
#define V_MUL  5L
#define V_LOAD 6L
#define V_LOGIC 7L
#define V_SUB  8L
#define FP_SUB 9L
#define INT_MUL 10L
#define INT_UMUL 11L
#define INT_DIV 12L
#define INT_UDIV 13L

#define FP_INT_OP 0L
#define FP_SINGLE_FP_OP 1L
#define FP_DOUBLE_FP_OP 2L
#define FP_VECT_OP	3L
#define FP_SV_OP	4L
#define FP_VS_OP	5L
#define FP_VEQ_OP	6L
#define FP_VNE_OP	7L
#define FP_VGT_OP	8L
#define FP_VLT_OP	9L
#define FP_VGE_OP	10L
#define FP_VLE_OP	11L
#define FP_SVEQ_OP	12L
#define FP_SVNE_OP	13L
#define FP_SVGT_OP	14L
#define FP_SVLT_OP	15L
#define FP_SVGE_OP	16L
#define FP_SVLE_OP	17L
#define FP_MASK_OP	18L
#define FP_LOAD_OP	19L
#define FP_STORE_OP	20L
#define FP_LSTD_OP	21L
#define FP_SSTD_OP	22L
#define FP_LINDEX_OP	23L
#define FP_SINDEX_OP	24L


#define CYC_CNT_RESET 16384L     /* Cada cuanto se ha de resetear el contador
				  * de ciclos para prevenir problemas de wrap
				  * around. */
#define FD_SIZE 32L              /* Numero de descriptores de fichero
				  * simulados. */

/*
 * Valores OpCode para la estructura MemWord .
 * Ademas tenemos los suiguientes codigos especiales:
 *
 * OP_NOT_COMPILED -	Significa que el valor de la posicion de memoria ha
 *			cambiado por lo que la instruccion necesita ser
 *			recompilada
 * OP_UNIMP -		Significa que la instruccion es legal, pero que
 *			todavia no ha sido implementada en el simulador,
 * OP_RES -		Es un opcode reservado (No esta soportado por la
 *			arquitectura).
 */

#define OP_ADD 		 1L
#define OP_ADDI 	 2L
#define OP_ADDU 	 3L
#define OP_ADDUI 	 4L
#define OP_AND 		 5L
#define OP_ANDI 	 6L
#define OP_BEQZ 	 7L
#define OP_BFPF 	 8L
#define OP_BFPT 	 9L
#define OP_BNEZ 	 10L
#define OP_J 		 11L
#define OP_JAL 		 12L
#define OP_JALR 	 13L
#define OP_JR 		 14L
#define OP_LB 		 15L
#define OP_LBU 		 16L
#define OP_LD 		 17L
#define OP_LF 		 18L
#define OP_LH 		 19L
#define OP_LHI 		 20L
#define OP_LHU 		 21L
#define OP_LW 		 22L
#define OP_MOVD 	 23L
#define OP_MOVF 	 24L
#define OP_MOVFP2I 	 25L
#define OP_MOVI2FP 	 26L
#define OP_MOVI2S 	 27L
#define OP_MOVS2I 	 28L
#define OP_OR 		 29L
#define OP_ORI 		 30L
#define OP_RFE 		 31L
#define OP_SB 		 32L
#define OP_SD 		 33L
#define OP_SEQ 		 34L
#define OP_SEQI 	 35L
#define OP_SEQU 	 36L
#define OP_SEQUI 	 37L
#define OP_SF 		 38L
#define OP_SGE 		 39L
#define OP_SGEI 	 40L
#define OP_SGEU 	 41L
#define OP_SGEUI 	 42L
#define OP_SGT 		 43L
#define OP_SGTI 	 44L
#define OP_SGTU 	 45L
#define OP_SGTUI 	 46L
#define OP_SH 		 47L
#define OP_SLE 		 48L
#define OP_SLEI 	 49L
#define OP_SLEU 	 50L
#define OP_SLEUI 	 51L
#define OP_SLL 		 52L
#define OP_SLLI 	 53L
#define OP_SLT 		 54L
#define OP_SLTI 	 55L
#define OP_SLTU 	 56L
#define OP_SLTUI 	 57L
#define OP_SNE 		 58L
#define OP_SNEI 	 59L
#define OP_SNEU 	 60L
#define OP_SNEUI 	 61L
#define OP_SRA 		 62L
#define OP_SRAI 	 63L
#define OP_SRL 		 64L
#define OP_SRLI 	 65L
#define OP_SUB 		 66L
#define OP_SUBI 	 67L
#define OP_SUBU 	 68L
#define OP_SUBUI 	 69L
#define OP_SW 		 70L
#define OP_TRAP 	 71L
#define OP_XOR 		 72L
#define OP_XORI 	 73L
#define OP_NOP       74L

/* Operaciones en coma flotante */

#define OP_ADDD 	 80L
#define OP_ADDF 	 81L
#define OP_CVTD2F 	 82L
#define OP_CVTD2I 	 83L
#define OP_CVTF2D 	 84L
#define OP_CVTF2I 	 85L
#define OP_CVTI2D 	 86L
#define OP_CVTI2F 	 87L
#define OP_DIV 		 88L
#define OP_DIVD 	 89L
#define OP_DIVF 	 90L
#define OP_DIVU 	 91L
#define OP_EQD 		 92L
#define OP_EQF 		 93L
#define OP_GED 		 94L
#define OP_GEF 		 95L
#define OP_GTD 		 96L
#define OP_GTF 		 97L
#define OP_LED 		 98L
#define OP_LEF 		 99L
#define OP_LTD 		 100L
#define OP_LTF 		 101L
#define OP_MULT 	 102L
#define OP_MULTD 	 103L
#define OP_MULTF 	 104L
#define OP_MULTU 	 105L
#define OP_NED 		 106L
#define OP_NEF 		 107L
#define OP_SUBD 	 108L
#define OP_SUBF 	 109L

/* Operaciones vectoriales aadidas */

#define OP_ADDSV	 111L
#define OP_ADDV		 112L
#define OP_CVI		 113L
#define OP_CVM		 114L
#define OP_DIVSV	 115L
#define OP_DIVV		 116L
#define OP_DIVVS	 117L
#define OP_LV		 118L
#define OP_LVI		 119L
#define OP_LVWS		 120L
#define OP_MOVF2S	 121L
#define OP_MOVS2F	 122L
#define OP_MULTSV	 123L
#define	OP_MULTV	 124L
#define OP_POP		 125L
#define OP_SEQSV	 126L
#define OP_SEQV		 127L
#define OP_SGESV	 128L
#define OP_SGEV		 129L
#define OP_SGTSV	 130L
#define OP_SGTV		 131L
#define OP_SLESV	 132L
#define OP_SLEV		 133L
#define OP_SLTSV	 134L
#define OP_SLTV		 135L
#define OP_SNESV	 136L
#define OP_SNEV		 137L
#define	OP_SUBSV	 138L
#define OP_SUBV		 139L
#define OP_SUBVS	 140L
#define OP_SV		 141L
#define OP_SVI		 142L
#define OP_SVWS		 143L

/* "Opcodes" especiales, hay que dar estos valores despues de los
 * anteriores */

#define OP_NOT_COMPILED  144L
#define OP_UNIMP 	 145L
#define OP_RES 		 146L
#define OP_LAST 	 147L

typedef struct {
    long int numWords;		/* Numero de palabras de memoria simuladas en
				 * la maquina. */
    long int numChars;          /* Numero de caracteres de memoria */
    MemWord huge *memPtr;	/* Array de MemWords, suficiente para
				 * proporcionar memSize bytes de
				 * almacenamiento. */
    char huge *memScratch;      /* Scratchpad a usar por los manejadores
				 * de excepciones. */
    char huge *endScratch;      /* Puntro al final del scratch pad */
    long int fd_map[FD_SIZE];   /* mapea los descriptores de fichero simulados
				 * con los df reales. Principalmente protege
				 * la entrada, salida, y salida de error
				 * estandar del programa. Un valor de -1
				 * significa que el df no esta en uso. */
    long int regs[TOTAL_REGS];	/* Registros de proposito general, seguidos por
				 * los registros el contador de programa y el
				 * npc, el registro vlr, el vm y vmlo
				 * Ambos pc's son almacenados como indices
				 * del array memPtr. */
    unsigned long int badPC;	/* Si ocurre un error de direccionamiento
				 * durante la busqueda de la instruccion, Esta
				 * variable almacena la direccion erronea.
				 * cero signifia que no hay error pendiente
				 * por direccion erronea. */
    long int addrErrNum;	/* Si badPC es no cero, Este da el numero de
				 * serie (valor de insCount) de la instruccion
				 * tras la cual se registro un error de
				 * direccionamiento. */
    long int loadReg1, loadReg2;/* Para cargas, registro cargado por la
				 * ultima instruccion. Usada para los ciclos de
				 * parada ( 0 significa que la ultima
				 * instruccion no fue de carga). loadReg2 es
				 * usado cuando se craga un valor de doble
				 * precision. */
    long int lastRdst;		/* Registro destino de la ultima instruccion. Usada
				   para los ciclos de parada en riesgos RAW en los
                                   que la 2a instruccion es un salto */ 
    long int loadValue1, loadValue2; /* Para cargas, valor del registro a
				 * cargar en el siguiente ciclo. */
    long int insCount;		/* Contador del # total de instrucciones
				 * ejecutadas en esta maquina ( tambien se
				 * utiliza como numero de serie de la
				 * instruccion actual. */
    long int firstIns;		/* Numero de serie correspondiente a la primera
				 * instruccion ejecutada; usada para ignorar
				 * los stops en la primera instruccion. */
    long int branchSerial;	/* Numero de serie de la instruccion mas
				 * reciente de salto; usada para poner el bit
				 * BD durante las excepciones. */
    long int branchPC;		/* PC de la instruccion dada por
				 * "branchSerial":
				 * Tambien usada durante las excepciones. */
    long int flags;		/* Usada para indicar condiciones especiales
				 * durante la simulacion (para mayor velocidad,
				 * debe ser normalmente cero). Ver abajo para
				 * las definiciones. */
    long int stopNum;		/* Usado para asignar los numeros de referencia
				 * de los stops de forma incremental. */
    Stop *stopList;		/* Primero en la cadena de todos los stops
				 * asociados con esta maquina (NULL significa
				 * ninguno). */
    CallBack *callBackList;	/* Primero en la lista de todas la llamadas
				 * retardadas actualmente registradas para esta
				 * maquina, ordenadas en orden creciente por
				 * SerialNum. */
    Hash_Table symbols;		/* Guarda las direcciones de todos los simbolos
				 * leidos por la maquina al ensamblar. */
    IoState ioState;		/* Informacion relacionada con la E/S de la
				 * maquina (Ver io.h e io.c para mas detalles*/
    Cop0 cop0;			/* Estado del coprocesador 0 (Ver cop0.h y
				 * cop0.c para mas detalles). */

    /* Estadisticas */

    long int loadStalls;          /* Contador de ciclos de parada por carga
				      *	de operaciones escalares */
    long int FPstalls;          /* Contador de ciclos de parada por dependencia
				 * de operaciones escalares. */
    long int Vstalls;		/* Contador de ciclos de parada por dependencia
				 * de operaciones vectoriales. */
    long int branchYes;              /* Contador de los saltos tomados. */
    long int branchNo;               /* Contador de los saltos no tomados. */
    long int Terminar;		     /* Contador de ciclos adicionales cuando
				      * se utiliza la trap #6. */
    long int	operationCount
      [OP_LAST+1];              /* Contador dinamico de instrucciones. */


    /* Estructuras relacionadas con las unidades de proceso */

    int	read_ports;		/* Numero de puertos de lectura de los
				 * registros vectoriales. */
    int rports[NUM_V_REGS];	/* Nmero de puertos de lectura ocupados
				 * por los registros vectorales. */
    int	write_ports;		/* Numero de puertos de escritura de los
				 * registros vectoriales. */
    int wports[NUM_V_REGS];	/* Nmero de puertos de lectura ocupados
				 * por los registros vectorales. */
    long int bancos;		/* Numero de bancos de memoria */
    long int num_add_units;     /* Numero de unidades de suma en coma
				 * flotante para esta maquina. */
    long int num_div_units;     /* Numero de unidades en PF de division para
				 * esta maquina. */
    long int num_mul_units;     /* Numero de unidades en PF de multiplicacion
				 * para esta maquina. */
    long int num_addv_units;	/* Numero de unidades vectoriales de suma para
				 * esta maquina. */
    long int num_divv_units;    /* Numero de unidades vectoriales de division
				 * para esta maquina. */
    long int num_mulv_units;	/* Numero de unidades vectoriales de
				 * multiplicacion para esta maquina. */
    long int num_loadv_units;	/* Numero de unidades vectoriales de
				 * carga/almacenamiento para esta maquina. */
    long int num_logic_units;	/* Numero de unidades logicas vectoriales
				 * para esta maquina. */

    long int fp_add_latency;    /* Latencia para suma/resta en PF de esta
				 * maquina. */
    long int fp_div_latency;    /* Latencia para division en PF de esta
				 * maquina. */
    long int fp_mul_latency;    /* Latencia para multiplicacion en PF de
				 * esta maquina. */
    long int v_add_latency;	/* Start up penalty de la unidad vectorial
				 * de suma. */
    long int v_div_latency;	/* Start up penalty de la unidad vectorial
				 * de division. */
    long int v_mul_latency;	/* Start up penalty de la unidad vectorial
				 * de multiplicacion. */
    long int v_load_latency;	/* Start up penalty de la unidad vectorial
				 * de carga/almacenamiento. */
    long int v_logic_latency;	/* Start up penalty de la unidad logica
				 * vectorial. */


    long int FPstatusReg;       /* Registro de estado de las operaciones
				 * en coma flotante. Este registro es
				 * modificado en las comparaciones en
				 * coma flotante, es leido en las
				 * instrucciones BFPT y BFPF. */
    long int fp_add_units
      [MAX_FP_UNITS];         /* 0 significa que la unidad esta disponible
			       * en otro caso indica el ciclo en el que
			       * la unidad de suma completara la operacion
			       * en curso. */
    long int fp_div_units
      [MAX_FP_UNITS];	      /* 0 significa que la unidad esta disponible
			       * en otro caso indica el ciclo en el que
			       * la unidad de division completara la operacion
			       * en curso. */
    long int fp_mul_units
      [MAX_FP_UNITS];         /* 0 significa que la unidad esta disponible
			       * en otro caso indica el ciclo en el que
			       * la unidad de multiplicacion suma completara
			       * la operacion en curso. */
    long int v_add_units
      [MAX_FP_UNITS];	      /* 0 indica unidad disponible, cualquier otro numero
				 indica el numero de ciclos tras los cuales la unidad
				 de suma completara la operacion actual */
    long int v_div_units
      [MAX_FP_UNITS];         /* 0 indica unidad disponible, cualquier otro numero
				 indica el numero de ciclos tras los cuales la unidad
				 de division completara la operacion actual */
    long int v_mul_units
      [MAX_FP_UNITS];         /* 0 indica unidad disponible, cualquier otro numero
				 indica el numero de ciclos tras los cuales la unidad
				 de multiplicacion completara la operacion actual */
    long int v_load_units
      [MAX_LD_UNITS];         /* 0 indica unidad disponible, cualquier otro numero
				 indica el numero de ciclos tras los cuales la unidad
				 de carga completara la operacion actual */
    long int v_logic_units
      [MAX_LG_UNITS];         /* 0 indica unidad disponible, cualquier otro numero
				 indica el numero de ciclos tras los cuales la unidad
				 logica (de comparacion de vectores) completara la
				 operacion actual */

    long int *fp_units[3+5];  /* Estructura en array utilizada para acceder a las
				 fp_add_units, fp_div_units, fp_mul_units,
				 y todas las unidades vectoriales aadidas */

    long int waiting_FPRs[32 + NUM_V_REGS + 1]; /* Un valor no-cero en el
			       * elemento i significa que el registro i esta
			       * esperando un resultado de una unidad de
			       * proceso. El valor indica el ciclo en el que
			       * el resultado estara listo. */

    long int waitingE[32 + NUM_V_REGS + 1];
				/* Esta estructura se utiliza como la
				 * anterior estructura pero cuando hay
				 * concatenacion de operaciones vectoriales. */

    long int waitS[NUM_V_REGS];	/* Esta estructura se utiliza junto con el
				 * nmero de puertos de lectura de los
				 * registros vectoriales. */

    long int vecregs[NUM_V_REGS][MVL*2]; /* Representacion de los registros
					  * vectoriales presentes en la
					  * maquina simulada */

    FPop *FPopsList;          /* Primero de una lista de todas las operaciones
			       * pendientes asociadas a esta maquina; NULL
			       * significa que no hay ninguna. */

    long int cycleCount;      /* Mantenemos el numero de ciclos que han sido
			       * ejecutados, para asi poder monitorizar las
			       * unidades de proceso. */

    long int checkFP;         /* Similar a la variable checkNum en sim.c.
			       * Representa el siguiente ciclo en el que una
			       * unidad de proceso acabara una operacion.
			       * Esto permite un rapido chequeo. */

    FILE *refTraceFile;       /* Puntero a un fichero en el cual escribir
			       * las referencias de memoria para el simulador
			       * de memoria de caches. */

   /* Variables utilizadas para la comunicacion de mensajes de error que se
    * producen o pueden producirse en las diferentes etapas del simulador. */

    long int dynamic;
    long int errorLine;
    char huge *result;
    char resultSpace[TCL_RESULT_SIZE];

} DLXV;

/*
 * Valores de los flags para las estructuras del DLXV:
 *
 * STOP_REQUESTED:		1 significa que el comando "Stop" ha sido
 *				ejecutado y la ejecucion debe ser detenida.
 *
 */

#define STOP_REQUESTED		0x4L

/*
 * Conversion entre direcciones DLXV e indices del array, los cuales son
 * almacenados en los campos pc/nextPc de la estructura del DLXV y son
 * tambien utilizados para direccionar los valores de memPtr:
 */

#define INDEX_TO_ADDR(index)	((unsigned long int) (index << 2))
#define ADDR_TO_INDEX(addr)	(addr >> 2)

/*
 * Definiciones varias:
 */

#define SIGN_BIT	0x80000000L
#define R31		31L


/*
 * Comandos proporcionados por el simulador:
 */

extern long int Gp_GetCmd(), Gp_PutCmd();
extern long int Gp_FGetCmd(), Gp_FPutCmd();
extern long int Gp_Vget();
extern long int Sim_GoCmd(), Sim_StepCmd();
extern long int Stop_StopCmd(), Sim_DumpStats();
extern long int Sim_TraceCmd();

/*
 * Otros procedimientos que son exportados de un fichero a otro:
 */

extern void		Sim_CallBack();
extern DLXV *		Sim_Create();
extern unsigned long int	Sim_GetPC();
extern void		Sim_Stop();

extern long int		Stop_Execute();
extern long int		Elegir();
extern long int		Simulate();
#endif /* _MIPS */
