/* hash.h --
 *
 * 	Este fichero contiene definiciones usadas por el modulo hash,
 * 	el cual mantiene las tablas hash.
 *
 * Copyright 1986, 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#ifndef	_HASH
#define	_HASH

#ifndef _LIST
#include "list.h"
#endif
#ifndef _SPRITE
#include "sprite.h"
#endif

/*
 * La siguiente estructura define una entrada de la tabla hash.
 */

typedef struct Hash_Entry {
    List_Links	      links;		/* Usado para reunir todas las
					 * entradas asociadas con el mismo
					 * bucket. */
    ClientData	      clientData;	/* Informacion de tamao arbitrario
					 * asociada con la clave. */
    union {
	Address	 ptr;			/* Valor de la clave, de una sola
					 * palabra, para identificar la
					 * entrada. */
					/* clave = key */
	long int words[1];		/* Valor de la clave de N palabras
					 * Nota: el tamao actual puede ser
					 * mas grande si es necesario almacenar
					 * la clave completa. */
	char 	 name[4];		/* Nombre (tipo texto) de esta entrada
					 * El tamao actual puede ser mayor si
					 * es necesario almacenar la cadena
					 * completa. Esta DEBE ser la ultima
					 * entrada de la estructura. */
    } key;
} Hash_Entry;

/*
 * Una tabla hash consiste en un array de punteros apuntando a listas hash.
 * Las tablas pueden ser organizadas de tres formas diferentes, dependiendo
 * del tipo de la calve que se utilice para la comparacion:
 *
 *	Strings:	  Terminadas por el caracter NULL; Su direccion
 *			  es pasada a HashFind como (char *).
 *	Claves de 1 Palabra: Pueden ser cualquier cosa, pero deben ser pasadas
 *			  a HashFind como una direccion.
 *	Claves multi palabra: Tambien pueden ser cualquier cosa, su direccion
 *			  se pasa a HashFind como una direccion.
 *
 *	Las claves de 1 palabra son las mas rapidas, pero tambien las mas
 *	restrictivas.
 */

#define HASH_STRING_KEYS	0L
#define HASH_ONE_WORD_KEYS	1L

typedef struct Hash_Table {
    List_Links 	*bucketPtr;	/* Puntero a un array de List_Links, uno por
				 * cada bucket en la tabla. */
    long int 	size;		/* Tamao real del array. */
    long int 	numEntries;	/* Numero de entradas en la tabla. */
    long int 	downShift;	/* Cuenta de desplazamiento, usado en la
				 * funcion hash.  */
    long int 	mask;		/* Usado para seleccionar los bits para
				 * realizar el ordenamiento hash. */
    long int 	keyType;	/* Tipo de claves usadas en la tabla:
				 * HASH_STRING_KEYS, HASH_ONE-WORD_KEYS,
				 * o >1 significa que keyType proporciona el
				 * numero de palabras que forman la clave. */
} Hash_Table;

/*
 * La siguiente estructura es utilizada por las rutinas de busqueda para
 * guardar el estado en el que se encuentra la busqueda.
 */

typedef struct Hash_Search {
    Hash_Table  *tablePtr;	/* Tabla en la que se esta buscando. */
    long int 	nextIndex;	/* Siguiente bucket a comprobar (despues
				 * del actual). */
    Hash_Entry 	*hashEntryPtr;	/* Siguiente entrada a comprobar en el
				 * bucket actual. */
    List_Links	*hashList;	/* Cadena hash que esta siendo comprobada
				 * actualmente. */
} Hash_Search;

/*
 * Macros utilizados.
 */

/*
 * ClientData Hash_GetValue(h)
 *     Hash_Entry *h;
 */

#define Hash_GetValue(h) ((h)->clientData)

/*
 * Hash_SetValue(h, val);
 *     HashEntry *h;
 *     char *val;
 */

#define Hash_SetValue(h, val) ((h)->clientData = (ClientData) (val))

/*
 * Hash_Size(n) devuelve el numero de palabras de un objeto de n bytes.
 */

#define	Hash_Size(n)	(((n) + sizeof (long int) - 1) / sizeof (long int))

/*
 * Las siguientes declaraciones de procedimientos y macros son las unicas
 * cosas que pueden ser necesarias fuera de los modulos de implementacion
 * de la tabla hash.
 */

Hash_Entry *	Hash_CreateEntry();
void		Hash_DeleteEntry();
void		Hash_DeleteTable();
Hash_Entry *	Hash_EnumFirst();
Hash_Entry *	Hash_EnumNext();
Hash_Entry *	Hash_FindEntry();
void		Hash_InitTable();
void		Hash_PrintStats();

#endif /* _HASH */
