/*
 * Hash_InitTable.c --
 *
 *	Codigo fuente para el procedimiento de libreria Hash_InitTable.
 *
 * Copyright 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <alloc.h>
#include "hash.h"
#include "list.h"

/*
 *---------------------------------------------------------
 *
 * Hash_InitTable --
 *
 *	Esta rutina inicializa la tabla hash.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	El espacio requerido por los buckets iniciales son alojados en
 *	memoria.
 *
 *---------------------------------------------------------
 */

void
Hash_InitTable(tablePtr, numBuckets, keyType)
    register Hash_Table *tablePtr;	/* Estructura usada para mantener la
					 * tabla. */
    long int 	        numBuckets;	/* Cuantos buckets se han de crear
					 * inicialmente. Este numero es
					 * redondeado a una potencia de 2
					 * Si es <=0 un vlor por defecto es
					 * utilizado. La tabla se hara mas
					 * grande si es necesario. */
    long int 	        keyType;	/* HASH_STRING_KEYS significa que el
					 * valor de la clave(key) que se pasa
					 * a HashFind es un string, que se pasa
					 * mediante un puntero (char *).
					 * HASH_ONE_WORD_KEYS significa que el
					 * valor de la clave es una palabra
					 * que se pasara como una direccion.
					 * > 1 siginifica que el valor de la
					 * clave es un valor de varias palabras
					 * pasadas como una direccion.
					 * En este ultimo caso, keyType es el
					 * numero de palabras que forman la
					 * clave, no el numero de bytes.
					 */
{
    register	long int 		i;
    register	List_Links 	*bucketPtr;

    /*
     * Redondea el tamao a una potencia de dos.
     */

    if (numBuckets <= 0) {
	numBuckets = 16;
    }
    tablePtr->numEntries = 0;
    tablePtr->keyType = keyType;
    tablePtr->size = 2;
    tablePtr->mask = 1;
    tablePtr->downShift = 29;
    while (tablePtr->size < numBuckets) {
	tablePtr->size <<= 1;
	tablePtr->mask = (tablePtr->mask << 1) + 1;
	tablePtr->downShift--;
    }

    tablePtr->bucketPtr = (List_Links far *) farcalloc(1, (unsigned long int) (sizeof(List_Links)
	    * tablePtr->size));
    for (i=0, bucketPtr = tablePtr->bucketPtr; i < tablePtr->size;
	    i++, bucketPtr++) {
	List_Init(bucketPtr);
    }
}
