/*
 * Hash_DeleteTable.c --
 *
 *	Codigo fuente para el procedimiento de libreria Hash_DeleteTable.
 *
 * Copyright 1988 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <alloc.h>
#include "hash.h"
#include "list.h"


/*
 *---------------------------------------------------------
 *
 * Hash_DeleteTable --
 *
 *	Esta rutina elimina de la tabla hash todo su contenido
 *	y libera el espacio de memoria que ocupaba (exceptuando
 *	el espacio ocupado por la estructura Hash_Table ).
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	Gran cantidad de memoria es liberada.
 *
 *---------------------------------------------------------
 */

void
Hash_DeleteTable(tablePtr)
    Hash_Table *tablePtr;		/* Tabla Hash cuyas entradas van
					 * a ser eliminadas.  */
{
    register List_Links *hashTableEnd;
    register Hash_Entry *hashEntryPtr;
    register List_Links *bucketPtr;

    bucketPtr = tablePtr->bucketPtr;
    hashTableEnd = &(bucketPtr[(unsigned int)tablePtr->size]);
    for (; bucketPtr < hashTableEnd; bucketPtr++) {
	while (!List_IsEmpty(bucketPtr)) {
	    hashEntryPtr = (Hash_Entry *) List_First(bucketPtr);
	    List_Remove((List_Links *) hashEntryPtr);
	    farfree((Address) hashEntryPtr);
	}
    }
    farfree((Address) tablePtr->bucketPtr);

    /*
     * Inicializa la tabla hash de manera que cause fallos de memoria en
     * cualquier acceso posterior mientras no se reinicialice.
     */

    tablePtr->bucketPtr = (List_Links *) NIL;
}
