/*
 * io.h --
 *
 *	Declaraciones para las facilidades de E/S proporcionadas como
 *	parte del simulador.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#ifndef _MIPSIM_IO
#define _MIPSIM_IO


/*
 * La siguiente estructura es parte de cada estructura DLXV, y describe
 * el estado de la E/S de la maquina:
 */

typedef struct IoState {
    char input;			/* Siguiente caracter de entrada. */
    long int flags;		/* Varios flags:  Ver abajo. */
} IoState;

/*
 * Valores de los flags utilizados en la estructura IoState:
 *
 * IO_TERM_INPUT_READY -	1 siginifica que hay un caracter valido
 *				en la entrada esperando en el campo "input".
 * IO_TERM_OUTPUT_READY -	1 significa que el buffer de salida del
 *				terminal esta vacio. 0 significa que un
 *				caracter ha sido recibido, pero estamos
 *				esperando unas cuantas instrucciones para
 *				simular la transmision del caracter.
 * IO_TERM_INPUT_IE - 		1 significa que las interrupciones estan
 *				habilitadas para la entrada.
 * IO_TERM_OUTPUT_IE -		1 significa que las interrupciones estan
 *				habilitadas para la salida.
 */

#define IO_TERM_INPUT_READY	1L
#define IO_TERM_OUTPUT_READY	2L
#define IO_TERM_INPUT_IE	4L
#define IO_TERM_OUTPUT_IE	8L

/*
 * Direcciones de los registros de dispositivo simulados:
 */

#define IO_RECV_CONTROL		0xfff0L
#define IO_RECV_DATA		0xfff4L
#define IO_TRANS_CONTROL	0xfff8L
#define IO_TRANS_DATA		0xfffcL

/*
 * Posiciones de los bits en los registros de dispositivo simulados:
 */

#define IO_READY	1L
#define IO_IE		2L

/*
 * Procedimientos exportados por io.c al resto del Mipsim:
 */

extern void	Io_BeginSim();
extern void	Io_EndSim();
extern void	Io_Init();
extern long int	Io_Read();
extern long int	Io_Write();
extern void	MemErr();

#endif /* _MIPSIM_IO */
