/*
 * tclUtil.c --
 *
 *	Este fichero contiene algunos procedimiento utiles para el
 *	procesamiento de los comandos.
 *
 * Copyright 1987, 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <alloc.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "dlx.h"

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Return --
 *
 *	Prepara el string para que sea el valor de retorno.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	machPtr->result queda apuntando al "string" (si "copy" es 0)
 *	o a una copia del string.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_Return(maq, string, status)
    DLXV *maq;		/* Maquina a la que asociar el valor devuelto. */
    char *string;	/* Valor a ser devuelto.  Si NULL, al resultado
			 * se le asigna un string vacio. */
    long int status;	/* Proporciona informacion sobre el string:
			 * TCL_STATIC, TCL_DYNAMIC, TCL_VOLATILE.
			 * Ignoredo si el string es NULL. */
{
    register DLXV *iPtr = (DLXV *) maq;
    long int length;
    long int wasDynamic = iPtr->dynamic;
    char huge *oldResult = iPtr->result;

    if (string == NULL) {
	iPtr->resultSpace[0] = 0;
	iPtr->result = iPtr->resultSpace;
	iPtr->dynamic = 0;
    } else if (status == TCL_STATIC) {
	iPtr->result = string;
	iPtr->dynamic = 0;
    } else if (status == TCL_DYNAMIC) {
	iPtr->result = string;
	iPtr->dynamic = 1;
    } else {
	length = _fstrlen(string);
	if (length > TCL_RESULT_SIZE) {
	    iPtr->result = (char far *) farmalloc((unsigned long int) length+1);
	    iPtr->dynamic = 1;
	} else {
	    iPtr->dynamic = 0;
	}
	strcpy((char far *)iPtr->result, string);
    }

    /*
     * Si el viejo resultado fue alojado dinamicamente, lo libera. Se hace
     * aqui, mas que al principio, por si acaso el nuevo resultado fuera
     * parte del viejo resultado.
     */

    if (wasDynamic) {
	farfree(oldResult);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_Backslash --
 *
 *	Indica como manejar las secuencias del backslash (simbolo "\").
 *
 * Resultados:
 *	El valor devuelto es el caracter que deberia substituir a la
 *	secuencia de backslash que comienza en src. Si readPtr no es
 *	NULL entonces es rellenado con el numero de caracteres que forman
 *	la secuencia backslash. Nota: Si el backslash no esta seguido por
 *	caracteres que puedan ser entendidos, entonces la secuencia de
 *	backslash es considerada como de longitud 1, y es reemplazada
 *	por el caracter backslash ("\").
 *
 * Efectos Laterales:
 *	Ninguno.
 *
 *----------------------------------------------------------------------
 */

char
Tcl_Backslash(src, readPtr)
    char *src;			/* Apunta al caracter "\" de la secuencia */
    long int *readPtr;		/* Rellenado con el numero de caracteres
				 * leidos de src, a menos que sea NULL. */
{
    register char *p = src+1;
    char result;
    long int count;

    count = 2;

    switch (*p) {
	case 'b':
	    result = '\b';
	    break;
	case 'e':
	    result = 033;
	    break;
	case 'n':
	    result = '\n';
	    break;
	case 't':
	    result = '\t';
	    break;
	case 'C':
	    p++;
	    if (isspace(*p) || (*p == 0)) {
		result = 'C';
		count = 1;
		break;
	    }
	    count = 3;
	    if (*p == 'M') {
		p++;
		if (isspace(*p) || (*p == 0)) {
		    result = 'M' & 037;
		    break;
		}
		count = 4;
		result = (*p & 037) | 0200;
		break;
	    }
	    count = 3;
	    result = *p & 037;
	    break;
	case 'M':
	    p++;
	    if (isspace(*p) || (*p == 0)) {
		result = 'M';
		count = 1;
		break;
	    }
	    count = 3;
	    result = *p + 0200;
	    break;
	case '}':
	case '{':
	case ']':
	case '[':
	case '$':
	case ' ':
	case ';':
	case '"':
	case '\\':
	    result = *p;
	    break;
	default:
	    if (isdigit(*p)) {
		result = *p - '0';
		p++;
		if (!isdigit(*p)) {
		    break;
		}
		count = 3;
		result = (result << 3) + (*p - '0');
		p++;
		if (!isdigit(*p)) {
		    break;
		}
		count = 4;
		result = (result << 3) + (*p - '0');
		break;
	    }
	    result = '\\';
	    count = 1;
	    break;
    }

    if (readPtr != NULL) {
	*readPtr = count;
    }
    return result;
}
